#ifdef KNOERRED
#include "knoerred.h"
#else
#include "knoerre.h"
#endif
#ifdef CONFIG_KNOERRE_CHECK_DIRENTRIES

/*
config KNOERRE_CHECK_DIRENTRIES
    bool "check_direntries"
    default y
    help
      This knoerre module counts directory entries.
      More information offers man-page knoerre(1).

config KNOERRE_CHECK_DIRENTRIES_KEY
    string "key string for check_direntries"
    default "direntries/"
    depends on KNOERRE_CHECK_DIRENTRIES
    help
      The key string for check_direntries as described in man-page knoerre(1).
*/

/*
 * do_check_direntries.c
 * $Id: do_check_direntries.c 691 2017-06-04 23:12:04Z fwbergm $
 */

/**
 * Step recursively into a directory and count direntries.
 *
 * This is a helper function of do_check_direntries.
 */
static int walkdir(
    const char *name,
    uint32_t *entries,
    uint32_t *entries_maxdir,
    char *pathname_maxdir,
    uint16_t pathname_size,
    int32_t depth,
    int32_t curdepth,
    const char *subdir,
    char *errfile,
    dev_t *topmostdev) {

  register DIR *dptr;
  register struct dirent *eptr;
  struct stat statbuf;
  register uint64_t entries_sum;
  register char *nptr;
  register char *p;
  register int dotdirs;
  register int32_t hlinks;
  register int ret;
  int r2;
  char pathname[PATHNAME_SIZE*2];

  dptr = opendir(name);
  if (dptr == NULL) return -1;
  ret = 0;
  dotdirs = 0;
  hlinks = 2;
  if (lstat(name, &statbuf) == 0) {
    /* simple cross-fs check */
    if (!*topmostdev) {
      *topmostdev = statbuf.st_dev;
    } else {
      if (*topmostdev != statbuf.st_dev) return 0;
    }
    hlinks = statbuf.st_nlink;
    if (hlinks >= 2) hlinks -= 2; /* leaf optimization */
  }
  for(entries_sum = 1;;) {
    eptr = readdir(dptr); if (eptr == NULL) break;
    nptr = eptr->d_name;
    if ((dotdirs<2)&&(*nptr == '.')) {
      if (nptr[1]=='\0') { dotdirs+=1;continue; }
      if (nptr[1]=='.') {
        if (nptr[2]=='\0') { dotdirs+=1;continue; }
      }
    }
    if (hlinks>0) { /* leaf optimization */
      p = stpcpy(pathname, name);
      *p++ = '/';
      p = stpcpy(p, nptr);
      if (subdir!=NULL) {
        if(*subdir!=0) {
          *p++ = '/';
          p = stpcpy(p, subdir);
        }
      }
      if (lstat(pathname,&statbuf)==0){
        if ((statbuf.st_mode&S_IFMT)==S_IFDIR) {
          hlinks--;
          if (depth>0) {
            if (curdepth<depth) {
              r2 = walkdir(pathname,entries,entries_maxdir,pathname_maxdir,pathname_size,depth,curdepth+1,NULL,errfile,topmostdev);
              if (r2 != -1) ret = r2;
            }
          } else {
            r2 = walkdir(pathname,entries,entries_maxdir,pathname_maxdir,pathname_size,depth,0,NULL,errfile,topmostdev);
            if (r2 != -1) ret = r2;
          } /* depth */
        } else {
          entries_sum++;
        } /* statbuf */
      } else {
        ret = errno;
        stpcpy(errfile, pathname);
      } /* lstat */
    } else {
      entries_sum++;
    } /* hlinks */
  }
  closedir(dptr);
  if ((entries_sum>*entries_maxdir)&&(subdir == NULL)) {
    *entries_maxdir = entries_sum;
    strcpy(pathname_maxdir,name);
  }
  *entries+=entries_sum;
  return ret;
}


/**
 * This function counts all direntries (not inodes) recursively in a dir and prints the count.
 *
 * This function is called by knoerre main().
 * It walks recursively through a dir and counts all entries in the dir and its subdirs and
 * writes it as http data out and exists.
 * $Id: do_check_direntries.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * @param buf The http data buf for returning data.
 * @param buflen The size of the http data buf.
 * @param arg1 The directory to analyze.
 * @param depth The recursion level for stepping into the directory.
 */
void do_check_direntries(register char *buf, const int buflen, char *bufp, char *arg1, uint16_t depth) {
  register char *p;
  register char *subdir;
  register int ret;
  uint32_t entries;
  register char *basedir;
  uint32_t entries_maxdir;
  char pathname_maxdir[PATHNAME_SIZE+1];
  char errfile[(2*PATHNAME_SIZE)+1];
  dev_t topmostdev;

  basedir = arg1;
  p = strchr(basedir, ' ');
  *p = 0;
  entries = 0; entries_maxdir=0;
  *errfile = 0;
  topmostdev = 0;

  /* wildcards? */
  subdir = strstr(basedir,"/./");
  if (!subdir) subdir = strstr(basedir,"/@/");
  /* strip trailing slash */
  ret = strlen(basedir);
  if(ret > 1) {
    p = basedir + ret - 1;
    if(*p == '/') *p = 0;
  }
  if (subdir) {
    *subdir = 0;
    subdir += 3;
    ret = walkdir(basedir,&entries,&entries_maxdir,pathname_maxdir,PATHNAME_SIZE,depth,0,subdir,errfile,&topmostdev);
    /* entries = entries_maxdir; */
  } else {
    ret = walkdir(basedir,&entries,&entries_maxdir,pathname_maxdir,PATHNAME_SIZE,depth,0,NULL,errfile,&topmostdev);
    if (ret!=0) goto error;
  }
  p = stpcpy(bufp, pathname_maxdir);
  *p++ = '=';
  p = sputlongs(p, entries_maxdir) - 1;
  *p++ = ' ';
  p = stpcpy(p, "[total = ");
  p = sputlongs(p, entries) - 1;
  p = stpcpy(p, "]\n");
  p = sputlongs(p, entries_maxdir);
  write1(buf, p-buf);
  _exit(ret);

  error:
  check_error_exit(ret);
#ifdef NULLDEF
  check_error_exit_ns(buf, buflen, 1, (uint32_t)ret, strerror(ret), errfile);
#endif
}

#endif
