#include "knoerre.h"
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS

#define PIPEBUF_SIZE 8192

static pid_t pid_child;

/**
 * This function calls an external command and writes its output as http data.
 *
 * This function is called by knoerre main().
 * It calls an external command as specified in RC_FILE (/etc/knoerrerc)
 * and writes its piped output as httpd data and exits.
 * It does not use the dirty popen() function but uses a small
 * and special "popen" implementation without the usage of
 * /bin/sh and any environment.
 * The main() function of knoerre has an internal list of PATHs where
 * to search for the command. The first match is used and the external
 * command is given to this function with full path. As an example
 * take an entry like this from RC_FILE:
 *     CMD loadavg cat /proc/loadavg
 * will lead to
 *     /bin/cat /proc/loadavg
 * if URL /loadavg is requested.
 * Because no shell is used on execve() (native call) you can't use any
 * shell features like wildcards and also environment variables but on the
 * other hand you get more security.
 * The number of arguments to the external command is limited due to
 * simplicity and security to 15 arguments (not including argv[0]).
 * $Id: do_check_external_command.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * @param buf The http data buf for returning data.
 * @param buflen The size of the http data buf.
 * @param arg1 pointer to the first arg in request_string
 * @param path The command to be called with full path and args
 */
void do_check_external_command(register char *buf, const int buflen, char *arg1, char *path) {
  register char *p;
  register int i;
  int fd[2];
  int status_child;
  char pipebuf[PIPEBUF_SIZE+1];
  register char *u;

  if (pipe(fd)<0) goto sys_error;
  if ((pid_child=fork())<0) {
    close(fd[0]); close(fd[1]);
    goto sys_error;
  }
  if (!pid_child) {
    const char *argv[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    close(fd[0]); close(0);
    dup2(fd[1],1); close(fd[1]); 
    p=path;
    i=0;
    argv[i]=p;
    /* split knoerrerc cmdstring into cmd and args */
    for(;(*p!='\0')&&(i<16);p++) {
      if (*p==' ') {
        *p++='\0';
        while(*p==' ') p++;
        if ((*p!='\0')&&(i<16)) argv[++i]=p;
      }
    }
    p = arg1;
    if ((*p) && (i<16)) {
      while(*p=='/') p++;
      argv[++i] = p;
      /* key has args, split arg1 into args and append after cmdstring args */
      for(;(*p!='\0')&&(i<16);p++) {
        if (*p=='/') {
          *p++='\0';
          while(*p=='/') p++;
          if ((*p!='\0')&&(i<16)) argv[++i]=p;
        }
      }
    }
    execve(argv[0],(char*const*)argv,NULL);
    close(1);
    /*goto sys_error;*/
    _exit(127);
  }
  close(fd[1]);
  p = pipebuf;
  i = read(fd[0],p,PIPEBUF_SIZE);
  if (i>0) { 
    u = p + PIPEBUF_SIZE;
    while (i > 0) {
      p += i;
      i = read(fd[0],p,u-p);
    }
    *p = '\0';
    
    p = buf + strlen(buf);
    p = stpcpy(p, pipebuf);
    write1(buf,p-buf);
  } else {
    p = stpcpy(buf, "no data from extcmd\n");
    p = stpcpy(p, ERROR_COMMON_VALUE_STRING);
    write1(buf, p-buf);
    /*_exit(0);*/
  }
  close(fd[0]);
  waitpid(pid_child, &status_child, 0);
  if (!WIFEXITED(status_child)) {
    kill(pid_child,SIGKILL);
  }
  _exit(0);

  sys_error:
  p = stpcpy(buf, ERROR_SYSCALL_VALUE_STRING);
  write1(buf, p-buf);
  _exit(0);
}

#endif
