#ifdef KNOERRED
#include "knoerred.h"
#else
#include "knoerre.h"
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOADUSER
#ifndef CONFIG_KNOERRE_PROC_ACCESS
#error PROC_ACCESS needed for this check
#endif

#define MAX_USERS_IN_PASSWD 10000

/*
config KNOERRE_CHECK_LOADUSER
    bool "check_loaduser"
    default y
    depends on KNOERRE_PROC_ACCESS
    help
      This knoerre module shows the max count of processes per user.
      It needs PROC_ACCESS. More information offers man-page knoerre(1).

config KNOERRE_CHECK_LOADUSER_KEY
    string "key string for check_loaduser"
    default "loaduser/"
    depends on KNOERRE_CHECK_LOADUSER
    help
      The key string for check_loaduser as described in man-page knoerre(1).
*/

/**
 * This function counts the maximum number of processes per account.
 *
 * Every file /proc/PID/status is read for the Uid (no stat call!)
 * and the processcounter for this uid is incremented if it matches
 * the given uid range by "/loaduser/UIDMIN/UIDMAX".
 * For the three top uids in this list we get the user name by
 * getpwuid_r() and print the users and the max count in the last
 * line.
 *
 * $Id: do_check_loaduser.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * @param buf The http data buf for returning data.
 * @param buflen The size of the http data buf.
 * @param bufp The pointer to char after http-header in buf.
 * @param arg1 The file/directory which lines to count
 */
void do_check_loaduser(char *buf, const int buflen, char *bufp, char *arg1) {
  register char *p;
  int16_t c;
  int rf;
  int64_t tval;
  uint64_t uid_min;
  uint64_t uid_max;
  DIR *dptr;
  struct dirent *eptr;
  struct passwd spwd;
  struct passwd *pspwd;
  char spwdbuf[STRING_SIZE+1];
  uint32_t uidp[MAX_USERS_IN_PASSWD+1];
  uint32_t uidcnt[MAX_USERS_IN_PASSWD];
  char name[2*PATHNAME_SIZE];
  char filebuf[BUF_SIZE+1];
  uint32_t max1uid, max2uid, max3uid, max1cnt, max2cnt, max3cnt;

  p = arg1;
  p[strlen(p)-1]='\0';

  if (!str2uint64(p, &uid_min)) goto err_exit;
  while((*p >= '0') && (*p <= '9')) p++;
  if (*p != '/') goto err_exit;

  p++;
  if (!str2uint64(p, &uid_max)) goto err_exit;
  while((*p >= '0') && (*p <= '9')) p++;

  if (uid_max < uid_min) goto err_exit;

  uidp[0] = uidcnt[0] = 0;

  /* scan all /proc/NUMBER/status */
  dptr=opendir("/proc");
  if(dptr==NULL) goto err_exit;

  for(;(eptr=readdir(dptr))!=NULL;) {

    if ((*(eptr->d_name) < '0') || (*(eptr->d_name) > '9')) continue;

    /* don't use stat/lstat(), read file .../status, works on 32- and 64-Bit os */
    p = stpcpy(name, "/proc/");
    p = stpcpy(p, eptr->d_name);
    p = stpcpy(p, "/status");
    p = name;

    rf = open(p, O_RDONLY);
    if (rf < 0) continue;
    p = filebuf;
    if (read(rf, p, BUF_SIZE) <2 ) { close(rf); continue; }
    p[BUF_SIZE] = '\0';
    close(rf);
    if ((c = str_index2(p, proc_status_uid_string)) < 0) continue;
    p += c; /* points to real uid */
    /* c = field_index(p, 1); */
    /* c = str_field(p, 1, '\t', 0); */
    p = str_field(p, 1, '\t', 0);
    if (!str2uint64(p, (uint64_t *) &tval)) goto err_exit; /* effective user id */
    if ((tval < uid_min) || (tval > uid_max)) continue;

    /* search for uid in our list and increment process counter for this uid */
    c = 0;
    for(;;) {
      if (c >= MAX_USERS_IN_PASSWD) break;
      if (uidp[c] == 0) {
        uidp[c] = tval; uidcnt[c] = 1;
        uidp[++c] = 0; break;
      }
      if (uidp[c] == tval) {
        uidcnt[c] += 1;
        break;
      }
      c++;
    }

  }
  closedir(dptr);

  /* walk through our list and get the 3 top users with most processes running */
  max1uid = max1cnt = max2uid = max2cnt = max3uid = max3cnt = 0;
  c = 0;
  for(;;) {
    if (c >= MAX_USERS_IN_PASSWD) break;
    if (uidp[c] == 0) { break; }
    if (uidcnt[c] > max1cnt) {
      max3uid = max2uid; max3cnt = max2cnt;
      max2uid = max1uid; max2cnt = max1cnt;
      max1uid = uidp[c]; max1cnt = uidcnt[c];
    } else {
      if (uidcnt[c] > max2cnt) {
        max3uid = max2uid; max3cnt = max2cnt;
        max2uid = uidp[c]; max2cnt = uidcnt[c];
      } else {
        if (uidcnt[c] > max3cnt) { max3uid = uidp[c]; max3cnt = uidcnt[c]; }
      }
    }
    c++;
  }

  getpwuid_r(max1uid, &spwd, spwdbuf, STRING_SIZE, &pspwd);
  if (pspwd) {
    p = stpcpy(bufp, spwd.pw_name);
  } else {
    p = sputlongs(bufp, max1uid); --p;
  }
  *p++ = '=';
  p = sputlongs(p, max1cnt);
  if (max2cnt != 0) {
    --p;
    *p++ = ' ';
    getpwuid_r(max2uid, &spwd, spwdbuf, STRING_SIZE, &pspwd);
    if (pspwd) {
      p = stpcpy(p, spwd.pw_name);
    } else {
      p = sputlongs(p, max2uid); --p;
    }
    *p++ = '=';
    p = sputlongs(p, max2cnt);
    if (max3cnt != 0) {
      --p;
      *p++ = ' ';
      getpwuid_r(max3uid, &spwd, spwdbuf, STRING_SIZE, &pspwd);
      if (pspwd) {
        p = stpcpy(p, spwd.pw_name);
      } else {
        p = sputlongs(p, max3uid); --p;
      }
      *p++ = '=';
      p = sputlongs(p, max3cnt);
    }
  }
  p = sputlongs(p, max1cnt);
  write1(buf, p - buf);
  _exit(0);

  err_exit:
  check_error_exit(0);
}
#endif
