#ifdef KNOERRED
#include "knoerred.h"
#else
#include "knoerre.h"
#endif
#ifdef CONFIG_KNOERRE_CHECK_NFS

/*
config KNOERRE_CHECK_NFS
    bool "check_nfs"
    default y
    help
      This knoerre module checks read access to a NFS-mounted filesystem.
      More information offers man-page knoerre(1).

config KNOERRE_CHECK_NFS_KEY
    string "key string for check_nfs"
    default "nfs/"
    depends on KNOERRE_CHECK_NFS
    help
      The key string for check_nfs as described in man-page knoerre(1).
*/

/*
 * $Id: do_check_nfs.c 691 2017-06-04 23:12:04Z fwbergm $
 */

static pid_t do_check_nfs_pid;

/**
 * This signal handler prints the content of check_nfs_filename.
 *
 * This is a helper function of do_check_nfs().
 * It will be executed if our forked child was able to read the contents
 * of the file and terminated itself well.
 * Now it should be safe enough to read the contents again and send it
 * back to the monitoring server.
 *
 * @param num The signal caught.
 */
static void check_nfs_sig_handler(int num) {
  register char *p;
  ssize_t bytes;
  int fd;
  int child_status;

  /* ignore all signals, check_nfs_sig_handler should only be called once */
  signal(SIGCHLD, SIG_IGN);

  /* match do_check_nfs_pid? */
  if (waitpid(do_check_nfs_pid, &child_status, 0) != do_check_nfs_pid) goto leave_handler;
  p = knoerre_global_buf_cursor;
  /* child killed? */
  if (WIFSIGNALED(child_status)) goto errexit;
  /* child exited normally? */
  if (!WIFEXITED(child_status)) goto leave_handler;
  /* return code is 0? */
  if (WEXITSTATUS(child_status) != 0) goto leave_handler;

  /*
   * at this point we are sure, that our child did a good job, terminated normally
   * and had a return value of 0, it should be safe to read the file contents
   */
  fd = open(knoerre_global_arg1, O_RDONLY);
  if (fd == -1) goto errexit;
  bytes = read(fd, p, PATHNAME_SIZE);
  close(fd);
  if (bytes <= 0) goto errexit;
  p += bytes;
  *p = 0;

  write1(knoerre_global_buf, p - knoerre_global_buf);
  _exit(0);

  errexit:
  p = stpcpy(p, ERROR_COMMON_VALUE_STRING);
  write1(knoerre_global_buf, p - knoerre_global_buf);
  _exit(0);

  leave_handler:
  return;
}

/**
 * This function tries to read the content of a NFS-mounted file and writes it as http data out.
 *
 * This function is called by knoerre main().
 * It reads the content of a (NFS-mounted) file and writes it to stdout.
 * To avoid blocking if the NFS-server vanished it does a fork() and open/read in the child process.
 * If the child blocks it will be terminated by the father after a sleep.
 * If the child can open/read the file it will exit and the check_nfs_sig_handler will do the actual open/read.
 */
void do_check_nfs() {
  register char *p;
  register ssize_t bytes;
  register int i;

  p = strchr(knoerre_global_arg1, ' ');
  *p = 0;

  p = stpcpy(knoerre_global_buf_cursor, knoerre_global_arg1);
  *p = '\n';
  ++p;
  knoerre_global_buf_cursor = p;

  /* catch a SIGCHILD of a well teminated child */
  if (signal(SIGCHLD, check_nfs_sig_handler) == SIG_ERR) goto error;

  if ((do_check_nfs_pid = fork()) < 0) {
    goto error;
  } else if (do_check_nfs_pid == 0) {
    /* child: try to open nfs-file */
    char dummybuf[STRING_SIZE+1];
    i = open(knoerre_global_arg1, O_RDONLY|O_NONBLOCK);
    if (i == -1) _exit(1);
    bytes = read(i, dummybuf, sizeof dummybuf - 1);
    if (bytes <= 0) _exit(2);
    close(i);
    _exit(0);
  }

  /* parent: sleep a short time while child maybe does his job */
  sleep(5);
  /* this kill will not be executed if check_nfs_sig_handler() will exit our process */
  if (kill(do_check_nfs_pid, 9) == 0) sleep(5);
  /*
   * if kill was successful (child did still exist and did not exit) then 
   * sleep a short time while sighandler maybe does his job
   */

  error:
  p = strcpy(p, ERROR_COMMON_VALUE_STRING);
  write1(knoerre_global_buf, p - knoerre_global_buf);
  _exit(1);
}

#endif
