#ifdef PLUGIN
#ifndef KNOERRED
#define KNOERRED
#endif
#endif
#ifdef KNOERRED
#include "knoerred.h"
#else
#include "knoerre.h"
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCMOUNTS
#ifndef CONFIG_KNOERRE_PROC_ACCESS
#error PROC_ACCESS needed for this check
#endif
#include <sys/mman.h>
#include <errno.h>

/*
config KNOERRE_CHECK_PROCMOUNTS
    bool "check_procmounts"
    default y
    depends on KNOERRE_PROC_ACCESS
    help
      This knoerre module checks all mounts listed in /proc/mounts for being alive.
      More information offers man-page knoerre(1).

config KNOERRE_CHECK_PROCMOUNTS_KEY
    string "key string for check_procmounts"
    default "procmounts "
    depends on KNOERRE_CHECK_PROCMOUNTS
    help
      The key string for check_procmounts as described in man-page knoerre(1).
*/


#ifdef SMALL_BUILD
/* MOUNTS_BUF_SIZE for Atmel, MIPS and other embedded systems */
#define MOUNTS_BUF_SIZE 65536
#else
#define MOUNTS_BUF_SIZE 262144
#endif

static pid_t do_check_procmounts_pid;
static char *checkmount;

typedef struct struct_child_buf {
  int32_t count;
  char *current_mount;
  char *output_cursor;
  char *notfound_cursor;
  char output_buf[MOUNTS_BUF_SIZE+4];
  char notfound_buf[MOUNTS_BUF_SIZE+1];
} CHILDBUF;

CHILDBUF *sharedbuf;

/**
 * This signal handler will reset checkmount if child terminated successfully.
 *
 * This is a helper function of do_check_procmounts().
 * It will be executed if our forked child was able to do a statfs()
 * and terminated itself well.
 *
 * @param num The signal caught.
 */
static void check_procmounts_sig_handler(int num) {
  register char *p;
  int child_status;
  int signaled_pid;

  /* ignore all signals, check_procmounts_sig_handler should only be called once */
  signal(SIGCHLD, SIG_IGN);

  /* match do_check_procmounts_pid? */
  signaled_pid = waitpid(do_check_procmounts_pid, &child_status, 0);
  if ((signaled_pid == -1) && (errno == ECHILD)) goto child_finished_normally;
  if (do_check_procmounts_pid && (signaled_pid != do_check_procmounts_pid)) goto leave_handler;
  p = knoerre_global_buf_cursor;
  /* child killed? */
  if (WIFSIGNALED(child_status)) goto errexit;
  /* child exited normally? */
  if (!WIFEXITED(child_status)) goto leave_handler;
  if (WEXITSTATUS(child_status) != 0) goto leave_handler;

  /*
   * at this point we are sure, that our child did a good job, terminated normally
   * and had a return value of 0. Do output and exit.
   */
  child_finished_normally:
  p = stpcpy(knoerre_global_buf_cursor, sharedbuf->output_buf);
  /* last line */
  p = sputuint32(p, 0);
  *p++ = '\n';
#ifdef KNOERRED
#ifdef PLUGIN
  ++p;
  *((uint32_t *)p) = (uint32_t)(0);
  --p;
#endif
  knoerre_global_buf_cursor = p;
  /* return p - knoerre_global_buf; */
  return;
#else
  write1(knoerre_global_buf, p - knoerre_global_buf);
  _exit(0);
#endif

  leave_handler:
  return;

  errexit:
  p = stpcpy(knoerre_global_buf_cursor, sharedbuf->output_buf);
  p = stpcpy(p, "\nmounted, but lost: ");
  p = stpcpy(p, sharedbuf->current_mount);
  *p++ = '\n';
  /* last line */
  p = sputuint32(p, 2);
  *p++ = '\n';
#ifdef KNOERRED
#ifdef PLUGIN
  ++p;
  *((uint32_t *)p) = (uint32_t)(2);
  --p;
#endif
  knoerre_global_buf_cursor = p;
  /* return p - knoerre_global_buf; */
  return;
#else
  write1(knoerre_global_buf, p - knoerre_global_buf);
  _exit(0);
#endif
}

/**
 * This function checks all mounts of /proc/mounts for being alive.
 *
 * This function is called by knoerre/knoerred main().
 * If we are called by knoerred then output is limited to two lines.
 * $Id: do_check_procmounts.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * @return size of data to be sent, 0 on error
 */
int do_check_procmounts() {
  register char *p;
  register char *r;
  char *t;
  struct statfs stat_buf;
  char mountlist[MOUNTS_BUF_SIZE+1];
  int32_t s;
  int i;

  sharedbuf = mmap(0, sizeof(CHILDBUF), PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_SHARED, -1, 0);
  if (sharedbuf == (void *) -1) goto error;
  sharedbuf->count = 0;
  sharedbuf->current_mount = 0;
  sharedbuf->output_cursor = 0;
  sharedbuf->notfound_cursor = 0;
  sharedbuf->output_buf[0] = 0;
  sharedbuf->notfound_buf[0] = 0;

  /* read in list of mounts */
  i = open("/proc/mounts", O_RDONLY);
  if (__unlikely(i < 0)) goto error;
  p = mountlist;
  r = p + MOUNTS_BUF_SIZE - 1;
  s = (int32_t) read(i, p, MOUNTS_BUF_SIZE);
  if (__unlikely(s < 2)) { close(i); goto error; }
  for (;s > 0;) {
    p += s; *p = 0;
    if (p >= r) break; /* ooops: buffer too small? */
    s = (int32_t) read(i, p, r - p);
  }
  close(i);
  if (p >= r) {
    p = stpcpy(knoerre_global_buf_cursor, ERROR_SMALLBUFFER_STRING);
#ifdef KNOERRED
    return p - knoerre_global_buf;
#else
    write1(knoerre_global_buf, p - knoerre_global_buf);
    _exit(2);
#endif
  }
  if (s > 0) p[s] = '\0';
  p = mountlist;

  /* re-format list of mounts: keep only mount points */
  r = p;
  for (;;) {
    /* skip til next white space */
    for (;;++p) if (*p == ' ') break;
    if (!*p) break;
    ++p;
    /* copy til next white space */
    for (;*p != ' ';++p) *r++ = *p;
    *r++ = 0;
    /* skip til LF */
    for (;;++p) if (*p == '\n') break;
    ++p;
    if (!*p) break;
  }
  *r = 0;
  t = r;

  /* catch a SIGCHILD of a well teminated child */
  if (signal(SIGCHLD, check_procmounts_sig_handler) == SIG_ERR) goto error;

  /* do statfs() on all fstab entries to see if mount does still work */
  /* like do_check_nfs() do the statfs() in a child to avoid blocking */
  checkmount = p;

  if ((do_check_procmounts_pid = fork()) < 0) {

    goto error;

  } else if (do_check_procmounts_pid == 0) {

    /* child: statfs on all /proc/mounts entries */

    (void) signal(SIGCHLD, SIG_IGN);
    sharedbuf->current_mount = 0;
    sharedbuf->output_cursor = sharedbuf->output_buf;

    /* step through the mountlist */
    p = mountlist;
    for (;;) {
      sharedbuf->current_mount = sharedbuf->output_cursor;
      sharedbuf->output_cursor = stpcpy(sharedbuf->output_cursor, p);
      sharedbuf->output_cursor[0] = 0;
      ++sharedbuf->count;

      /* child: try to make statfs */
      (void) statfs(p, &stat_buf);

      sharedbuf->output_cursor[0] = ' '; ++sharedbuf->output_cursor;

      for (;*p;)++p;
      ++p;
      if (p >= t) break;
    }


    p = stpcpy(sharedbuf->output_cursor, "- mountcount: ");
    p = sputuint32(p, sharedbuf->count);
    *p++ = '\n';
    *p = 0;
    sharedbuf->output_cursor = p;
    _exit(0); /* child */
  }


  /* parent: sleep a short time while child maybe does his job */
  /* it will be interrupted by the signal */
  sleep(4);
  /* this kill will not be executed if sighandler did reset pid on success */
  if (kill(do_check_procmounts_pid, 9) == 0) sleep(1);
  /*
   * if kill was successful (child did still exist and did not exit) then 
   * sleep a short time while sighandler maybe does his job
   */

  error:
#ifndef KNOERRED
  check_error_exit(0);
#endif
  return 0;
}
#endif
