/**
 * $Id: knoerre.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * knoerre - small "web-server" for serving nagios remote values
 *
 * knoerre should be run through tcpserver from DJB's ucspi-tcp suite.
 * Only the brave among yourselves will have the heart to do the daring deed of using (x)inetd.
 * Compile it with: diet gcc -O2 -Wall [-D_DEBUG_] ... ; sstrip knoerre
 * License is GNU GPL v2.
 *
 * Frank Bergmann, http://www.tuxad.com
 *
 * ToDo:
 *
 * - clean up source after including all checks (i.e. str_chr/str_index)
 * - more beautify source code...
 *
 * History:
 *
 * 0.8.119: - added do_check_countsockets set of protocols, "all" protocols and all ports
 * 0.8.118: - added: check_dmesg
 * 0.8.117: - added: check_mounts, check_logcheckerr, check_procmounts, check_time
 * 0.8.116: - added knoerrehd
 * 0.8.115: - added do_check_mounts()
 * 0.8.114: - added do_check_diskusagelocal()
 * 0.8.113: - do_check_kernellog: limit read()-calls to avoid timeout on very big logfiles
 *          - do_check_kernellog: "reboot is needed" 1000 => 50, "EXT3-fs error" 300 => 500
 *          - do_check_proccountovz: bugfix output top instances
 *          - do_check_kernellog: limit scanning of very big logfiles to some MB
 * 0.8.112: - if no cpu-key matches try also knoerrerc
 * 0.8.111: - added do_check_tcp()
 * 0.8.110: - added do_check_uptime()
 * 0.8.109: - do_check_diskusage(): better percentage calculation, show number of blocks
 * 0.8.108: - do_check_kernellog() add " data lost"
 *          - do_postfix_mailqsize() maximum mode
 *          - check_remote_by_http_time set tcp timeouts and minor updates
 *          - other plugins also minor updates
 *          - small bugfix on do_check_proccountovz()
 *          - do_check_proccounttg() to count only "processes" (pid==tgid) and no "threads" (share resources with clone()-caller)
 * 0.8.107: - switched to new API: do_check_diskusage()
 *          - do_check_diskusage() also prints all and used space in GB
 * 0.8.106: - switched to new API:
 *          - do_check_cat(), do_check_nettraf() and check_error_exit()
 * 0.8.105: - reuse do_check_nfs() from svn release 10,
 *            killing a forked and blocked child is more reliable than alarm()
 * 0.8.104: - include static const version strings
 * 0.8.103: - common_timeout() before first (http-) read()
 * 0.8.102: - do_check_countsockets() shows up to three top remote addresses
 * 0.8.101: - kernellog add some checks for hosts without RAID
 * 0.8.100: - do_check_countsockets() as option count also TIME_WAIT
 *          - do_check_countsockets() does now support tcp, udp, tcp6 and udp6
 * 0.8.99:  - do_check_dirlevels() and do_check_direntries(): no cross-fs counts
 * 0.8.98:  - do_check_nettraf() bugfix parse error /proc/net/dev on low traffic
 * 0.8.97:  - do_check_netlinksdown()
 * 0.8.96:  - do_check_fileexists()
 * 0.8.95:  - kernellog check also "Neighbour table overflow"
 *          - do_check_ctxtswitch()
 * 0.8.94:  - do_postfix_mailqsize(): if dirs given show scanned dirs
 * 0.8.93:  - do_postfix_mailqsize(): specify single subdirs by key
 * 0.8.92:  - do_check_kernellog(): new kernel string "reboot is needed", weights edited
 * 0.8.91:  - do_check_nettraf() bugfix: detect carry over on 32 Bit kernels and correct it
 * 0.8.90:  - some minor changes
 * 0.8.89:  - do_check_nettraf()
 * 0.8.88:  - do_check_countsockets: support one excluded address
 * 0.8.87:  - do_check_countsockets: support also "remote"
 * 0.8.86:  - filesizesbypattern, filesizesbypatternmaxage
 * 0.8.85:  - do_check_proccountovz()
 *          - LONGPROCS disabled
 * 0.8.84:  - alarm/timeout honours CLI mode
 * 0.8.83:  - maxfilesizes with sum_mode
 * 0.8.82:  - some more *KEY defines
 * 0.8.81:  - KNOERRE_COMMON_CLI_TIMEOUT
 * 0.8.80:  - kernellog: skip page alloc failures if NO_PAGE_ALLOC_ERROR was defined
 *          - do_check_direntries/do_check_dirlevels/do_check_maxsizes: no "//" in filepath outputs
 * 0.8.79:  - for OpenVZ usage: divide load-values by cpu-cores
 * 0.8.78:  - do_check_cachedvalue()
 * 0.8.77:  - do_check_filesizes(): enable '/@/' at end of path, correct depth usage
 * 0.8.76:  - do_check_kernellog(): recognize "Unit hang" on eth-devices
 * 0.8.75:  - main(): no HTTP header in CLI mode with external commands
 * 0.8.74:  - do_check_loaduser(): fixed bug in getting euid
 * 0.8.73:  - do_check_longprocp(): fixed bug in getting euid
 * 0.8.72:  - for nagios-cache: do_check_cmdline() detect cmdline mode and decrement result by 1
 * 0.8.71:  - do_check_process() can match OpenVZ CTIDs
 * 0.8.70:  - more cleanups and enhancements
 *          - more modularizing: more keys defined in knoerre/config.h 
 *          - make code GNUish: compile with gcc and without dietlibc
 * 0.8.69:  - some cleanups and enhancements
 * 0.8.68:  - do_check_wc-l(): use open64()
 *          - do_check_countsockets(): count only tcp socket not in state TIME_WAIT
 * 0.8.67:  - do_check_process(): as an option count only 'real' daemons
 * 0.8.66:  - do_check_countsockets()
 * 0.8.65:  - do_check_mountopts()
 * 0.8.64:  - bug: remove "leaf optimization" from do_check_maxsizes()
 * 0.8.63:  - raise_stacklimit() on do_check_dirlevels()
 * 0.8.62:  - do_check_dirlevels()
 * 0.8.61:  - do_check_maxsizes()
 * 0.8.60:  - small bugfix: check_error_exit() no HTTP header output in CLI
 * 0.8.59:  - do_check_swap()
 * 0.8.58:  - small "bug": all modules usind uids did take real and not effective uid. Fixed.
 * 0.8.57:  - do_check_direntries(): return error code to help debugging
 *          - check_error_exit_ns(): with strerror and more debug out
 *          - undef 'longprocs' by default, is deprecated, use 'longprocp'
 * 0.8.56:  - small changes in do_check_loaduser()
 * 0.8.55:  - new _KEY defines in knoerre.c and knoerre/config.h
 *          - disable rsbackup and mailbackup per undef in knoerre/config.h
 * 0.8.54:  - do_check_filetimestamp(): print also content of file if possible
 * 0.8.53:  - do_check_cpu(): optional parameter seconds for measuring interval
 * 0.8.52:  - do_check_mysqllog(): fixed datestr bug
 * 0.8.51:  - do_check_direntries(): "bugfix": don't show total count but only entries in maxdir
 * 0.8.50:  - do_check_loaduser() added: count processes per user
 * 0.8.49:  - do_check_wcl(): also print last bytes of file in last-before line, requested by jraschke
 * 0.8.48:  - do_check_cpu()
 * 0.8.47:  - do_check_kernellog(): no lightweightbadline as lastline if highweightbadlines exist
 * 0.8.46:  - new do_check_mysqllog(); do_check_kernellog(): some defines moved to knoerre.h
 * 0.8.45:  - do_check_kernellog(): increase MATCH_LINES_CACHE; just few low-weight lines in output
 * 0.8.44:  - do_check_kernellog(): bad log strings with weight
 * 0.8.43:  - do_check_kernellog(): look also for " segfault at "
 * 0.8.42:  - do_check_cat(): use ERROR_COMMON_VALUE_STRING on error
 * 0.8.41:  - more security fixes/enhancements, configurable access to /etc /proc /sys
 *            default is undef'd ENABLE_*_ACCESS
 * 0.8.40:  - security fix/enhancement: no access to "/etc", "../", it's not necessary
 * 0.8.39:  - do_check_cmp()
 * 0.8.38:  - do_check_external_command() now passes args to extcmd given by request_string
 * 0.8.37:  - do_check_cat() prints also mtime string in last before line
 * 0.8.36:  - eliminate http header with CLI on some checks
 * 0.8.35:  - do_check_cat(), do_check_nfs() is deprecated
 * 0.8.34:  - do_check_load15()
 * 0.8.33:  - str_len() => strlen()
 *          - jraschke: do_check_mailbackup()
 * 0.8.32:  - do_check_kernellog(): don't search for "ommand failed"
 * 0.8.31:  - less rounding on diskusage
 *          - statfs64 replaces statfs
 * 0.8.30:  - do_check_swaps() detect empty /proc/swaps
 * 0.8.29:  - replace lstat() by lstat64()
 *          - do_check_cmdline()
 * 0.8.28:  - do_check_proccount()
 *          - do_check_process() small changes
 *          - do_check_swaps()
 * 0.8.27:  - do_check_filetimestamp(): selection of oldest ts possible
 * 0.8.26:  - small code changes
 *          - do_check_longprocp()
 * 0.8.25:  - do_check_timediff(): error checking
 * 0.8.24:  - do_check_timediff()
 *          - nagios plugin check_remote_by_http_time
 * 0.8.23:  - do_check_load5()
 *          - cacti_remote_by_http tool
 * 0.8.22:  - split up file do_check_longuserprocesses.c
 *          - disable logging setup in knoerre-conf
 * 0.8.21:  - substitue the last stat() calls by lstat()
 *          - do_check_filetimestamp()
 *          - more error strings in do_check_kernellog()
 * 0.8.20:  - different checks of input length for CLI and stdin mode
 *          - more ifdefs for monolithic version
 *          - no stat/lstat() call in check_longuserprocesses, use /proc/.../status
 *            compatible for 32 and 64 bit os
 * 0.8.19:  - again changes in timeout handling
 * 0.8.18:  - again changes in timeout handling
 * 0.8.17:  - do_check_external_command(): error on no data from pipe
 * 0.8.16:  - add do_check_diskinodes
 *          - small changes do_check_load1()
 *          - change calculation of diskusage again:
 *            should now also work on very big and very full fs
 * 0.8.15:  - change calculation of diskusage
 * 0.8.14:  - speed up startup and other things in knoerre.c
 *          - accept request-key as commandline argument
 * 0.8.13:  - check_diskusage: show free space in GB
 *          - remove return values for nagios on redirect
 *          - use a nearly complete "fake" redirect
 * 0.8.12:  - add do_check_wcl()
 * 0.8.11:  - some bugfixes and adjustments
 * 0.8.10:  - add do_check_kernellog()
 * 0.8.9:   - also use '@' as wildcard because webbrowsers strip './' sequences
 *          - add stpcpy()
 *          - strip some str_cat(), use stpcpy()
 *          - include do_check_logentriesdaily
 *          - some "bug"fixes
 *          - add more information in line before last
 * 0.8.8:   - add common timeout with alarm sighandler
 * 0.8.7:   - include check_filesizes
 * 0.8.6:   - include check_external_command
 *          - substitute popen(), don't use MONOLITHIC as default in Makefile
 *          - LONGPROC_UID_MIN and LONGPROC_UID_MAX in RC_FILE
 * 0.8.5:   - include check_maxdirentries
 * 0.8.4:   - include check_direntries
 * 0.8.3:   - include check_nfs
 * 0.8.2:   - include check_process
 *          - include check_mailqsize
 *          - include check_backup
 *          - include check_longuserprocesses
 *          - remove check_kugelfisch - use check_process as substitution
 * 0.8.1:   - include check_diskusage
 *          - include check_kugelfisch
 */


#include "knoerre.h"


static const char version_string[] = "version_string: $Id: knoerre.c 691 2017-06-04 23:12:04Z fwbergm $";

/**
 * This is the default type 200 http header to be returned.
 *
 * After the header a line with the remote check value will be sent.
 * This value will be added by any of the internal do_check_*() functions
 * or external commands.
 */
const char str200[]="HTTP/1.0 200 OK\r\nServer: knoerre/"VERSION"\r\nContent-Type: text/plain\r\n\r\n";



/**
 * This is the type 302 http redirect header.
 *
 * It is just a default http output on any error when there's no error value or string
 * from an internal do_check_*() function or an external command.
 * Mostly it will be sent on unknown URLs or on syscall errors.
const char str302[]="HTTP/1.0 302 Found\r\nServer: knoerre/"VERSION"\r\nLocation: http://"CONFIG_REDIRECT_HOSTNAME"/\r\nContent-Type: text/html\r\n\r\n<title>302 Found</title>\r\nThe document is <a href=\"http://"CONFIG_REDIRECT_HOSTNAME"/\">here</a>.\n";
 */
const char str302[]="HTTP/1.1 302 Found\r\nDate: Thu, 01 Jan 1970 01:00:00 CET\r\nServer: Apache/1.3.3 (Unix) (Red Hat/Linux)\r\nLocation: http://"CONFIG_REDIRECT_HOSTNAME"/\r\nAccept-Ranges: bytes\r\nContent-Length: 143\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\r\n<html><head>\n<title>302 Found</title>\n</head><body>\n<h1>Found</h1>\nThe document has moved <a href=\"http://"CONFIG_REDIRECT_HOSTNAME"/\">here</a>.\n</body></html>\n";


char *knoerre_global_buf;
const int knoerre_global_bufsize = BUF_SIZE;
char *knoerre_global_buf_cursor;
char *knoerre_global_arg1;


/**
 * knoerre internal search PATHs for external commands.
 *
 * We don't trust entries in RC_FILE and use our own default PATHs.
 * Command entries in RC_FILE containing "../" are ignored.
 */
const char *paths[]= { "/bin/", "/sbin/", "/usr/sbin/", "/usr/bin/", "/usr/local/sbin/", "/usr/local/bin/", 0 };



/**
 * knoerre main() function, all other funcs are moved to separate files
 *
 */
int main(int argc,char *const argv[]) {
  char buf[3*BUF_SIZE];
  char key_buf[STRING_SIZE+1];
  register char *cptr, *u;
  int i;
  char *request_string;
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
  int l;
  int s;
  struct statfs stat_buf;
#endif


  /* cmdline or stdin? */
  if (__unlikely(argc > 1)) {

    u = buf;
    knoerre_global_buf = buf;
    knoerre_global_buf_cursor = u;

    /* command line mode: expect request key as argument */
#ifdef CONFIG_KNOERRE_CHECK_LOAD15
    if (!strncmp(argv[1], CONFIG_KNOERRE_CHECK_LOAD15_KEY, sizeof CONFIG_KNOERRE_CHECK_LOAD15_KEY - 2)) {
      do_check_load15(u, BUF_SIZE, u);
    }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOAD1
    if (!strncmp(argv[1], CONFIG_KNOERRE_CHECK_LOAD1_KEY, sizeof CONFIG_KNOERRE_CHECK_LOAD1_KEY - 2)) {
      do_check_load1(0);
    }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOAD5
    if (!strncmp(argv[1], CONFIG_KNOERRE_CHECK_LOAD5_KEY, sizeof CONFIG_KNOERRE_CHECK_LOAD5_KEY - 2)) {
      do_check_load5(u, BUF_SIZE, u);
    }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOAD1OVZ
    if (!strncmp(argv[1], "load1ovz", 9)) {
      do_check_load1(1);
    }
#endif

    cptr = u + BUF_SIZE;
    if (strlen(argv[1]) >= STRING_SIZE) { argv[1][STRING_SIZE-2] = '\0'; }
    u = stpcpy(cptr, argv[1]);
    *u++ = ' ';
    *u = '\0';
    request_string = cptr;
    i = u - cptr;
    u = buf;
    *u = '\0';
    common_timeout(KNOERRE_COMMON_CLI_TIMEOUT, 0);

  } else {

    common_timeout(KNOERRE_COMMON_TIMEOUT, 1);

    /* "normal" http-mode, expect http-request from stdin per tcpserver or inetd */
    cptr = key_buf;
    i = http_to_key(0, key_buf, STRING_SIZE);
    if (i == 0) goto hard_exit;
    if (i < 0) goto redirect_exit;
    request_string = cptr;
    u = stpcpy(buf, str200);

  }

  /* common part for cmdline and stdin */

  knoerre_global_buf = buf;
  knoerre_global_buf_cursor = u;
  knoerre_global_arg1 = request_string;

  /* do some basic checks */
  if (__unlikely(strstr(request_string, "../"))) goto redirect_exit;
#ifndef CONFIG_KNOERRE_ETC_ACCESS
  if (__unlikely(strstr(request_string, "/etc"))) goto redirect_exit;
#endif
#ifndef CONFIG_KNOERRE_PROC_ACCESS
  if (__unlikely(strstr(request_string, "/proc/"))) goto redirect_exit;
#else
  if (__unlikely(strstr(request_string, "/proc/"))) goto checks_with_parms;
#endif
#ifndef CONFIG_SYS_ACCESS
  if (__unlikely(strstr(request_string, "/sys/"))) goto redirect_exit;
#endif

    /* load* will be pre-checked before other compares */
#ifdef CONFIG_KNOERRE_CHECK_LOAD1
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOAD1_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOAD1_KEY) - 1)) {
    knoerre_global_buf_cursor = stpcpy(buf, str200);
    do_check_load1(0);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOAD5
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOAD5_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOAD5_KEY) - 1)) {
    u = stpcpy(buf, str200);
    do_check_load5(buf, BUF_SIZE, u);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOAD15
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOAD15_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOAD15_KEY) - 1)) {
    u = stpcpy(buf, str200);
    do_check_load15(buf, BUF_SIZE, u);
  }
#endif

  /* CPU percentage for user, nice, ... */
#ifdef CONFIG_KNOERRE_CHECK_CPU
  if (!strncmp(request_string, "cpu", 3)) {
    request_string += 3;
    if (*request_string == 'u') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUUSERTOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUUSERCURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
    if (*request_string == 'n') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUNICETOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUNICECURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
    if (*request_string == 's') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUSYSTEMTOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUSYSTEMCURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
    if (*request_string == 'i') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUIDLETOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUIDLECURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
    if (*request_string == 'w') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUWAITTOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUWAITCURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
    if (*request_string == 'I') {
      request_string++;
      if (*request_string == 't') {
        do_check_cpu(buf, BUF_SIZE, u, CPUIDLEINVERSETOTAL, request_string);
      }
      if (*request_string == 'c') {
        do_check_cpu(buf, BUF_SIZE, u, CPUIDLEINVERSECURR, request_string);
      }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
      request_string -= 4;
      goto parse_knoerrerc;
#else
      goto redirect_exit;
#endif
    }
#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
    request_string -= 3;
    goto parse_knoerrerc;
#else
    goto redirect_exit;
#endif
  }
#endif


#ifdef CONFIG_KNOERRE_PROC_ACCESS

  checks_with_parms:

#endif

#ifdef CONFIG_KNOERRE_CHECK_CACHEDVALUE
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_CACHEDVALUE_KEY, sizeof(CONFIG_KNOERRE_CHECK_CACHEDVALUE_KEY) - 1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_CACHEDVALUE_KEY - 1;
    do_check_cachedvalue();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_TIMEDIFF
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY, sizeof(CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY)-1)) {
    do_check_timediff(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_TIME
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_TIME_KEY, sizeof(CONFIG_KNOERRE_CHECK_TIME_KEY)-1)) {
    do_check_time(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_TIME_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCESS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCESS_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCESS_KEY)-2)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCESS_KEY - 2;
    do_check_process();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCESS0
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCESS0_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCESS0_KEY)-1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCESS0_KEY - 3;
    do_check_process();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCESSD
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCESSD_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCESSD_KEY)-1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCESSD_KEY - 3;
    do_check_process();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCCOUNTOVZ
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCCOUNTOVZ_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCCOUNTOVZ_KEY))) {
    /* knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCCOUNTOVZ_KEY; */
    do_check_proccountovz();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCCOUNTTG
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCCOUNTTG_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCCOUNTTG_KEY))) {
    /* knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCCOUNTTG_KEY; */
    do_check_proccounttg();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCCOUNT
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCCOUNT_KEY, sizeof(CONFIG_KNOERRE_CHECK_PROCCOUNT_KEY))) {
    /* knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_PROCCOUNT_KEY; */
    do_check_proccount();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_SWAPS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_SWAPS_KEY, sizeof(CONFIG_KNOERRE_CHECK_SWAPS_KEY))) {
    do_check_swaps(buf, BUF_SIZE, u);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_SWAP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_SWAP_KEY, sizeof(CONFIG_KNOERRE_CHECK_SWAP_KEY))) {
    do_check_swap(buf, BUF_SIZE, u);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MAILQSIZE
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MAILQSIZE_KEY, sizeof(CONFIG_KNOERRE_CHECK_MAILQSIZE_KEY) - 1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_MAILQSIZE_KEY - 1;
    do_postfix_mailqsize();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_TCP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_TCP_KEY, sizeof(CONFIG_KNOERRE_CHECK_TCP_KEY) - 1)) {
    /* knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_TCP_KEY - 1; */
    do_check_tcp();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_UPTIME
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_UPTIME_KEY, sizeof(CONFIG_KNOERRE_CHECK_UPTIME_KEY) - 1)) {
    /* knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_UPTIME_KEY - 1; */
    knoerre_global_arg1 = request_string;
    do_check_uptime();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_CTXTSWITCH
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_CTXTSWITCH_KEY, sizeof(CONFIG_KNOERRE_CHECK_CTXTSWITCH_KEY) - 1)) {
        do_check_ctxtswitch(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_CTXTSWITCH_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_WC_L
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_WC_L_KEY, sizeof(CONFIG_KNOERRE_CHECK_WC_L_KEY)-1)) {
    do_check_wcl(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_WC_L_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DISKUSAGELOCAL
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DISKUSAGELOCAL_KEY, sizeof CONFIG_KNOERRE_CHECK_DISKUSAGELOCAL_KEY - 1)) {
    do_check_diskusagelocal();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MOUNTS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MOUNTS_KEY, sizeof CONFIG_KNOERRE_CHECK_MOUNTS_KEY - 1)) {
    do_check_mounts();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_PROCMOUNTS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_PROCMOUNTS_KEY, sizeof CONFIG_KNOERRE_CHECK_PROCMOUNTS_KEY - 1)) {
    do_check_procmounts();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DISKUSAGE
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DISKUSAGE_KEY, sizeof CONFIG_KNOERRE_CHECK_DISKUSAGE_KEY - 1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_DISKUSAGE_KEY - 2;
    do_check_diskusage();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DISKINODES
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DISKINODES_KEY, sizeof(CONFIG_KNOERRE_CHECK_DISKINODES_KEY)-1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_DISKINODES_KEY - 2;
    do_check_diskinodes();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_RSBACKUP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_RSBACKUP_KEY, sizeof(CONFIG_KNOERRE_CHECK_RSBACKUP_KEY))) {
    do_check_backup(buf, BUF_SIZE, u);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MAILBACKUP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MAILBACKUP_KEY, sizeof(CONFIG_KNOERRE_CHECK_MAILBACKUP_KEY)-1)) {
    do_check_mailbackup(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_MAILBACKUP_KEY)-1);
  }
#endif

  /* internal cmds with execution times maybe longer than some msec */
#ifdef CONFIG_KNOERRE_CHECK_LOGENTRIES
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOGENTRIES_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOGENTRIES_KEY)-1)) {
    do_check_logentriesdaily(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_LOGENTRIES_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_FILEEXISTS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_FILEEXISTS_KEY, sizeof(CONFIG_KNOERRE_CHECK_FILEEXISTS_KEY)-1)) {
    do_check_fileexists(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_FILEEXISTS_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_FILETIMESTAMP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_FILETIMESTAMP_KEY, sizeof(CONFIG_KNOERRE_CHECK_FILETIMESTAMP_KEY)-1)) {
    do_check_filetimestamp(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_FILETIMESTAMP_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DIRLEVELS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DIRLEVELS_KEY, sizeof(CONFIG_KNOERRE_CHECK_DIRLEVELS_KEY)-1)) {
    do_check_dirlevels(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_DIRLEVELS_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DIRENTRIES
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DIRENTRIES_KEY, sizeof(CONFIG_KNOERRE_CHECK_DIRENTRIES_KEY)-1)) {
    do_check_direntries(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_DIRENTRIES_KEY)-2, 0);
  }
  if (!strncmp(request_string, "maxdirentries/", 14)) {
    do_check_direntries(buf, BUF_SIZE, u, request_string+15,(uint16_t)(request_string[14]-48));
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MAXFILESIZESSUM
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MAXFILESIZESSUM_KEY, sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZESSUM_KEY)-1)) {
    do_check_maxsizes(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZESSUM_KEY),(uint16_t)(request_string[sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZESSUM_KEY)-1]-48), 1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MAXFILESIZES_KEY
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MAXFILESIZES_KEY, sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZES_KEY)-1)) {
    do_check_maxsizes(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZES_KEY),(uint16_t)(request_string[sizeof(CONFIG_KNOERRE_CHECK_MAXFILESIZES_KEY)-1]-48), 0);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_FILESIZES
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_FILESIZES_KEY, sizeof(CONFIG_KNOERRE_CHECK_FILESIZES_KEY)-1)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_FILESIZES_KEY)-2, -1);
  }
  if (!strncmp(request_string, "filesizesrecursive/", 19)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+18, 0);
  }
  if (!strncmp(request_string, "filesizesbysuffix/", 18)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+17, 1);
  }
  if (!strncmp(request_string, "filesizesbyprefix/", 18)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+17, 2);
  }
  if (!strncmp(request_string, "filesizesbypattern/", 19)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+18, 3);
  }
  if (!strncmp(request_string, "filesizesbypatternmaxage/", 25)) {
    do_check_filesizes(buf, BUF_SIZE, u, request_string+24, 4);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DMESG
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DMESG_KEY, sizeof(CONFIG_KNOERRE_CHECK_DMESG_KEY))) {
    do_check_dmesg();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_KERNELLOG
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_KERNELLOG_KEY, sizeof(CONFIG_KNOERRE_CHECK_KERNELLOG_KEY)-1)) {
    do_check_kernellog(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_KERNELLOG_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOGCHECKERR
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOGCHECKERR_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOGCHECKERR_KEY)-1)) {
    do_check_logcheckerr(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_LOGCHECKERR_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MYSQLERR
  /* if (!strncmp(request_string, "mysqlerr/", 9)) { */
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MYSQLERR_KEY, sizeof(CONFIG_KNOERRE_CHECK_MYSQLERR_KEY)-1)) {
    do_check_mysqllog(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_MYSQLERR_KEY)-2);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_SOCKETS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_SOCKETS_KEY, sizeof(CONFIG_KNOERRE_CHECK_SOCKETS_KEY)-1)) {
    do_check_countsockets(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_SOCKETS_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_NETLINKSDOWN
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_NETLINKSDOWN_KEY, sizeof(CONFIG_KNOERRE_CHECK_NETLINKSDOWN_KEY) - 1)) {
    do_check_netlinksdown();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_NETTRAF
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_NETTRAF_KEY, sizeof CONFIG_KNOERRE_CHECK_NETTRAF_KEY - 1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_NETTRAF_KEY - 1;
    do_check_nettraf();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LOADUSER
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LOADUSER_KEY, sizeof(CONFIG_KNOERRE_CHECK_LOADUSER_KEY)-1)) {
    do_check_loaduser(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_LOADUSER_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_LONGPROCP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_LONGPROCP_KEY, sizeof(CONFIG_KNOERRE_CHECK_LONGPROCP_KEY)-1)) {
    do_check_longprocp(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_LONGPROCP_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_CMDLINE
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_CMDLINE_KEY, sizeof(CONFIG_KNOERRE_CHECK_CMDLINE_KEY)-1)) {
    do_check_cmdline(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_CMDLINE_KEY)-1);
  }
#endif

  /* internal commands which MUST use alarm timeout */
#ifdef CONFIG_KNOERRE_CHECK_NFS
  /* do_check_nfs() does also have a separate internal timeout and does fork() */
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_NFS_KEY, sizeof CONFIG_KNOERRE_CHECK_NFS_KEY - 1 )) {
    /* "sizeof CONFIG_KNOERRE_CHECK_NFS_KEY - 2" because '/' in KEY belongs to pathname */
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_NFS_KEY - 2;
    do_check_nfs();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_CAT
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_CAT_KEY, sizeof CONFIG_KNOERRE_CHECK_CAT_KEY - 1)) {
    knoerre_global_arg1 = request_string + sizeof CONFIG_KNOERRE_CHECK_CAT_KEY - 2;
    do_check_cat();
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_CMP
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_CMP_KEY, sizeof(CONFIG_KNOERRE_CHECK_CMP_KEY)-1)) {
    do_check_cmp(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_CMP_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_MOUNTOPTS
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_MOUNTOPTS_KEY, sizeof(CONFIG_KNOERRE_CHECK_MOUNTOPTS_KEY)-1)) {
    do_check_mountopts(buf, BUF_SIZE, u, request_string+sizeof(CONFIG_KNOERRE_CHECK_MOUNTOPTS_KEY)-1);
  }
#endif
#ifdef CONFIG_KNOERRE_CHECK_DUMMY
  if (!strncmp(request_string, CONFIG_KNOERRE_CHECK_DUMMY_KEY, sizeof CONFIG_KNOERRE_CHECK_DUMMY_KEY)) {
    knoerre_global_arg1 = request_string;
    do_check_dummy();
  }
#endif

#ifdef CONFIG_KNOERRE_EXTERNAL_COMMANDS
  parse_knoerrerc:
  /* read rcfile - NOT in _MONOLITHIC_ version included! */
  l = open(CONFIG_KNOERRE_RC_FILE, O_RDONLY);
  if (l < 0) main_error_redirect_exit(12);
  /* cptr = buf+BUF_SIZE; */
  cptr = buf;
  i = (int)read(l, cptr, BUF_SIZE);
  close(l);
  if (!*cptr) main_error_redirect_exit(9);
  cptr[i] = '\0';

  /* small parsing of key for ' ' and '/' */
  u = strchr(request_string, ' ');
  if (!u) main_error_redirect_exit(13);
  *u = 0;
  s = u - request_string;
  u = strchr(request_string, '/');
  if (u) {
    s = u - request_string + 1;
  }

  /* parse lines from rcfile */
  for(;;) {


    /* does the line start with CMD? */
    if (!strncmp("CMD ", cptr, 4)) {
      cptr += 4;
      /* does the URL match the field after CMD? */
      if (!strncmp(request_string, cptr, s)) {
        /* yes, got the third field as the program to be called */
        cptr = next_token(cptr, ' '); if (!cptr) goto next_rc_line;
        u = cptr;
        while(*u && (*u != '\n')) u++;
        *u = '\0';
	u = cptr;
	while(*u && (*u != ' ')) u++;
	l = u-cptr;
        /* step through our PATHs and search the command */
        for(i = 0;paths[i] != 0;i++) {
          u = buf+(2*BUF_SIZE);
          strcpy(u, paths[i]);
          strncat(u, cptr, l);
	  /* parent dirs are forbidden */
	  if (strstr(u, "../")) goto next_rc_line;
	  if (statfs(u,&stat_buf) == 0) {
            /* file with path prefix exists, call it */
            strcpy(u, paths[i]);
            strcat(u, cptr);

            if (__unlikely(argc > 1)) {
              *buf = 0;
              do_check_external_command(buf, BUF_SIZE, request_string+s, u);
            } else {
              strcpy(buf, str200);
              do_check_external_command(buf, BUF_SIZE-sizeof(str200), request_string+s, u);
            }
          }
        }
      } /* diffn u cptr / cmd match? */
      goto next_rc_line;
    } /* diffn CMD cptr */


    next_rc_line:
    cptr = next_token(cptr, '\n');
    if (!cptr) break;
  }

  if (argc > 1) {
    u = buf;
    *u = 0;
  } else {
    u = stpcpy(buf, str200);
  }
#endif

  redirect_exit:
  main_error_redirect_exit(10); /* no entry found/unknown URL: redirect and error */

  hard_exit:
  write2(version_string, sizeof version_string - 1);
  _exit(1);

}
