#ifndef __KNOERRE_H
#define __KNOERRE_H
#ifdef FWBNULL
#ifdef CONFIG_KNOERRED
#define CONFIG_KNOERRE_DAEMONS
#endif
#ifdef CONFIG_KNOERREHD
#undef CONFIG_KNOERRE_DAEMONS
#define CONFIG_KNOERRE_DAEMONS
#endif
#endif

/*
$Id: knoerre.h 691 2017-06-04 23:12:04Z fwbergm $
 */

/**
 * knoerre.h is used as a unique header file for all knoerre source code files.
 *
 * Some common definitions are documented in javadoc/doxygen
 * style, but for documentation of the functions and "plugins"
 * you must look in their specific files.
 * For author, copyright and more information about knoerre
 * please take a look into knoerre.c.
 */

/**
 * Start of the output buffer
 *
 * This is the place where check functions may store their
 * output before write()ing it out.
 */
extern char *knoerre_global_buf;

/**
 * Size of the output buffer
 */
extern const int knoerre_global_bufsize;

/**
 * Cursor to the next free byte in output buffer
 *
 * In HTTP mode the buffer will be filled with a HTTP header
 * before a check routine is called. In this case the cursor
 * will point to the first byte after the header.
 * In CLI mode the cursor will point to the start of the buffer
 * before calling the check function.
 */
extern char *knoerre_global_buf_cursor;

/**
 * Pointer to the argument 1 after the key
 *
 * In some cases it will point to the '/' itself (in case that a
 * path will follow the key) but in other cases it will point to
 * the first char after "key/".
 */
extern char *knoerre_global_arg1;


extern const char str200[];
extern const char str302[];

/**
 * We need to handle large files
 */
#ifndef _LARGEFILE_SOURCE
#define _LARGEFILE_SOURCE
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#define _FILE_OFFSET_BITS 64
#include <features.h>

/* for measurements while developing */
#define rdtscl(low) __asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

#define __STDC_LIMIT_MACROS 1

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <stdint.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <pwd.h>

/* don't change this order! */
#ifndef _USE_DEPRECATED_STR2UINT63
#define _USE_DEPRECATED_STR2UINT63
#endif
#include "../lib/libfwb.h"
#include "../config.h"
#include "config.h"

/**
 * This is the base version string.
 *
 * The base version is the version number of a common source code.
 * If knoerre is compiled as a monolithic version then "m" is appended
 * to the base version.
 */
#define BASE_VERSION "0.8.119"

/**
 * The common knoerre timeout.
 *
 * This timeout must be specified in seconds.
 * It is a generic timeout for all builtin or external checks of knoerre
 * called in http mode.
 * Its value should be less than nagios timeout for check_remote_by_http
 * and greater than 2.
 */
#define KNOERRE_COMMON_TIMEOUT 40

/**
 * The common knoerre CLI timeout.
 *
 * This timeout must be specified in seconds.
 * It is a generic timeout for all builtin or external checks of knoerre
 * called in CLI mode.
 */
#define KNOERRE_COMMON_CLI_TIMEOUT 3598

/**
 * A common error return value for nagios.
 *
 * This is a common error return value for nagios checks. It is used
 * i.e. to check the number of running instances of a program: If no
 * process (cron, monit or other) is running, then "0" will be mapped
 * to this.
 * The default string "999999999999999999" is the biggest number
 * consisting of cipher "9" which fits in 63 bits (64 bits minus
 * 1 sign bit).
 * static const char ERROR_COMMON_VALUE_STRING[]="999999999999999999\n";
 */
#define ERROR_COMMON_VALUE_STRING "999999999999999999\n"

/**
 * A knoerre common timeout return value for nagios.
 *
 * If knoerre runs into its common timeout by receiving a sigalrm
 * it will return this string.
 * static const char ERROR_COMMON_TIMEOUT_STRING[]="999999999999999998\n";
 */
#define ERROR_COMMON_TIMEOUT_STRING "timeout\n999999999999999998\n"

#define ERROR_OPEN_STRING "open-error\n999999999999999997\n"
#define ERROR_READ_STRING "read-error\n999999999999999996\n"
#define ERROR_SEEK_STRING "seek-error\n999999999999999995\n"
#define ERROR_WRONG_PID_STRING "pid-error\n999999999999999994\n"
#define ERROR_CONNECT_STRING "connect-error\n999999999999999993\n"
#define ERROR_SMALLBUFFER_STRING "buffer too small\n999999999999999992\n"

/**
 * A syscall error return value for nagios.
 *
 * Its default is built like ERROR_COMMON_VALUE_STRING but with cipher
 * "8" and it is returned as error value if an knoerre-internal error
 * on a syscall did happen.
 * static const char ERROR_SYSCALL_VALUE_STRING[]="888888888888888888\n";
 */
#define ERROR_SYSCALL_VALUE_STRING "syscall-error\n888888888888888888\n"

/**
 * The size of the large file buffer.
 *
 * For reading potential large files of a size above approx. 100KB
 * buffers with this size are used.
 */
#ifdef SMALL_BUILD
/* LARGEBUF_SIZE for Atmel, MIPS and other embedded systems */
#define LARGEBUF_SIZE 262144
#else
#define LARGEBUF_SIZE 1048576
#endif

/**
 * The size of the generic buffer.
 *
 * A generic buffer like i.e. read buffer for reading in the RC_FILE
 * has this size. It should fit for common needs.
 */
#ifdef SMALL_BUILD
/* BUF_SIZE for Atmel, MIPS and other embedded systems */
#define BUF_SIZE 16384
#else
#define BUF_SIZE 65536
#endif

/**
 * The max length of a pathname.
 *
 * Filenames with complete path can't be longer than this size.
 */
#ifdef SMALL_BUILD
/* PATHNAME_SIZE for Atmel, MIPS and other embedded systems */
#define PATHNAME_SIZE 4096
#else
#define PATHNAME_SIZE 16384
#endif

/**
 * The max length of a generic string.
 *
 * Generic strings with this max size hold i.e. a line of
 * input or output. It's also a limit for some lines of the
 * RC_FILE like i.e. LONGPROC_EXCLUDES.
 */
#define STRING_SIZE 1024

/**
 * The max length of a complete http request.
 */
#define MAX_HTTP_REQUEST_SIZE 512

/**
 * The max amount of OpenVZ instances
 */
#define MAX_OVZ_INSTANCES 1024

/**
 * Used by do_check_kernellog() and do_check_mysqllog().
 */
#define LOGCHECK_MAX_NOMATCH_LINES 24

/**
 * Used by do_check_kernellog() and do_check_mysqllog().
 */
#define LOGCHECK_MATCH_LINES_CACHE 20

/**
 * Used by do_check_kernellog() and do_check_mysqllog().
 */
#define CONFIG_KNOERRE_LEN_OF_BADLINE_IN_NAGIOS 160

/**
 * Used by knoerred for interval measuring of CPU% and more.
 * Duration of one single measure interval.
 * 3 to 5 seconds is a good choice.
 */
#define MEASURE_INTERVAL_LENGTH 5

/**
 * Used by knoerred for interval measuring of CPU% and more
 * Number of intervals of MEASURE_INTERVAL_LENGTH for
 * calculating differences of jiffies and more.
 * 10 to 20 is a good choice.
 * Example:
 * MEASURE_INTERVAL_LENGTH is 5 and MEASURE_INTERVAL_COUNT
 * is 12 then CPU user%, CPU system% and more will be
 * calculated as average of the last minute.
 */
#define MEASURE_INTERVAL_COUNT 12

/**
 * Used by do_check_kernellog() and do_check_mysqllog().
 */
typedef struct struct_log_strings {
  const char *string;
  const uint16_t weight;
} KLE;

/**
 * Used by do_check_diskusagelocal().
 * List of local rw fs types to be recognized
 */
extern const char const *localrwfs[];

/**
 * Used by do_check_diskusagelocal() and do_check_mounts().
 */
typedef struct struct_fs_stats{
  const char *path;
  uint32_t used;
  uint32_t free;
  uint32_t all;
  uint16_t percent_used;
} FS_STAT;


static const char proc_status_uid_string[]="Uid:\011";

#define CURRSLEEPINTERVAL 10
enum { CPUUSERTOTAL, CPUNICETOTAL, CPUSYSTEMTOTAL, CPUIDLETOTAL, CPUWAITTOTAL, CPUIDLEINVERSETOTAL,
       CPUUSERCURR, CPUNICECURR, CPUSYSTEMCURR, CPUIDLECURR, CPUWAITCURR, CPUIDLEINVERSECURR };

#ifndef CONFIG_KNOERRE_EXTERNAL_COMMANDS
#define VERSION BASE_VERSION"m"
#else
#define VERSION BASE_VERSION
#define MAX_CMDS 64
#endif

/**
 * dummy check:
 * this is an example check
 */
int do_check_dummy();

void do_check_external_command(register char *buf, const int buflen, char *arg1, char *path);
void do_check_backup(register char *buf, const int buflen, char *bufp);
int do_check_cachedvalue();
int do_check_cat();
int do_check_cmp(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_cmdline(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_countsockets(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_cpu(register char *buf, const int buflen, char *bufp, int reqtype, char *arg1);
void do_check_ctxtswitch(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_direntries(register char *buf, const int buflen, char *bufp, char *arg1, uint16_t depth);
void do_check_dirlevels(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_diskinodes();
int do_check_diskusage();
int do_check_diskusagelocal();
int do_check_dmesg();
int do_check_fileexists(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_filesizes(register char *buf, const int buflen, char *bufp, char *arg1, int mode);
void do_check_kernellog(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_kugelfisch(register char *buf, const int buflen, char *bufp);
int do_check_load1(int ovz);
int do_check_load5(register char *buf, const int buflen, char *bufp);
int do_check_load15(register char *buf, const int buflen, char *bufp);
int do_check_loadmulti(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_loaduser(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_logcheckerr(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_logentriesdaily(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_longprocp(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_mailbackup(register char *buf, const int buflen, char *bufp, char *arg1);
void do_check_maxsizes(register char *buf, const int buflen, char *bufp, char *arg1, uint16_t depth, int sum_mode);
void do_check_mountopts(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_mounts();
int do_check_procmounts();
void do_check_mysqllog(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_netlinksdown();
int do_check_nettraf();
void do_check_nfs();
int do_check_proccount();
int do_check_proccounttg();
void do_check_proccountovz();
int do_check_process();
int do_check_filetimestamp(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_swaps(register char *buf, const int buflen, char *bufp);
void do_check_swap(register char *buf, const int buflen, char *bufp);
int do_check_tcp();
int do_check_uptime();
int do_check_timediff(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_time(register char *buf, const int buflen, char *bufp, char *arg1);
int do_check_wcl(register char *buf, const int buflen, char *bufp, char *arg1);
int do_postfix_mailqsize();
int walkdir_filesizes(
    const char *name,
    off_t *maxfilesize,
    char *maxfilesize_pathname,
    uint16_t pathname_size,
    int16_t depth,
    int16_t curdepth,
    const char *subdir,
    const char *pattern,
    int mode,
    time_t ts);
void check_error_exit(int ret);
void check_error_exit_ns(register char *buf, const int buflen, int ret, uint32_t errcode, char *errstring, char *errfile);
int16_t get_fs_stats(FS_STAT *fs_stat);
uint16_t get_all_local_fs_stats(FS_STAT *top1stat, FS_STAT *top2stat);
void my_err_exit(const int st);
char *next_token(register char *p, register const char sep);
void common_timeout(unsigned int ctimeout, int print_http_header);
void raise_stacklimit();
void main_error_redirect_exit(const int st);
int http_to_key(int readsock, char key[], int keysize);


#endif
