#ifndef __BUFFER_H
#define __BUFFER_H

/**
 * libfwb buffer.h
 * FWB, http://www.tuxad.com
 *
 * Lightweight and fast alternative functions to
 * substitute widely used stdio-functions like
 * i.e. 'printf()'. The source code is bigger when
 * using libfwb-buffer but the compiled code misses
 * the bloated stdio code and gives you fewer syscalls.
 * Therefore the resulting code is smaller and runs faster.
 *
 * example init and usage code for substituting
 * i.e. 'printf("%s = %u\n, a, b);' :
 * 
 *   #include "lib/buffer.h"
 *   #define READBUFSIZE 4096
 *   char buf0space[READBUFSIZE+1];
 *   buffer buf0;
 *   #define WRITEBUFSIZE 8192
 *   char buf1space[WRITEBUFSIZE+1];
 *   buffer buf1;
 *
 *   buf1.s = buf1space;
 *   buf1.l = 0;
 *   buf1.t = buf1space + WRITEBUFSIZE;
 *   buf1.c = buf1space;
 * 
 *   buf0.s = buf0space;
 *   buf0.l = READBUFSIZE;
 *   buf0.t = buf0space + buf0.l;
 *   buf0.c = buf0.t;
 *
 *   buffer_putstr(&buf1, a);
 *   buffer_putstr(&buf1, " = ");
 *   buffer_putuint32(&buf1, b);
 *   buffer_putc(&buf1, '\n');
 *   buffer_flush(&buf1);
 *
 * alternative init method:
 *   p = buf1space; buffer buf1 = { p, p, p + WRITEBUFSIZE, 1 };
 *
 * $Id: buffer.h 691 2017-06-04 23:12:04Z fwbergm $
 *
 */

#include <stdint.h>
#include <string.h>
#include <unistd.h>

typedef struct buffer {
  char *s;     /* first char in bufspace */
  char *c;     /* cursor */
  char *t;     /* first char after bufspace */
  int fd;      /* er... maybe the file descriptor? ;-) */
  uint32_t l;  /* for read buffers: size of the buffer */
} buffer;

#ifndef EOF
#define EOF (-1)
#endif

int buffer_flush(buffer *b);
int buffer_flushlines(buffer *b);
int buffer_put(buffer *b, register const char *a, register uint32_t l);
int buffer_putc(buffer *b, register const char a);
int buffer_putuint8(buffer *b, uint16_t num);
int buffer_putuint32(buffer *b, uint32_t num);
int buffer_putuint64(buffer *b, uint64_t num);
int buffer_getchar(buffer *b);
int buffer_getline(buffer *b, char *str, int str_size);

int buffer_putmem(buffer *b, register const char *a, register uint32_t l);
int buffer_putstr(buffer *b, register const char *a);

int buffer_openread(buffer *b, const char *filename);
int buffer_openwrite(buffer *b, char *filename);
int buffer_close(buffer *b);
int buffer_catfromfile(buffer *b, char *filename);
int buffer_filecmp(const char *filename1, const char *filename2);

#endif
