#include "libfwb.h"

/**
 * Convert kernel ip address notation to human readable form
 *
 * IP address string from /proc/net/(tcp6|udp6|tcp|udp) will
 * be converted to human readable IP address.
 * If the string has more than 8 chars then IPv6 conversion
 * will be used.
 *
 * IPv6 rules:
 *
 * rule1: 16 bit hex numbers in 8 colon separated blocks.
 * rule2: leading zeros in a block may be ommitted.
 * rule3: one or more blocks with zeros may be ommitted (causing
 * "::" in the address string).
 * rule4: ommitting zeros with rule3 may only be done once.
 *
 * $Id: procnetipaddr_to_humanreadable.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * @param procnets The address string from /proc/net/(tcp6|udp6|tcp|udp)
 * @param ips The buffer to which the human readable address will be written.
 * @return a pointer to the end of the human readable string
 */
char *procnetipaddr_to_humanreadable(const char *procnets, char *ips) {
  char c;
  int i,x,g,d;
  int part;
  int rule2, rule3, ipv4_mapped;
  char *p;

  if ((!procnets) || (!ips)) return 0;
  p = strchr(procnets, ':');
  if (p) {
    i = p - procnets;
  } else {
    i = strlen(procnets);
  }

  if (i > 8) {

    /* IPv6 conversion */

    x = 0;
    d = 0;
    part = 0;
    rule3 = 0;
    ipv4_mapped = 0;
    for (;;) {
      i = 6;
      g = 0;
      rule2 = 1;
      for (;;) {
        c = procnets[part * 8 + i];
        if (c > '9') c += 32;
        if (c != '0') {
          rule2 = 0;
          if (rule3 == 1) {
            rule3 = 2;
            if ((part == 2) && (i == 2)) ipv4_mapped = 1;
            ips[x] = ':'; ++x;
          }
        }
        if (ipv4_mapped && (part == 2) && (i < 4) && (c == 'f')) ++ipv4_mapped;
        if (ipv4_mapped == 6) {
          d = c - '0'; if (d > 9) d -= 39;
          d *= 16;
        } else {
          if (!rule2) {
            ips[x] = c; ++x;
          }
        }
        ++g;
        if (g == 4) {
          if (rule2) {
            if (!rule3) {
              rule3 = 1;
              ips[x] = ':'; ++x;
            }
          } else {
            ips[x] = ':'; ++x;
          }
          rule2 = 1; g = 0;
        }
        ++i;
        c = procnets[part * 8 + i];
        if (c > '9') c += 32;
        if (c != '0') {
          rule2 = 0;
          if (rule3 == 1) {
            rule3 = 2;
            ips[x] = ':'; ++x;
          }
        }
        if (ipv4_mapped && (part == 2) && (i < 4) && (c == 'f')) ++ipv4_mapped;
        if (ipv4_mapped == 6) {
          int printflag = 0;
          c -= '0'; if (c > 9) c -= 39;
          d += c;
          if (d > 99) {
            c = d / 100;
            d -= c * 100;
            c += '0';
            ips[x] = c; ++x;
            printflag = 1;
          }
          if (printflag || (d > 9)) {
            c = d / 10;
            d -= c * 10;
            c += '0';
            ips[x] = c; ++x;
          }
          ips[x] = d + '0'; ++x;
          if (i != 1) {
            ips[x] = '.'; ++x;
          }
        } else {
          if (!rule2) {
            ips[x] = c; ++x;
          }
        }
        ++g;
        if (g == 4) {
          if (rule2) {
            if (!rule3) {
              rule3 = 1;
              ips[x] = ':'; ++x;
            }
          } else {
            if ((i == 1) && (part == 3)) break;
            if (ipv4_mapped == 5) {
              ++ipv4_mapped;
              ips[x] = ':'; ++x;
            }
          }
          rule2 = 1; g = 0;
        }
        if (i == 1) break;
        i -= 3;
      }
      if (part == 3) break;
      ++part;
    }
    if (rule3 == 1) { ips[x] = ':'; ++x; }

  } else {

    /* IPv4 conversion */

    x = 0;
    i = 7;
    for (;;) {
  
      d = procnets[i] - '0';
      if (d > 9) d -= 7;
      --i;
  
      c = procnets[i] - '0';
      if (c > 9) c -= 7;
      d += c * 16;
      int printflag = 0;
      if (d > 99) {
        c = d / 100;
        d -= c * 100;
        c += '0';
        ips[x] = c; ++x;
        printflag = 1;
      }
      if (printflag || (d > 9)) {
        c = d / 10;
        d -= c * 10;
        c += '0';
        ips[x] = c; ++x;
      }
      ips[x] = d + '0'; ++x;
  
      if (!i) break;
  
      ips[x] = '.'; ++x;
      --i;
  
    }

  }

  ips[x] = 0;
  return ips + x;
}
