/**
 * cacti_remote_by_http.c
 *
 * cacti_remote_by_http is a cacti data input script which does remote checks by http.
 * It queries the remote host with simple http-requests like i.e.
 *    "GET /load1 HTTP/1.1"
 * and takes the last line returned as a result for nagios.
 *
 * compile with i.e.
 *   diet gcc -O2 -Wall -o cacti_remote_by_http cacti_remote_by_http.c;sstrip cacti_remote_by_http
 *
 * Frank Bergmann, 2008-09-09
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define MAXRETRIES 6
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <stdint.h>

#include "lib/libfwb.h"

#define STR_HTTP_GET1 "GET /"
#define STR_HTTP_GET2 " HTTP/1.1\nHost: "
#define STR_HTTP_GET3 "\n\n"
#define STR_URL1 " (<a href=\"http://"
#define STR_URL2 "\">"
#define STR_URL3 "</a>)"
#define READBYTES 1024

static const char version_string[] = "version_string: $Id: cacti_remote_by_http.c 691 2017-06-04 23:12:04Z fwbergm $";

static const char usage[]="usage: cacti_remote_by_http [-D] -H ip-address -p port [-u UNIT] key\n";
static void printusage() { write1(usage, sizeof usage -1); _exit(2); }

static void print(const char *str) {
  register uint32_t i;

  if (!str) return;
  if (!*str) return;
  i = 1;
  for (;;) { if (!str[i]) break; ++i; }
  write1(str, i);
}

int main(int argc,char *const argv[]) {
  register char *p;
  register char *r;
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char http_buf[4*READBYTES];
  ssize_t buflen;
  int c;
  int port;
  const char *ip_address;
  const char *key;
  const char *unit;
  const char *portstr;
  char *http_body;
  int tcp_defer_accept;
  int retrycount;
  uint64_t val64;

  unit = NULL;
  port = tcp_defer_accept = 0;
  ip_address = key = portstr = 0;

  if (argc < 2) printusage();
  for(c=1; c<argc; c++) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') printusage();
      switch (argv[c][1]) {
        case 'h': printusage(); break;
        case 'D': tcp_defer_accept = 1; break;
        case 'H': c++; ip_address=argv[c]; break;
        case 'u': c++; unit=argv[c]; break;
        case 'p': c++; portstr = argv[c]; str2uint64((char *)portstr,&val64);port=(int)val64; break;
        default: break;
      }
    } else {
      key=argv[c];
      if (strlen(key) > READBYTES) { print("key too large\n"); _exit(3); }
    }
  }

  if (!key) { print("missing key - help with '-h'\n"); _exit(3); }
  if (!port) { print("missing port - help with '-h'\n"); _exit(3); }
  if (!ip_address) { print("missing ip-address - help with '-h'\n"); _exit(3); }
  p = stpcpy(http_buf, STR_HTTP_GET1);
  p = stpcpy(p, key);
  p = stpcpy(p, STR_HTTP_GET2);
  p = stpcpy(p, ip_address);
  p = stpcpy(p, STR_HTTP_GET3);
  p = http_buf;
  if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) _exit(3);
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(port);

  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) { _exit(3); }
  if (tcp_defer_accept) {
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) { close(sockfd); print("error on setsockopt"); _exit(3); }
  }
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
    close(sockfd);
    _exit(3);
  }
  if(write(sockfd,p,strlen(p)) == -1) { close(sockfd); _exit(3); }
  p = http_buf;
  r = http_buf+16;
  for(retrycount=0;;) {
    buflen = read(sockfd, p, READBYTES);
    if (buflen == -1) goto err_exit;
    if (buflen > 0) p += buflen;
    if (p < r) {
      if (retrycount >= MAXRETRIES) {
        err_exit:
        close(sockfd);
        if (buflen == -1) { _exit(3); }
        _exit(3);
      }
      usleep(100);
      retrycount++;
    } else break;
  }

  close(sockfd);
  *p = 0;
  while (*r != '\0') {
    r++;
    if (*r == '\n') {
      if (*(r-1) == '\n') break;
      if ((*(r-2) == '\n') && (*(r-1) == '\r')) break;
    }
  }
  http_body = r+1;
  if (p < http_body) http_body -= 2; /* this should never happen (TM) */
  if (http_body < http_buf) http_body = http_buf; /* this should never happen (TM) */
  /* get last line */
  for(;;) { p--; if (p == http_body) break; if ((unsigned char)*p >= 32) break; }
  p[1]='\0';
  for(;;) { p--; if (p < http_body) break; if ((unsigned char)*p < 32) break; }
  /* get value from last line */
  p++;
  r = http_buf;
  if (unit != NULL) {
    r = stpcpy(r, unit);
    *r++ = ':';
  }
  r = stpcpy(r, p);
  *r++ = '\n';
  write1(http_buf, r-http_buf);
  _exit(0);
}
