/**
 * $Id: check_http_fwb.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * check_http_fwb is a nagios plugin
 *
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define MAXRETRIES 6

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/times.h>
#include <arpa/inet.h>
#include <setjmp.h>
#include <stdint.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>

#include "lib/libfwb.h"

#define READBYTES 16384

#define NAGIOSRESULTPREFIXSTRING "HTTP "
#define NAGIOSRESULTOKSTRING "OK "
#define NAGIOSRESULTWARNSTRING "WARNING "
#define NAGIOSRESULTCRITSTRING "CRITICAL "
#define NAGIOSRESULTUNKNOWNSTRING "UNKNOWN "

static const char version_string[] = "version_string: $Id: check_http_fwb.c 691 2017-06-04 23:12:04Z fwbergm $";

static const char usage[] = 
  "usage: check_http_fwb [-0] [-T(O|W|C)] [-C(4)] [-P] -I ip-address -w warnlevel -c critlevel [-p port] [-H http11_hostname] [-u remotefile] [-A user_agent] [-t timeout] [-G|-R requestmethod] [-D]\n"
  "example: check_http_fwb -I 127.0.0.1 -w 2 -c 4 -p 8080 -H www.testdomain.de:8080 -u /myfile.txt -G\n";
static void printusage() { write1(usage, sizeof usage - 1); exit(1); }

enum { OK, WARNING, CRITICAL, UNKNOWN };

static jmp_buf check_remote_procstate;

static void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}

static void print(const char *str) {
  register uint32_t i;

  if (!str) return;
  if (!*str) return;
  i = 1;
  for (;;) { if (!str[i]) break; ++i; }
  write1(str, i);
}

static char *get_http_body(register char *r) {
  for (;;) {
    if (*r == 0) return 0;
    ++r;
    if (*r == '\n') {
      if (r[-1] == '\n') return r+1;
      if ((*(r-3) == '\r') && (*(r-2) == '\n') && (*(r-1) == '\r')) return r+1;
    }
  }
}

static char *sputint(register char *buf, uint32_t num, int leadzeroes) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;

  i=num; pbuf=buf;
  o = 0 - leadzeroes;
  if(o) {
    for(d=1;leadzeroes>0;leadzeroes--)d*=10;
  } else {
    d=1000000000;
  }
  for(;d>1;d/=10) {
    c=i/d;
    if((c>0)||(o)) {
      *pbuf++=c+'0'; i=i-(c*d);
      if(c>0)o=1;
      else ++o;
    }
  }
  *pbuf++=i+'0'; *pbuf='\0';
  return pbuf;
}


int main(int argc,char *const argv[]) {
  register char *p;
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char http_buf[2*READBYTES+1];
  char linebuf[READBYTES+1];
  char *r;
  int retvalue;
  ssize_t buflen;
  int c;
  uint64_t warn,crit;
  int port;
  int use_http10;
  const char *ip_address;
  const char *portstr;
  char *http_body;
  const char *http11_hostname;
  const char *user_agent;
  const char *url_file_path;
  int tcp_defer_accept;
  int use_get;
  const char *req_method;
  int retrycount;
  int timeout;
  int timeout_state;
  int64_t runtime;
  int runtime_negativeflag;
/*
  struct timeval rt_tval;
  struct timezone rt_tz;
*/
  struct timespec rt_tspec;
  int ping_mode;
  int allow_4xx;
  uint64_t val64;


  rt_tspec.tv_sec = 0;
  rt_tspec.tv_nsec = 0;
  runtime = -1;
  runtime_negativeflag = 0;
  ping_mode = port = tcp_defer_accept = use_get = timeout = use_http10 = 0;
  warn = crit = 0;
  http11_hostname = url_file_path = ip_address = req_method = (const char *) 0;
  user_agent = "check_http_fwb/1 (ngtx)";
  portstr = "80";
  retvalue = 2;
  timeout_state = UNKNOWN;

  if (argc < 2) { print("missing arguments - help with '-h'\n"); _exit(UNKNOWN); }
  for(c=1; c<argc; c++) {
    if (*argv[c] == '-') {
      switch (argv[c][1]) {
        case 'h': printusage(); break;
        case '0': use_http10 = 1; break;
        case 'D': tcp_defer_accept = 1; break;
        case 'P': ping_mode = 1; break;
        case 'G': use_get = 1; break;
        case 'R': c++; req_method = argv[c]; break;
        case 'w': c++; str2uint64(argv[c],&val64);warn=val64; break;
        case 'c': c++; str2uint64(argv[c],&val64);crit=val64; break;
        case 'I': c++; ip_address = argv[c]; break;
        case 'A': c++; user_agent = argv[c]; break;
        case 'H': c++; http11_hostname = argv[c]; break;
        case 'u': c++; url_file_path = argv[c]; break;
        case 'p': c++; portstr = argv[c]; str2uint64((char*)portstr,&val64);port=(int)val64; break;
        case 't': c++; str2uint64(argv[c],&val64);timeout=val64; break;
        case 'C': switch (argv[c][2]) {
                    case '4': allow_4xx = 1; break;
                    default: break;
                  } break;
        case 'T': switch (argv[c][2]) {
                    case 'O': timeout_state = OK; break;
                    case 'W': timeout_state = WARNING; break;
                    case 'C': timeout_state = CRITICAL; break;
                    default: break;
                  } break;
        default: break;
      }
    } else {
      printusage();
    }
  }
  if (ip_address == 0) print(usage);
  if (strlen(user_agent) > 255) { print("user_agent too long\n"); _exit(UNKNOWN); }
  if (warn == 0) warn = 5;
  if (crit == 0) crit = 10;
  if (timeout == 0) timeout = 10;
  if (port == 0) port = 80;
  if (ping_mode) {
    warn *= 1000;
    crit *= 1000;
  } else {
    warn *= 1000000;
    crit *= 1000000;
  }
  if (use_get && req_method) printusage();

  if (ping_mode) goto requeststringcreated;
  /* create http request */
  if (use_get) {
    p = stpcpy(http_buf, "GET ");
  } else {
    if (req_method) {
      p = stpcpy(http_buf, req_method);
      *p++ = ' ';
    } else {
      p = stpcpy(http_buf, "HEAD ");
    }
  }
  if (url_file_path) {
    p = stpcpy(p, url_file_path);
  } else {
    *p++ = '/';
  }
  p = stpcpy(p, " HTTP/1.");
  if (use_http10) {
    /* use_http10 overrides http-1.1-mode even if a http11_hostname is given */
    *p = '0';
  } else {
    *p = '1';
  }
  ++p;
  p = stpcpy(p, "\r\nConnection: close\r\n");
  p = stpcpy(p, "User-Agent: ");
  p = stpcpy(p, user_agent);
  *p++ = '\r'; *p++ = '\n';
  if (http11_hostname) {
    if (strlen(http11_hostname) > 255) { print("http11_hostname too long\n"); _exit(UNKNOWN); }
    p = stpcpy(p, "Host: ");
    p = stpcpy(p, http11_hostname);
    *p++ = '\r'; *p++ = '\n';
  }
  *p++ = '\r'; *p++ = '\n';
  *p = 0;
  p = http_buf;

  requeststringcreated:

  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    if (ping_mode) timeout_state = CRITICAL;
    switch (timeout_state) {
      case OK: print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTOKSTRING "- plugin internal timeout\n"); break;
      case WARNING: print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTWARNSTRING "- plugin internal timeout\n"); break;
      case CRITICAL: print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- plugin internal timeout\n"); break;
      default: print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- plugin internal timeout\n"); break;
    }
    _exit(timeout_state);
  }

  /* send request */
  if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on inet_aton\n");
    _exit(UNKNOWN);
  }
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(port);

  /* obsoleted in POSIX.1-2008:  if (!gettimeofday(&rt_tval, &rt_tz)) ... */
  if (!clock_gettime(CLOCK_REALTIME, &rt_tspec)) runtime = (1000000 * rt_tspec.tv_sec) + (rt_tspec.tv_nsec / 1000);

  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on opening socket\n");
    _exit(UNKNOWN);
  }
  if (tcp_defer_accept) {
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) {
      close(sockfd);
      print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- error on setsockopt");
      _exit(UNKNOWN);
    }
  }
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- error on connect\n");
    _exit(CRITICAL);
  }

  if (ping_mode) goto requestdone;

  if(write(sockfd,p,strlen(p)) == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on write\n");
    _exit(UNKNOWN);
  }

  /* get reply */
  p = http_buf;
  r = http_buf + 17;
  http_body = 0;
  for(retrycount=0;;) {
    buflen = read(sockfd, p, READBYTES);
    if (buflen == -1) goto err_exit;
    if (buflen > 0) p += buflen;
    *p = 0;
    if (p >= r) http_body = get_http_body(r);
    if (buflen > 500) {
      /* only do more read()-calls if first read got some hundred bytes */
      int mrb = READBYTES-buflen;
      buflen = read(sockfd, p, mrb);
      while (buflen > 0) {
        /* discard read data */
        buflen = read(sockfd, p, mrb);
      }
    } else {
      if (http_body) goto requestdone;
    }
    if ((p < r) || buflen) {
      if (retrycount >= MAXRETRIES) {
        err_exit:
        close(sockfd);
        if (buflen == -1) { print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on read\n"); _exit(3); }
        if (buflen == -1) {
          print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on read\n");
          _exit(UNKNOWN);
        }
        print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- received data underrun #");
        write1(http_buf,p-http_buf);
        print("#\n");
        _exit(UNKNOWN);
      }
      /* usleep(100); */
      /*nanosleep(&sleep_between_retries, 0);*/
      retrycount++;
    } else break;
  }

  *p = 0;
  http_body = get_http_body(r);

  requestdone:

  /*close(sockfd);*/
  /* obsoleted in POSIX.1-2008:  if (!gettimeofday(&rt_tval, &rt_tz)) ... */
  if (!clock_gettime(CLOCK_REALTIME, &rt_tspec)) runtime = ((1000000 * rt_tspec.tv_sec) + (rt_tspec.tv_nsec / 1000)) - runtime;

  if (ping_mode) {
    retvalue = 0;
    *http_buf = 0;
    goto parsingdone;
  }

  /* parse reply */
  while (*r != '\0') {
    r++;
    if (*r == '\n') {
      if (*(r-1) == '\n') break;
      if ((*(r-2) == '\n') && (*(r-1) == '\r')) break;
    }
  }
  http_body = r+1;
  if (p < http_body) http_body -= 2; /* this should never happen (TM) */
  if (http_body < http_buf) http_body = http_buf; /* this should never happen (TM) */

  /* http reply? */
  p = http_buf;
  r = "HTTP/1.";
  for(;;) {
    if (*p != *r) goto invalid_response;
    ++p; ++r; if (!*r) break;
  }

  /* actually http? */
  if ((*p != '1') && (*p != '0')) goto invalid_response;
  if (*++p != ' ') goto invalid_response;
  ++p;

  /* http code, only <400 is OK */
  if ((*p > '3') && !allow_4xx) retvalue = 2;
  else retvalue = 0;
  for(;;) {
    if ((*p == '\r') || (*p == '\n')) break;
    ++p; if (!*p) break;
  }
  *p = 0;

  parsingdone:

  /* slow response? */
  if (runtime < -1) {
    /* negative runtime: ntp time drift? */
    runtime_negativeflag = 1; /* most simple :-) */
    runtime = 0 - runtime;
  } else {
    /* only take care of runtime if there's no time drift */
    if ((retvalue < 1) && (runtime >= warn)) retvalue = 1;
    if ((retvalue < 2) && (runtime >= crit)) retvalue = 2;
  }

  /* create nagios reply string */
  p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING);
  switch(retvalue) {
    case 0: p = stpcpy(p, NAGIOSRESULTOKSTRING); break;
    case 1: p = stpcpy(p, NAGIOSRESULTWARNSTRING); break;
    case 2: p = stpcpy(p, NAGIOSRESULTCRITSTRING); break;
    default: retvalue = 3;
  }
  if (*http_buf) {
    p = stpcpy(p, http_buf);
    *p = ' '; ++p;
  }

  if (runtime != -1) {
    p = stpcpy(p, "- ");
    if (runtime_negativeflag) { *p = '-'; ++p; } /* ntp time drift? */
    r = sputint(http_buf, (uint32_t)(runtime/1000000), 0);
    *r++ = '.';
    sputint(r, (uint32_t)(runtime - (1000000 * (runtime/1000000))), 5);
    r = http_buf;
    p = stpcpy(p, r) - 3;
    p = stpcpy(p, " second response time |time=");
    p = stpcpy(p, r);
    p = stpcpy(p, "s\n");
  }
  write1(linebuf, p - linebuf);
  _exit(retvalue);

  invalid_response:
  print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- Invalid HTTP response received\n");
  _exit(2);
}
