/**
 * $Id: check_remote_by_http.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * check_remote_by_http is a nagios plugin which does remote checks by http.
 * It queries the remote host with simple http-requests like i.e.
 *    "GET /load1 HTTP/1.1"
 * and takes the last line returned as a result for nagios.
 *
 * compile with i.e.
 *   diet gcc -O2 -Wall -o check_remote_by_http check_remote_by_http.c;sstrip check_remote_by_http
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define TESTMULTIREQUESTS
#undef TESTMULTIREQUESTS

#define MAXRETRIES 6

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <setjmp.h>
#include <stdint.h>

#include "lib/libfwb.h"

#define NAGIOSRESULTPREFIXSTRING "HTTP "

#define STR_HTTP_GET1 "GET /"
#define STR_HTTP_GET2 " HTTP/1.1\nHost: "
#define STR_HTTP_GET3 "\n\n"
#define STR_URL1 " (<a href=\"http://"
#define STR_URL2 "\"> "
#define STR_URL3 " </a>)"
#define READBYTES 16384

static const char version_string[] = "version_string: $Id: check_remote_by_http.c 691 2017-06-04 23:12:04Z fwbergm $";

static jmp_buf check_remote_procstate;

enum { OK, WARNING, CRITICAL, UNKNOWN };

const char usage[]="usage: check_remote_by_http [-D] [-T(O|W|C)] [-L] [-A(0|1|2|3|4)] [-W warnlevellow] [-C critlevellow] -w warnlevel -c critlevel -H ip-address -p port [-t timeout] [-u UNIT] key\n";
static void printusage() { write1(usage, sizeof usage - 1); exit(1); }

static void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}

static void print(const char *str) { write1(str,strlen(str)); }

static char *get_http_body(register char *r) {
  for (;;) {
    if (*r == 0) return 0;
    ++r;
    if (*r == '\n') {
      if (r[-1] == '\n') goto check_end;
      if ((*(r-3) == '\r') && (*(r-2) == '\n') && (*(r-1) == '\r')) goto check_end;
    }
  }
  check_end:
  ++r;
  if (*r) return r;
  return 0;
}

int main(int argc,char *const argv[]) {
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char http_buf[4*READBYTES];
  char linebuf[2*READBYTES];
  register char *p;
  register char *r;
  register ssize_t buflen;
  register int c;
  register long warn,crit;
  register int retvalue;
  long warnlow, critlow;
  int port;
  const char *ip_address;
  const char *key;
  char *unit;
  const char *portstr;
  char *http_body;
  int tcp_defer_accept;
  int warni,criti;
  int retrycount;
  int timeout;
  int timeout_state;
  int show_link;
  int mode_show_additional_data;
  uint64_t val64;
#ifdef TESTMULTIREQUESTS
  int testmultirequests;

  for(testmultirequests=0;testmultirequests<20000;testmultirequests++) {
#endif

  unit = NULL;
  warnlow = critlow = warn = warni = crit = criti = port = tcp_defer_accept =
    timeout = show_link = 0;
  mode_show_additional_data = 1;
  ip_address = argv[0];
  portstr = "80";
  retvalue = 2;
  key = ip_address;
  timeout_state = UNKNOWN;

  if (argc < 2) { print("missing arguments - help with '-h'\n"); _exit(3); }
  for(c=1; c<argc; c++) {
    if (*argv[c] == '-') {
      switch (argv[c][1]) {
        case 'h': printusage(); exit(0); break;
        case 'D': tcp_defer_accept = 1; break;
        case 'W': c++; str2uint64(argv[c],&val64);warnlow=val64; break;
        case 'C': c++; str2uint64(argv[c],&val64);critlow=val64; break;
        case 'w': c++; warni=c;str2uint64(argv[c],&val64);warn=val64; break;
        case 'c': c++; criti=c;str2uint64(argv[c],&val64);crit=val64; break;
        case 'H': c++; ip_address=argv[c]; break;
        case 'u': c++; unit=argv[c]; break;
        case 'p': c++; portstr = argv[c]; str2uint64((char *)portstr,&val64);port=(int)val64; break;
        case 't': c++; str2uint64(argv[c],&val64);timeout=val64; break;
        case 'L': show_link = 1; break;
        case 'A': mode_show_additional_data = argv[c][2] - '0';
                  if (mode_show_additional_data < 0) printusage();
                  if (mode_show_additional_data > 4) printusage();
                  break;
        case 'T': switch (argv[c][2]) {
                    case 'O': timeout_state = OK; break;
                    case 'W': timeout_state = WARNING; break;
                    case 'C': timeout_state = CRITICAL; break;
                    default: break;
                  } break;
        default: printusage(); break;
      }
    } else {
      key = argv[c];
      if (strlen(key) > READBYTES) { print("key too large\n"); _exit(CRITICAL); }
    }
  }
  if (timeout == 0) timeout = 10;


  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    switch (timeout_state) {
      case OK: print(NAGIOSRESULTPREFIXSTRING "OK - plugin internal timeout\n"); break;
      case WARNING: print(NAGIOSRESULTPREFIXSTRING "WARNING - plugin internal timeout\n"); break;
      case CRITICAL: print(NAGIOSRESULTPREFIXSTRING "CRITICAL - plugin internal timeout\n"); break;
      default: print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - plugin internal timeout\n"); break;
    }
    _exit(timeout_state);
  }

  p = stpcpy(http_buf, STR_HTTP_GET1);
  p = stpcpy(p, key);
  p = stpcpy(p, STR_HTTP_GET2);
  p = stpcpy(p, ip_address);
  p = stpcpy(p, STR_HTTP_GET3);
  p = http_buf;
  if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) _exit(3);
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(port);

#ifdef TESTMULTIREQUESTS
  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on opening socket\n"); continue; }
#else
  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on opening socket\n"); _exit(3); }
#endif
  if (tcp_defer_accept) {
#ifdef TESTMULTIREQUESTS
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) { close(sockfd); print("error on setsockopt"); continue; }
#else
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) { close(sockfd); print("error on setsockopt"); _exit(3); }
#endif
  }
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
    close(sockfd);
#ifdef TESTMULTIREQUESTS
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on connect\n"); continue;
  }
  if(write(sockfd,p,strlen(p)) == -1) { close(sockfd); print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on write\n"); continue; }
#else
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on connect\n"); _exit(3);
  }
  if(write(sockfd,p,strlen(p)) == -1) { close(sockfd); print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on write\n"); _exit(3); }
#endif
  p = http_buf;
  r = http_buf + 17;
  http_body = 0;
  for(retrycount=0;;) {
    buflen = read(sockfd, p, READBYTES);
    if (buflen == -1) goto err_exit;
    if (buflen > 0) p += buflen;
    *p = 0;
    if (p >= r) http_body = get_http_body(r);
    if (buflen > 500) {
      /* do more read()-calls if read got some hundred bytes */
      int mrb = READBYTES-buflen;
      buflen = read(sockfd, p, mrb);
      while (buflen > 0) {
        mrb = mrb-buflen;
        if (mrb<0) break;
        p += buflen;
        buflen = read(sockfd, p, mrb);
      }
      if (buflen > 0) p += buflen;
    } else {
      if (http_body) goto got_http_body;
    }
    if ((p < r) || buflen) {
      if (retrycount >= MAXRETRIES) {
        err_exit:
        close(sockfd);
#ifdef TESTMULTIREQUESTS
        if (buflen == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on read\n"); continue; }
        print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - received data underrun #");
        write1(http_buf,p-http_buf);
        print("#\n");
        continue;
#else
        if (buflen == -1) { print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on read\n"); _exit(3); }
        print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - received data underrun #");
        write1(http_buf,p-http_buf);
        print("#\n");
        _exit(3);
#endif
      }
      /*usleep(100);*/
      retrycount++;
    } else break;
  }

  *p = 0;
  http_body = get_http_body(r);
  got_http_body:
  /*close(sockfd);*/
  if ((p-http_buf) == READBYTES) {
    #define OVERRUN "Buffer overrun, please adjust READBYTES and re-compile.\n"
    write1(OVERRUN, sizeof(OVERRUN) - 1);
    _exit(3);
  }
  if (p < http_body) http_body -= 2; /* this should never happen (TM) */
  if (http_body < http_buf) http_body = http_buf; /* this should never happen (TM) */
  /* get last line */
  for(;;) { p--; if (p == http_body) break; if ((unsigned char)*p >= 32) break; }
  p[1]='\0';
  for(;;) { p--; if (p < http_body) break; if ((unsigned char)*p < 32) break; }
  /* skip blanks and ctrl chars */
  for(;;) { p++; if ((unsigned char)*p > 32) break; }
  /* construct output line */
  r = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING);
  if (str2uint64(p,&val64) != 0) {
    retvalue = 0;
    if ((val64 > warn) || ((warnlow) && (val64 < warnlow))) retvalue = 1;
    if ((val64 > crit) || ((critlow) && (val64 < critlow))) retvalue = 2;
  } else {
    retvalue=3;
  }
  switch (retvalue) {
    case 0: r = stpcpy(r,"OK - "); break;
    case 1: r = stpcpy(r,"WARNING - ");break;
    case 2: r = stpcpy(r,"CRITICAL - "); break;
    default: r = stpcpy(r,"UNKNOWN - "); retvalue=3; break;
  }
  r = stpcpy(r,p);
  if (unit!=NULL) {
    r = stpcpy(r,unit);
  }
  if (tcp_defer_accept) { r = stpcpy(r, ", *TDA* active"); } /* for debugging */
  /* append additional data if retvalue >= mode_show_additional_data */
  if (retvalue >= mode_show_additional_data) {
    register char *s = p - 1;
    /* get line before last line */
    for(;;) { s--; if (s<=http_body) break; if ((unsigned char)*s>=32) break; }
    s[1]='\0';

    /* for(;;) { s--; if (s<http_body) break; if ((unsigned char)*s<32) break; } */
    for(;;) { s--; if (s<http_body) break; if ((unsigned char)*s=='\n') break; }

    s++;
    if (s<http_body) s=http_body;

    /* construct link */
    if (show_link) {
      r = stpcpy(r, STR_URL1);
      r = stpcpy(r, ip_address);
      *r++ = ':';
      r = stpcpy(r, portstr);
      *r++ = '/';
      r = stpcpy(r, key);
      r = stpcpy(r, STR_URL2);
      r = stpcpy(r, s);
      r = stpcpy(r, STR_URL3);
    } else {
      *r = ' '; ++r;
      *r = '('; ++r;
      r = stpcpy(r, s);
      *r = ')'; ++r;
    }
  }
  *r++ = ' '; *r++ = '|'; *r++ = ' ';
  r = stpcpy(r,key);
  *r++ = '=';
  r = stpcpy(r,p);
  if (unit!=NULL) {
    while (*unit == ' ') ++unit;
    p = unit;
    while (*p != ' ') ++p;
    *p = 0;
    r = stpcpy(r,unit);
  }
  *r++ = ';';
  r = stpcpy(r,argv[warni]);
  *r++ = ';';
  r = stpcpy(r,argv[criti]);
  *r++ = '\n';
  print(linebuf);
#ifdef TESTMULTIREQUESTS
  }
#endif
  _exit(retvalue);
}
