/**
 * check_remote_by_http_time.c
 *
 * check_remote_by_http_time is a nagios plugin which does a remote current time check by http.
 * It queries the remote host with a simple http-requests like that:
 *    "GET /timediff/1222253926 HTTP/1.1"
 * and takes the last line returned as a result for nagios.
 *
 * compile with i.e.
 *   diet gcc -O2 -Wall -o check_remote_by_http_time check_remote_by_http_time.c;sstrip check_remote_by_http_time
 *
 * Frank Bergmann, 2008-09-24
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define MAXRETRIES 6

#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <stdint.h>
#include <time.h>

#include "knoerre/knoerre.h"
#include "lib/libfwb.h"

#define STR_HTTP_GET1 "GET /"CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY
#define STR_HTTP_GET2 " HTTP/1.1\nHost: "
#define STR_HTTP_GET3 "\n\n"
#define STR_URL1 " (<a href=\"http://"
#define STR_URL2 "\"> "
#define STR_URL3 " </a>)"
#define READBYTES 1024

#define NAGIOSRESULTPREFIXSTRING "HTTP "
#define NAGIOSRESULTOKSTRING "OK "
#define NAGIOSRESULTWARNSTRING "WARNING "
#define NAGIOSRESULTCRITSTRING "CRITICAL "
#define NAGIOSRESULTUNKNOWNSTRING "UNKNOWN "
enum { OK, WARNING, CRITICAL, UNKNOWN };

static const struct timespec sleep_between_retries = { 0, 200000000 };
static const struct timeval tcp_timeout = { 1, 0 };

static const char version_string[] = "version_string: $Id: check_remote_by_http_time.c 691 2017-06-04 23:12:04Z fwbergm $";

static const char usage[]="usage: check_remote_by_http_time [-D] -w warnlevel -c critlevel -H ip-address -p port [-u UNIT]\n";
static void printusage() { write1(usage, sizeof usage - 1); exit(1); }

#ifdef NULLDEF
static const int timeout = 10;

static jmp_buf check_remote_procstate;
static void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}
#endif

static void print(const char *str) {
  register uint32_t i;

  if (!str) return;
  if (!*str) return;
  i = 1;
  for (;;) { if (!str[i]) break; ++i; }
  write1(str, i);
}

static char *get_http_body(register char *r) {
  for (;;) {
    if (*r == 0) return 0;
    ++r;
    if (*r == '\n') {
      if (*(r-1) == '\n') goto check_end;
      if ((*(r-3) == '\r') && (*(r-2) == '\n') && (*(r-1) == '\r')) goto check_end;
    }
  }
  check_end:
  ++r;
  if (*r) return r;
  return 0;
}

int main(int argc,char *const argv[]) {
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char http_buf[4*READBYTES];
  char linebuf[255];
  register char *p;
  register char *r;
  register ssize_t buflen;
  register int c;
  register long warn,crit;
  register int retvalue;
  int port;
  const char *ip_address;
  const char *unit;
  const char *portstr;
  char *http_body;
  int tcp_defer_accept;
  int retrycount;
  time_t tnow;
  uint64_t val64;

  unit = NULL;
  warn = crit = port = tcp_defer_accept = 0;
  ip_address = argv[0];
  portstr = "80";
  retvalue = 2;

  if (argc < 2) { print("missing arguments - help with '-h'\n"); _exit(UNKNOWN); }
  for(c=1; c<argc; c++) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') { printusage(); exit(UNKNOWN); }
      switch (argv[c][1]) {
        case 'h': printusage(); exit(0); break;
        case 'D': tcp_defer_accept = 1; break;
        case 'w': c++; str2uint64(argv[c],&val64);warn=val64; break;
        case 'c': c++; str2uint64(argv[c],&val64);crit=val64; break;
        case 'H': c++; ip_address=argv[c]; break;
        case 'u': c++; unit=argv[c]; break;
        case 'p': c++; portstr = argv[c]; str2uint64((char*)portstr,&val64);port=(int)val64; break;
        default: break;
      }
    } else {
      printusage(); exit(UNKNOWN);
    }
  }

  tnow = time(NULL);
  p = stpcpy(http_buf, STR_HTTP_GET1);
  p = sputlongs(p, tnow);
  --p;
  p = stpcpy(p, STR_HTTP_GET2);
  p = stpcpy(p, ip_address);
  stpcpy(p, STR_HTTP_GET3);

  if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on inet_aton\n");
    _exit(UNKNOWN);
  }
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(port);

  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on opening socket\n");
    _exit(UNKNOWN);
  }
  if (tcp_defer_accept) {
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) {
      close(sockfd);
      print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- error on setsockopt");
      _exit(UNKNOWN);
    }
  }
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on connect\n");
    _exit(UNKNOWN);
  }
  p = http_buf;

  setsockopt(sockfd, SOL_SOCKET, SO_SNDTIMEO, &tcp_timeout, sizeof(tcp_timeout));
  setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, &tcp_timeout, sizeof(tcp_timeout));

  if(write(sockfd,p,strlen(p)) == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on write\n");
    _exit(UNKNOWN);
  }
  r = http_buf + 17;
  retrycount = 0;

#ifdef NULLDEF
  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    goto also_timeout;
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- plugin internal timeout\n");
    _exit(UNKNOWN);
  }
#endif

  http_body = 0;
  for(;;) {
    buflen = read(sockfd, p, READBYTES);
    if (buflen == -1) goto err_exit;
    if (buflen > 0) p += buflen;
    *p = 0;
    if (p >= r) http_body = get_http_body(r);
    if (buflen > 500) {
      /* do more read()-calls if read got some hundred bytes */
      int mrb = READBYTES-buflen;
      buflen = read(sockfd, p, mrb);
      while (buflen > 0) {
        mrb = mrb-buflen;
        if (mrb<0) break;
        p += buflen;
        buflen = read(sockfd, p, mrb);
      }
      if (buflen > 0) p += buflen;
    } else {
      if (http_body) goto got_http_body;
    }
    if ((p < r) || buflen) {
      if (retrycount >= MAXRETRIES) {
        err_exit:
        close(sockfd);
        if (buflen == -1) {
          print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on read\n");
          _exit(UNKNOWN);
        }
        print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- received data underrun #");
        write1(http_buf,p-http_buf);
        print("#\n");
        _exit(UNKNOWN);
      }
      /* usleep(100); */
      nanosleep(&sleep_between_retries, 0);
      retrycount++;
    } else break;
  }

  *p = 0;
  http_body = get_http_body(r);
  got_http_body:
  close(sockfd);
  if ((p-http_buf) == READBYTES) {
    #define OVERRUN "Buffer overrun, please adjust READBYTES and re-compile.\n"
    write1( OVERRUN, sizeof(OVERRUN) - 1);
    _exit(3);
  }
  if (p < http_body) http_body -= 2; /* this should never happen (TM) */
  if (http_body < http_buf) http_body = http_buf; /* this should never happen (TM) */
  /* get last line */
  for(;;) { --p; if (p == http_body) break; if ((unsigned char)*p >= 32) break; }
  p[1] = 0;
  for(;;) { --p; if (p < http_body) break; if ((unsigned char)*p < 32) break; }
  /* get value from last line */
  p++;
  /* construct output line */
  r = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING);
  if (str2uint64(p,&val64) != 0) {
    if (val64 > warn) {
      if (val64 > crit) {
        r = stpcpy(r, NAGIOSRESULTCRITSTRING "- ");
        retvalue = CRITICAL;
      } else {
        r = stpcpy(r, NAGIOSRESULTWARNSTRING "- ");
        retvalue = WARNING;
      }
    } else {
      r = stpcpy(r, NAGIOSRESULTOKSTRING "- ");
      retvalue = OK;
    }
  } else {
    r = stpcpy(r, NAGIOSRESULTUNKNOWNSTRING "- ");
    retvalue = UNKNOWN;
  }
  r = stpcpy(r,p);
  if (unit!=NULL) {
    r = stpcpy(r,unit);
  }
  if (tcp_defer_accept) { r = stpcpy(r, ", *TDA* active"); } /* for debugging */
  /* append link if retvalue not "OK" */
  if (retvalue > 0) {
    register char *s = p - 1;
    /* get line before last line */
    for(;;) { s--; if (s<=http_body) break; if ((unsigned char)*s>=32) break; }
    s[1]='\0';
    for(;;) { s--; if (s<http_body) break; if ((unsigned char)*s<32) break; }
    s++;
    if (s<http_body) s=http_body;
    /* construct link */
    r = stpcpy(r, STR_URL1);
    r = stpcpy(r, ip_address);
    *r++ = ':';
    r = stpcpy(r, portstr);
    *r++ = '/';
    r = stpcpy(r, CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY);
    r = sputlongs(r, tnow);
    --r;
    r = stpcpy(r, STR_URL2);
    r = stpcpy(r, s);
    r = stpcpy(r, STR_URL3);
  }
  *r++ = ' '; *r++ = '|'; *r++ = ' ';
  r = stpcpy(r, CONFIG_KNOERRE_CHECK_TIMEDIFF_KEY); --r;
  *r++ = '=';
  r = stpcpy(r,p);
  *r++ = '\n';
  *r = 0;
  write1( linebuf, r - linebuf);
  _exit(retvalue);
}
