/**
 * $Id: check_remote_by_udp.c 691 2017-06-04 23:12:04Z fwbergm $
 * check_remote_by_udp.c
 *
 * check_remote_by_udp is a nagios plugin which does remote checks by raw udp.
 * It queries the remote host with simple requests like i.e.
 *    "load1\n"
 * and takes the last line returned as a result for nagios.
 *
 * compile with i.e.
 *   diet gcc -O2 -Wall -o check_remote_by_udp check_remote_by_udp.c;sstrip check_remote_by_udp
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define TESTMULTIREQUESTS
#undef TESTMULTIREQUESTS

#undef NO_PORT_RANDOMIZATION
#define NO_PORT_RANDOMIZATION

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <poll.h>
#include <setjmp.h>
#include <time.h>
#include <string.h>
#include <stdint.h>

#include "lib/libfwb.h"

#define READBYTES 4096

#define STR_URL1 " (<a href=\"http://"
#define STR_URL2 "\"> "
#define STR_URL3 " </a>)"
#define NAGIOSRESULTPREFIXSTRING "UDP "
#define NAGIOSRESULTOKSTRING "OK "
#define NAGIOSRESULTWARNSTRING "WARNING "
#define NAGIOSRESULTCRITSTRING "CRITICAL "
#define NAGIOSRESULTUNKNOWNSTRING "UNKNOWN "
enum { OK, WARNING, CRITICAL, UNKNOWN };

static const char version_string[] = "version_string: $Id: check_remote_by_udp.c 691 2017-06-04 23:12:04Z fwbergm $";

static const char usage[]="usage: check_remote_by_udp [-L] [-A(0|1|2|3|4)] [-W warnlevellow] [-C critlevellow] -w warnlevel -c critlevel -H ip-address -p port [-t timeout] [-u UNIT] key\n";
static void printusage() { write1(usage, sizeof usage - 1); exit(1); }

static jmp_buf check_remote_procstate;
static void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}

static void print(const char *str) {
  register uint32_t i;

  if (!str) return;
  if (!*str) return;
  i = 1;
  for (;;) { if (!str[i]) break; ++i; }
  write1(str, i);
}

int main(int argc,char *const argv[]) {
  register char *p;
  int tmpint;
  ssize_t bytecount;
  int sockfd;
  struct sockaddr_in servin;
  struct sockaddr_in clntin;
  struct pollfd st_poll;
  char udp_buf[READBYTES+1];
  char linebuf[READBYTES+1];
  long warn,crit;
  uint64_t value;
  int retvalue;
  long warnlow, critlow;
  int port;
  const char *ip_address;
  const char *key;
  char *unit;
  const char *portstr;
  int warni,criti;
  int timeout;
  char *lastline;
  int show_link;
  int mode_show_additional_data;
#ifndef NO_PORT_RANDOMIZATION
  struct timeval tv;
  uint32_t mrand;
#endif
#ifdef TESTMULTIREQUESTS
  int testmultirequests;

  for(testmultirequests=0;testmultirequests<20000;testmultirequests++) {
#endif

  unit = NULL;
  warnlow = critlow = warn = warni = crit = criti = port = timeout = show_link = 0;
  mode_show_additional_data = 1;
  ip_address = argv[0];
  portstr = "80";
  retvalue = 2;
  key = ip_address;

  if (argc < 2) printusage();
  for(tmpint=1; tmpint<argc; ++tmpint) {
    if (*argv[tmpint] == '-') {
/*
      if (argv[tmpint][2] != '\0') printusage();
*/
      switch (argv[tmpint][1]) {
        case 'h': printusage(); break;
        case 'w': ++tmpint; warni=tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); warn=(long) value; break;
        case 'W': ++tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); warnlow=(long) value; break;
        case 'c': ++tmpint; criti=tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); crit=(long) value; break;
        case 'C': ++tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); critlow=(long) value; break;
        case 't': ++tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); timeout=(long) value; break;
        case 'H': ++tmpint; ip_address=argv[tmpint]; break;
        case 'u': ++tmpint; unit=argv[tmpint]; break;
        case 'p': ++tmpint; portstr = argv[tmpint]; if(!str2uint64(argv[tmpint], &value)) printusage(); port=(int) value; break;
        case 'L': show_link = 1; break;
        case 'A': mode_show_additional_data = argv[tmpint][2] - '0';
                  if (mode_show_additional_data < 0) printusage();
                  if (mode_show_additional_data > 4) printusage();
                  break;
        default: printusage(); break;
      }
    } else {
      key = argv[tmpint];
      if (strlen(key) > READBYTES) { print("key too large\n"); _exit(CRITICAL); }
    }
  }
  if (timeout == 0) timeout = 10;


  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- plugin internal timeout\n");
    _exit(UNKNOWN);
  }

  /* setup sockaddr */
  memset(&servin, 0, sizeof servin);
  servin.sin_family = AF_INET;
  servin.sin_port = htons(port);
  if (inet_aton(ip_address, &servin.sin_addr) == 0) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on inet_aton\n");
    _exit(UNKNOWN);
  }

  /* socket */
  sockfd = tmpint = socket(PF_INET, SOCK_DGRAM, 0);
  if (sockfd == -1) {
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on opening socket\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(UNKNOWN);
#endif
  }

  tmpint = fcntl(sockfd, F_GETFL);
  tmpint |= O_NONBLOCK;
  tmpint = fcntl(sockfd, F_SETFL, tmpint);
  if (tmpint) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on fcntl\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(UNKNOWN);
#endif
  }

  memset(&clntin, 0, sizeof clntin);
  clntin.sin_family = AF_INET;
#ifndef NO_PORT_RANDOMIZATION
  /* dirty port randomization */
  mrand = getpid() % 32766;
  ++mrand;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  /* obsoleted in POSIX.1-2008: */
  if (!gettimeofday(&tv, 0)) {
    mrand = tv.tv_sec % mrand;
    mrand = (tv.tv_usec % 131075) * mrand;
  } else {
    mrand = time(0) % mrand;
  }
  mrand %= (65535 - 1024);
  mrand += 1024;
  clntin.sin_port = htons(mrand);
  tmpint = bind(sockfd, (struct sockaddr *) &clntin, sizeof clntin);
  if (tmpint) {
#endif
    /* let os choose port */
    clntin.sin_port = 0;
    tmpint = bind(sockfd, (struct sockaddr *) &clntin, sizeof clntin);
    if (tmpint) {
      close(sockfd);
      print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on bind\n");
#ifdef TESTMULTIREQUESTS
      continue;
#else
      _exit(UNKNOWN);
#endif
    }
#ifndef NO_PORT_RANDOMIZATION
  }
#endif

  tmpint = connect(sockfd, (struct sockaddr *) &servin, sizeof servin);
  if (tmpint == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on connect\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(UNKNOWN);
#endif
  }

  /* send */
  p = stpcpy(udp_buf, key);
  *p = '\n'; ++p;

  tmpint = send(sockfd, udp_buf, p - udp_buf, 0);
  if (tmpint == -1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on send\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(UNKNOWN);
#endif
  }

  /* recv */
  st_poll.fd = sockfd;
  st_poll.events = POLLIN;
  tmpint = poll(&st_poll, 1, -1); /* no check for POLLIN */
  bytecount = recv(sockfd, udp_buf, sizeof udp_buf, 0);
  if (bytecount < 1) {
    close(sockfd);
    print(NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- error on recv\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(UNKNOWN);
#endif
  }
  udp_buf[bytecount] = 0;

  /* get last line */
  for (tmpint = bytecount;;) {
    --tmpint;
    if (!tmpint) break;
    if ((unsigned char) udp_buf[tmpint] >= 32) break;
  }
  p = udp_buf + tmpint + 1;
  *p = 0;
  for (;;) {
    --tmpint;
    if (tmpint < 0) break;
    if ((unsigned char) udp_buf[tmpint] < 32) break;
  }
  /* skip blanks and ctrl chars */
  for (;;) {
    ++tmpint;
    if ((unsigned char) udp_buf[tmpint] > 32) break;
  }
  p = udp_buf + tmpint;
  lastline = p;

  /* construct output line */
  value = 0;
  if (str2uint64(p, &value)) {
    retvalue = OK;
    if ((value > warn) || ((warnlow) && (value < warnlow))) retvalue = WARNING;
    if ((value > crit) || ((critlow) && (value < critlow))) retvalue = CRITICAL;
  } else {
    retvalue = UNKNOWN;
  }
  switch (retvalue) {
    case 0: p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING NAGIOSRESULTOKSTRING "- "); break;
    case 1: p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING NAGIOSRESULTWARNSTRING "- ");break;
    case 2: p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING NAGIOSRESULTCRITSTRING "- "); break;
    default: p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING NAGIOSRESULTUNKNOWNSTRING "- "); retvalue=3; break;
  }

  p = stpcpy(p, lastline);
  if (unit != NULL) {
    p = stpcpy(p, unit);
  }
  /* append additional data if retvalue >= mode_show_additional_data */
  if (retvalue >= mode_show_additional_data) {
    /* get line before last line */
    for (tmpint = lastline - udp_buf;;) {
      if (!tmpint) break;
      --tmpint;
      if ((unsigned char) udp_buf[tmpint] >= 32) break;
    }

    if (lastline != (udp_buf+tmpint)) {
      udp_buf[tmpint + 1] = 0;
      for (;;) {
        --tmpint;
        if (tmpint < 0) break;
        if ((unsigned char) udp_buf[tmpint] < 32) break;
      }
      /* skip blanks and ctrl chars */
      for (;;) {
        ++tmpint;
        if ((unsigned char) udp_buf[tmpint] > 32) break;
      }
    }
    if (show_link) {
      p = stpcpy(p, STR_URL1);
      p = stpcpy(p, ip_address);
      *p = ':'; ++p;
      p = stpcpy(p, portstr);
      *p = '/'; ++p;
      p = stpcpy(p, key);
      p = stpcpy(p, STR_URL2);
      p = stpcpy(p, udp_buf + tmpint);
      p = stpcpy(p, STR_URL3);
    } else {
      *p = ' '; ++p;
      *p = '('; ++p;
      p = stpcpy(p, udp_buf + tmpint);
      *p = ')'; ++p;
    }
  }
  /* performance data */
  p = stpcpy(p, " | ");
  p = stpcpy(p, key);
  *p = '='; ++p;
  p = stpcpy(p, lastline);
  if (unit != NULL) {
    while (*unit == ' ') ++unit;
    for (tmpint = 0;; ++tmpint)
      if (unit[tmpint] <= 32) break;
    unit[tmpint] = 0;
    p = stpcpy(p, unit);
  }
  *p = ';'; ++p;
  p = stpcpy(p, argv[warni]);
  *p = ';'; ++p;
  p = stpcpy(p, argv[criti]);
  *p = '\n'; ++p;
  *p = 0;
  print(linebuf);
  close(sockfd);
#ifdef TESTMULTIREQUESTS
  }
#endif
  _exit(retvalue);
}
