/**
 * $Id: check_remote_by_udp_multi_net.c 691 2017-06-04 23:12:04Z fwbergm $
 * check_remote_by_udp_multi_net.c
 *
 * check_remote_by_udp_multi_net is a nagios plugin which does remote checks by raw udp.
 * It queries the remote host with simple requests like i.e.
 *    "loadmulti\n"
 * and takes the last line returned as a result for nagios for the FIRST check/key specified.
 * Other keys specified on commandline will be searched for in the reply data and sent to
 * nagios as a passive check result.
 *
 * result data format must be:
 *   ...
 *    passivekey1 additional_information
 *   VALUE
 *    passivekey2 additional_information
 *   VALUE
 *    activekey
 *   additional_information
 *   VALUE
 *
 * compile with i.e.
 *   diet gcc -O2 -Wall -o check_remote_by_udp_multi_net check_remote_by_udp_multi_net.c;sstrip check_remote_by_udp_multi_net
 */

/*
#define NAGIOS_COMMAND_PIPE "/var/lib/nagios3/rw/nagios.cmd"
*/
#define NAGIOS_COMMAND_PIPE "/tmp/nagios.cmd"

#define TESTMULTIREQUESTS
#undef TESTMULTIREQUESTS

#undef NO_PORT_RANDOMIZATION
#define NO_PORT_RANDOMIZATION

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <features.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <poll.h>
#include <setjmp.h>
#include <time.h>
#include <string.h>
#include <stdint.h>
#include <sysexits.h>
#include <sys/epoll.h>
#include <sched.h>

#ifndef _USE_DEPRECATED_STR2UINT63
#define _USE_DEPRECATED_STR2UINT63
#endif
#include "lib/libfwb.h"

/* printf for debug */
#include <stdio.h>

#define NAGIOSRESULTPREFIXSTRING "UDPMULTI "

#define STR_URL1 " (<a href=\"http://"
#define STR_URL2 "\">"
#define STR_URL3 "</a>)"

#ifndef CLONE_IO
#define CLONE_IO 0x80000000
#endif
#ifndef CLONE_SYSVSEM
#define CLONE_SYSVSEM 0x00040000
#endif

/**
 * we use only one udp packet
 * UDPREADBYTES must be as big as or bigger than
 * UDP_BUFSIZE of knoerre(d)
 */
#define UDPREADBYTES 2047

/* up to MAX_KEYS per host will be recognized and searched for in result data */
#define MAX_KEYS 80

/* must be big enough for MAX_KEYS lines "warnlevel critlevel servicename key [unit]" if keysfile was given */
/* it may be bigger as UDPREADBYTES if the contents of the optional MULTI_RESULTS_INCLUDE_FILE will vary */
#define KEYSFILEBUFSIZE 4095

/* up to MAX_HOSTS will be checked in "parallel" if multihostfile was given */
#define MAX_HOSTS 1000

/* must be big enough for MAX_HOSTS lines "ip-address [shorthostname]" */
#define HOSTSFILEBUFSIZE 32767

#define MAX_THREADS 512

#undef STACKSIZE
#define STACKSIZE 512

char child_stack[MAX_THREADS][STACKSIZE];
int child_id[MAX_THREADS];

long warn[MAX_KEYS],crit[MAX_KEYS];
char *servicename[MAX_KEYS];
char *unit[MAX_KEYS];
const char *key[MAX_KEYS];
int key_count;
struct child_data {
  char *mh_ip_address[MAX_HOSTS];
  char *mh_shorthostname[MAX_HOSTS];
} st_child;

static const char version_string[] = "version_string: $Id: check_remote_by_udp_multi_net.c 691 2017-06-04 23:12:04Z fwbergm $";

const char usage[] =
  "usage:\n"
  "check_remote_by_udp_multi_net -I ip-address -H shorthostname -p port [-L] [-O] [-t timeout] [-D] -w warnlevel -c critlevel [-s servicename] [-u UNIT] -k key [-w warnlevel -c critlevel [-s servicename] [-u UNIT] -k key] ... multiplekey\n"
  "check_remote_by_udp_multi_net -i multihostfile -p port [-L] [-O] [-t timeout] [-D] -f keysfile -w warnlevel -c critlevel [-s servicename] [-u UNIT] multiplekey\n"
  "check_remote_by_udp_multi_net -I ip-address -H shorthostname -p port [-L] [-O] [-t timeout] [-D] -w warnlevel -c critlevel [-s servicename] [-u UNIT] -k key [-w warnlevel -c critlevel [-s servicename] [-u UNIT] -k key] ... multiplekey\n"
  "example:\n"
  "check_remote_by_udp_multi_net -i hosts-to-check-list-file -H localhost -p 8888 -w 7 -c 20 -k load1 -f passive-checks-list-file loadmulti\n"
  "check_remote_by_udp_multi_net -I 127.0.0.1 -H localhost -p 8888 -w 7 -c 20 -k load1   -w 80 -c 90 -s dfroot -k diskusage/ -w 10 -c 20 -k load5   loadmulti\n";

/* don't care which thread gets the signal, it doesn't matter */
static jmp_buf check_remote_procstate;
static void check_remote_timeout_alarm_sighandler(int num) { longjmp(check_remote_procstate,1); }

static void print(const char *str) { write1(str,strlen(str)); }

static void printusage() { print(usage); exit(1); }



/* the thread routine which does actually the remote check */
#ifdef __dietlibc__
void *run(void *notused) {
#else
int run(void *notused) {
#endif
  register char *p;
  int tmpint;
  ssize_t bytecount;
  int fd;
  struct sockaddr_in servin;
  struct sockaddr_in clntin;
  time_t checktime;
  char udp_buf[UDPREADBYTES+2];
  char linebuf[UDPREADBYTES+3];
  char pipebuf[UDPREADBYTES+1];
  char *pipe_cursor;
  uint64_t value;
  int retvalue;
  int port;
  const char *ip_address;
  const char *shorthostname;
  const char *portstr;
  const char *keymulti;
  long warn[MAX_KEYS],crit[MAX_KEYS];
  char *servicename[MAX_KEYS];
  char *unit[MAX_KEYS];
  const char *key[MAX_KEYS];
  int show_http_link;
  int debug;
  char *lastline;
  char *lastbeforeline;
  char *retvalpos;
  int no_ok_passivechecks;
  struct epoll_event ev, events[1];
  int efd;

  write1("c\n",2);
  for(tmpint=0;tmpint<20;++tmpint) {
    usleep(1);
    printf("%d %d p%p\n", *((int *) notused), tmpint, &efd);
    usleep(1);
  }
  printf("stackrun %p %p\n", &tmpint, &efd);
  _exit(0);

  /* setup sockaddr */
  memset(&servin, 0, sizeof servin);
  servin.sin_family = AF_INET;
  servin.sin_port = htons(port);
  if (inet_aton(ip_address, &servin.sin_addr) == 0) _exit(4);

  /* socket */
  fd = tmpint = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd == -1) {
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on opening socket\n");
    _exit(5);
  }

  if (fd) {
    close(fd);
    _exit(6);
  }

  memset(&clntin, 0, sizeof clntin);
  clntin.sin_family = AF_INET;
#ifndef NO_PORT_RANDOMIZATION
  /* dirty port randomization */
  mrand = getpid() % 32766;
  ++mrand;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  /* obsoleted in POSIX.1-2008: */
  if (!gettimeofday(&tv, 0)) {
    mrand = tv.tv_sec % mrand;
    mrand = (tv.tv_usec % 131075) * mrand;
  } else {
    mrand = time(0) % mrand;
  }
  mrand %= (65535 - 1024);
  mrand += 1024;
  clntin.sin_port = htons(mrand);
  tmpint = bind(fd, (struct sockaddr *) &clntin, sizeof clntin);
  if (tmpint) {
#endif
    /* let os choose port */
    clntin.sin_port = 0;
    tmpint = bind(fd, (struct sockaddr *) &clntin, sizeof clntin);
    if (tmpint) {
      close(fd);
      print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on bind\n");
      _exit(7);
    }
#ifndef NO_PORT_RANDOMIZATION
  }
#endif

  tmpint = connect(fd, (struct sockaddr *) &servin, sizeof servin);
  if (tmpint == -1) {
    close(fd);
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on connect\n");
    _exit(8);
  }

  /* send */
  checktime = time(0);
  p = stpcpy(udp_buf, keymulti);
  *p = '/'; ++p;
  p = sputuint32(p, checktime);
  *p = '/'; ++p;
  p = stpcpy(p, shorthostname);
  *p = '\n'; ++p;
  tmpint = send(fd, udp_buf, p - udp_buf, 0);

  if (tmpint == -1) {
    close(fd);
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on write\n");
    _exit(9);
  }

  /* recv */
/*
  st_poll.fd = fd;
  st_poll.events = POLLIN;
  tmpint = poll(&st_poll, 1, -1);
*/
  efd = epoll_create(1);
  ev.events = EPOLLIN;
  ev.data.fd = fd;
  if (epoll_ctl(efd, EPOLL_CTL_ADD, fd, &ev) == -1) {
    exit(EXIT_FAILURE);
  }
  if (epoll_wait(efd, events, 1, 2000) < 1) {
    close(efd);
    _exit(21);
  }

  /* for searching key strings set first byte to '\n' */
  *udp_buf = '\n';
  bytecount = recv(fd, udp_buf+1, sizeof udp_buf - 1, 0);
  close(fd);
  if (bytecount < 1) {
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on read\n");
    _exit(10);
  }
  ++bytecount; /* adjust for starting '\n' */
  udp_buf[bytecount] = 0;


  /* now we got the result(s), step through all "passive" keys and submit data */

  fd = open(NAGIOS_COMMAND_PIPE, O_WRONLY);
  if (fd == -1) goto passive_checks_done;

  for (tmpint=1; tmpint < key_count; ++tmpint) {

    /* create a search string "\n KEY" in linebuf to search for in udp_buf */
    p = stpcpy(linebuf, "\n ");
    p = stpcpy(p, key[tmpint]);
    /* old version: key string terminated the line */
    /* *p = '\n'; ++p; *p = 0; */

    p = udp_buf;

    search_next:
    p = strstr(p, linebuf);
    if (!p) continue;

    p += strlen(linebuf);
    /* SPACE or LF terminates key string */
    if ((*p != ' ') && (*p != '\n')) goto search_next;
    if (*p == ' ') ++p;
    if (*p == '\n') {
      lastbeforeline = 0;
      ++p;
    } else {
      lastbeforeline = p;
      for (;;) { if (!*p) break; if (*p == '\n') { ++p; break; } ++p; }
    }
    lastline = p;

    p = pipe_cursor;
    *p = '['; ++p;
    p = sputuint32(p, checktime);
    p = stpcpy(p, "] PROCESS_SERVICE_CHECK_RESULT;");
    p = stpcpy(p, shorthostname);
    *p = ';'; ++p;
    p = stpcpy(p, servicename[tmpint]);
    retvalpos = p + 1;
    p = stpcpy(p, ";0;" NAGIOSRESULTPREFIXSTRING);

    value = 0;
    if (str2uint64(lastline, &value)) {
      if (value > warn[tmpint]) {
        if (value > crit[tmpint]) {
          p = stpcpy(p, "CRITICAL - ");
          retvalue = 2;
        } else {
          p = stpcpy(p, "WARNING - ");
          retvalue = 1;
        }
      } else {
        if (no_ok_passivechecks) continue; /* should be made before filling pipe_buf */
        p = stpcpy(p, "OK - ");
        retvalue = 0;
      }
    } else {
      p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING "UNKNOWN - ");
      retvalue=3;
    }
    *retvalpos = '0' + retvalue;

    for (;;) { *p = *lastline; ++p; ++lastline; if ((!*lastline) || (*lastline == '\n')) break; }
    if (unit[tmpint] != NULL) {
      p = stpcpy(p, unit[tmpint]);
    }

    /* append more data if retvalue not "OK" */
    if (retvalue > 0) {

      if (show_http_link) {
        /* construct link */
        p = stpcpy(p, STR_URL1);
        p = stpcpy(p, ip_address);
        *p = ':'; ++p;
        p = stpcpy(p, portstr);
        *p = '/'; ++p;
        p = stpcpy(p, key[tmpint]);
        p = stpcpy(p, STR_URL2);
        if (lastbeforeline) {
          for (;;) {
            *p = *lastbeforeline; ++p; ++lastbeforeline; if ((!*lastbeforeline) || (*lastbeforeline == '\n')) break;
          }
        } else {
          p = stpcpy(p, key[tmpint]);
        }
        p = stpcpy(p, STR_URL3);
      } else {
        if (lastbeforeline) {
          p = stpcpy(p, " (");
          for (;;) {
            *p = *lastbeforeline; ++p; ++lastbeforeline; if ((!*lastbeforeline) || (*lastbeforeline == '\n')) break;
          }
          *p = ')'; ++p;
        }
      }

    }

    *p = '\n'; ++p;
    /* "[$(date +%s)] PROCESS_SERVICE_CHECK_RESULT;mon2;rsbackup;0;UDPMULTI OK - 666 minutes_ago blablabla" */
    pipe_cursor = p;

  }
  if (debug) write1(pipebuf, pipe_cursor - pipebuf);
  write(fd, pipebuf, pipe_cursor - pipebuf);
  close(fd);

  passive_checks_done:
  /* passive checks done */

  _exit(retvalue);

  return 0;
}




int main(int argc,char *const argv[]) {
/*
  register char *cptr;
  int tmpint;
  ssize_t bytecount;
  uint64_t value;
  int mode1, mode2;
*/
  register char *p;
  char *stack;
  uint32_t flags;
  int tmpint;
  ssize_t bytecount;
  int fd;
  char pipebuf[UDPREADBYTES+1];
  char *pipe_cursor;
  uint64_t value;
  int retvalue;
  int port;
  const char *ip_address;
  const char *shorthostname;
  const char *portstr;
  const char *keymulti;
  int warni,criti;
  int timeout;
  int host_count;
  int show_http_link;
  int debug;
  int no_ok_passivechecks;
  char keysfilebuf[KEYSFILEBUFSIZE+1];
  char hostsfilebuf[HOSTSFILEBUFSIZE+1];
  char *mh_ip_address[MAX_HOSTS];
  char *mh_shorthostname[MAX_HOSTS];
  struct sockaddr_in servin;
  struct sockaddr_in clntin;
#ifndef NO_PORT_RANDOMIZATION
  struct timeval tv;
  uint32_t mrand;
#endif
#ifdef TESTMULTIREQUESTS
  int testmultirequests;

  for(testmultirequests=0;testmultirequests<20000;testmultirequests++) {
#endif

  no_ok_passivechecks = 0;
  unit[0] = NULL;
  warn[0] = warni = crit[0] = criti = port = timeout =
    key_count = host_count = show_http_link = debug = 0;
  ip_address = argv[0];
  portstr = "80";
  retvalue = 2;
  keymulti = shorthostname = ip_address;
  pipe_cursor = pipebuf;
  *keysfilebuf = *hostsfilebuf = 0;

  if (argc < 2) printusage();
  tmpint = 0;
  for (;tmpint < MAX_KEYS; ++tmpint) {
    warn[tmpint] = crit[tmpint] = 0;
    servicename[tmpint] = unit[tmpint] = 0;
    key[tmpint] = 0;
  }
  for(tmpint=1; tmpint<argc; ++tmpint) {
    if (*argv[tmpint] == '-') {
      if (argv[tmpint][2] != '\0') printusage();
      switch (argv[tmpint][1]) {
        case 'h': printusage(); break;
        case 't': ++tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); timeout=(long) value; break;
        case 'I': ++tmpint; ip_address=argv[tmpint]; break;
        case 'i': ++tmpint;
                  fd = open(argv[tmpint], O_RDONLY);
                  if (fd < 0) _exit(EX_NOINPUT);
                  bytecount = read(fd, hostsfilebuf, HOSTSFILEBUFSIZE);
                  if (bytecount < 9) _exit(EX_NOINPUT);
                  hostsfilebuf[bytecount] = 0;
                  close(fd);
                  p = hostsfilebuf;
                  readhosts:
                  mh_ip_address[host_count] = p; for(;;) { ++p; if ((*p == ' ') || (*p == '\n') || (!*p)) break; }
                  if (*p == ' ') {
                    *p = 0; ++p; mh_shorthostname[host_count] = p;
                    for(;;) { ++p; if ((*p == '\n') || (*p == ' ') || (!*p)) break; }
                  }
                  if (*p == '\n') { *p = 0;
/*
*/
#include <stdio.h>
if (mh_shorthostname[host_count])
printf("i=%s h=%s\n", mh_ip_address[host_count], mh_shorthostname[host_count]);
else
printf("i=%s\n", mh_ip_address[host_count]);
 ++p; }
                  ++host_count; if (host_count >= MAX_HOSTS) --host_count;
                  if (*p) goto readhosts;
                  break;
        case 'L': show_http_link = 1; break;
        case 'D': debug = 1; break;
        case 'O': no_ok_passivechecks = 1; break;
        case 'H': ++tmpint; shorthostname=argv[tmpint]; break;
        case 'p': ++tmpint; portstr = argv[tmpint]; if(!str2uint64(argv[tmpint], &value)) printusage(); port=(int) value; break;
        case 'f': ++tmpint;
                  fd = open(argv[tmpint], O_RDONLY);
                  if (fd < 0) _exit(EX_NOINPUT);
                  bytecount = read(fd, keysfilebuf, KEYSFILEBUFSIZE);
                  if (bytecount < 9) _exit(EX_NOINPUT);
                  keysfilebuf[bytecount] = 0;
                  close(fd);
                  p = keysfilebuf;
                  readkeys:
                  p = str2uint64(p, &value); if(!p) printusage(); warn[key_count]=(long) value; ++p;
                  p = str2uint64(p, &value); if(!p) printusage(); crit[key_count]=(long) value; ++p;
                  servicename[key_count] = p; for(;;) { ++p; if (*p == ' ') break; if ((*p == '\n') || (!*p)) printusage(); } *p = 0; ++p;
                  if ((!*p) || (*p == '\n')) printusage();
                  key[key_count] = p; for(;;) { ++p; if ((*p == '\n') || (*p == ' ') || (!*p)) break; }
                  if ((p - key[key_count]) >= UDPREADBYTES) { print("key too large\n"); _exit(2); }
                  if (*p == ' ') {
                    *p = 0; ++p; unit[key_count] = p;
                    for(;;) { ++p; if ((*p == '\n') || (*p == ' ') || (!*p)) break; }
                  }
                  if (*p == '\n') { *p = 0; ++p; }
/*
#include <stdio.h>
if (unit[key_count])
printf("w=%ld c=%ld s=%s k=%s u=%s\n", warn[key_count], crit[key_count], servicename[key_count], key[key_count], unit[key_count]);
else
printf("w=%ld c=%ld s=%s k=%s\n", warn[key_count], crit[key_count], servicename[key_count], key[key_count]);
*/
                  ++key_count; if (key_count >= MAX_KEYS) --key_count;
                  if (*p) goto readkeys;
                  break;
        case 'w': ++tmpint; if (!warni) warni=tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); warn[key_count]=(long) value; break;
        case 'c': ++tmpint; if (!criti) criti=tmpint; if(!str2uint64(argv[tmpint], &value)) printusage(); crit[key_count]=(long) value; break;
        case 's': ++tmpint; servicename[key_count]=argv[tmpint]; break;
        case 'u': ++tmpint; unit[key_count]=argv[tmpint]; break;
        case 'k': ++tmpint; key[key_count]=argv[tmpint];
                  if (!servicename[key_count]) servicename[key_count] = argv[tmpint];
                  if (strlen(key[key_count]) >= UDPREADBYTES) { print("key too large\n"); _exit(2); }
                  ++key_count; if (key_count >= MAX_KEYS) --key_count;
                  break;
        default: break;
      }
    } else {
      keymulti=argv[tmpint];
      if (strlen(keymulti) >= UDPREADBYTES) { print("key too large\n"); _exit(2); }
    }
  }
  if (timeout == 0) timeout = 10;
  if (ip_address == argv[0]) printusage();
  if (shorthostname == argv[0]) printusage();
  if (keymulti == argv[0]) printusage();
  if (key_count == 0) printusage();
  if (warni == 0) printusage();
  if (criti == 0) printusage();
  tmpint = 0;
  for (;tmpint < key_count; ++tmpint) {
    if (tmpint && !servicename[tmpint]) printusage();
    if (!key[tmpint]) printusage();
  }

  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - plugin internal timeout\n");
    _exit(3);
  }

/*
knoerre/knoerred.c-  child_stack = stack + STACKSIZE;
knoerre/knoerred.c-  flags =
knoerre/knoerred.c:    CLONE_FILES |
knoerre/knoerred.c:    CLONE_FS |
knoerre/knoerred.c:    CLONE_IO |
knoerre/knoerred.c:    CLONE_SIGHAND |
knoerre/knoerred.c:    CLONE_SYSVSEM |
knoerre/knoerred.c:    CLONE_THREAD |
knoerre/knoerred.c:    CLONE_VM;
knoerre/knoerred.c-  clone(interval_measure, child_stack, flags, 0);
CLONE_PTRACE
*/

  stack = child_stack[0] + STACKSIZE;
  flags =
    CLONE_FILES |
    CLONE_FS |
    CLONE_IO |
    CLONE_SIGHAND |
    CLONE_SYSVSEM |
    CLONE_THREAD |
    CLONE_VM;
  printf("stack1 %p %p\n", child_stack[0], stack);
  clone(run, stack, flags, &(child_stack[0]));
  stack = child_stack[1] + STACKSIZE;
  printf("stack2 %p %p\n", child_stack[1], stack);
  clone(run, stack, flags, &(child_stack[0]));
  /* NOTE: dietlibc defines 'void *(*)(void *)' but glibc defines 'int (*)(void *)' as first arg */

  for(tmpint=0;tmpint<20;++tmpint) {
    usleep(1);
    printf("%d m %d\n", getpid(), tmpint);
    usleep(1);
  }
  _exit(0);

  /* setup sockaddr */
  memset(&servin, 0, sizeof servin);
  servin.sin_family = AF_INET;
  servin.sin_port = htons(port);
  if (inet_aton(ip_address, &servin.sin_addr) == 0) _exit(4);

  /* socket */
  fd = tmpint = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd == -1) {
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on opening socket\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(5);
#endif
  }

  if (fd) {
    close(fd);
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(6);
#endif
  }

  memset(&clntin, 0, sizeof clntin);
  clntin.sin_family = AF_INET;
#ifndef NO_PORT_RANDOMIZATION
  /* dirty port randomization */
  mrand = getpid() % 32766;
  ++mrand;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  /* obsoleted in POSIX.1-2008: */
  if (!gettimeofday(&tv, 0)) {
    mrand = tv.tv_sec % mrand;
    mrand = (tv.tv_usec % 131075) * mrand;
  } else {
    mrand = time(0) % mrand;
  }
  mrand %= (65535 - 1024);
  mrand += 1024;
  clntin.sin_port = htons(mrand);
  tmpint = bind(fd, (struct sockaddr *) &clntin, sizeof clntin);
  if (tmpint) {
#endif
    /* let os choose port */
    clntin.sin_port = 0;
    tmpint = bind(fd, (struct sockaddr *) &clntin, sizeof clntin);
    if (tmpint) {
      close(fd);
      print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on bind\n");
#ifdef TESTMULTIREQUESTS
      continue;
#else
      _exit(7);
#endif
    }
#ifndef NO_PORT_RANDOMIZATION
  }
#endif

  tmpint = connect(fd, (struct sockaddr *) &servin, sizeof servin);
  if (tmpint == -1) {
    close(fd);
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on connect\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(8);
#endif
  }

  /* send */
  checktime = time(0);
  p = stpcpy(udp_buf, keymulti);
  *p = '/'; ++p;
  p = sputuint32(p, checktime);
  *p = '/'; ++p;
  p = stpcpy(p, shorthostname);
  *p = '\n'; ++p;
  tmpint = send(fd, udp_buf, p - udp_buf, 0);

  if (tmpint == -1) {
    close(fd);
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on write\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(9);
#endif
  }

  /* recv */
  efd = epoll_create(1);
  ev.events = EPOLLIN;
  ev.data.fd = fd;
  if (epoll_ctl(efd, EPOLL_CTL_ADD, fd, &ev) == -1) {
    exit(EXIT_FAILURE);
  }
  if (epoll_wait(efd, events, 1, 2000) < 1) {
    close(efd);
    _exit(21);
  }

  /* for searching key strings set first byte to '\n' */
  *udp_buf = '\n';
  bytecount = recv(fd, udp_buf+1, sizeof udp_buf - 1, 0);
  close(fd);
  if (bytecount < 1) {
    print(NAGIOSRESULTPREFIXSTRING "UNKNOWN - error on read\n");
#ifdef TESTMULTIREQUESTS
    continue;
#else
    _exit(10);
#endif
  }
  ++bytecount; /* adjust for starting '\n' */
  udp_buf[bytecount] = 0;


  /* now we got the result(s), step through all "passive" keys and submit data */

  fd = open(NAGIOS_COMMAND_PIPE, O_WRONLY);
  if (fd == -1) goto passive_checks_done;

  for (tmpint=1; tmpint < key_count; ++tmpint) {

    /* create a search string "\n KEY" in linebuf to search for in udp_buf */
    p = stpcpy(linebuf, "\n ");
    p = stpcpy(p, key[tmpint]);
    /* old version: key string terminated the line */
    /* *p = '\n'; ++p; *p = 0; */

    p = udp_buf;

    search_next:
    p = strstr(p, linebuf);
    if (!p) continue;

    p += strlen(linebuf);
    /* SPACE or LF terminates key string */
    if ((*p != ' ') && (*p != '\n')) goto search_next;
    if (*p == ' ') ++p;
    if (*p == '\n') {
      lastbeforeline = 0;
      ++p;
    } else {
      lastbeforeline = p;
      for (;;) { if (!*p) break; if (*p == '\n') { ++p; break; } ++p; }
    }
    lastline = p;

    p = pipe_cursor;
    *p = '['; ++p;
    p = sputuint32(p, checktime);
    p = stpcpy(p, "] PROCESS_SERVICE_CHECK_RESULT;");
    p = stpcpy(p, shorthostname);
    *p = ';'; ++p;
    p = stpcpy(p, servicename[tmpint]);
    retvalpos = p + 1;
    p = stpcpy(p, ";0;" NAGIOSRESULTPREFIXSTRING);

    value = 0;
    if (str2uint64(lastline, &value)) {
      if (value > warn[tmpint]) {
        if (value > crit[tmpint]) {
          p = stpcpy(p, "CRITICAL - ");
          retvalue = 2;
        } else {
          p = stpcpy(p, "WARNING - ");
          retvalue = 1;
        }
      } else {
        if (no_ok_passivechecks) continue; /* should be made before filling pipe_buf */
        p = stpcpy(p, "OK - ");
        retvalue = 0;
      }
    } else {
      p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING "UNKNOWN - ");
      retvalue=3;
    }
    *retvalpos = '0' + retvalue;

    for (;;) { *p = *lastline; ++p; ++lastline; if ((!*lastline) || (*lastline == '\n')) break; }
    if (unit[tmpint] != NULL) {
      p = stpcpy(p, unit[tmpint]);
    }

    /* append more data if retvalue not "OK" */
    if (retvalue > 0) {

      if (show_http_link) {
        /* construct link */
        p = stpcpy(p, STR_URL1);
        p = stpcpy(p, ip_address);
        *p = ':'; ++p;
        p = stpcpy(p, portstr);
        *p = '/'; ++p;
        p = stpcpy(p, key[tmpint]);
        p = stpcpy(p, STR_URL2);
        if (lastbeforeline) {
          for (;;) {
            *p = *lastbeforeline; ++p; ++lastbeforeline; if ((!*lastbeforeline) || (*lastbeforeline == '\n')) break;
          }
        } else {
          p = stpcpy(p, key[tmpint]);
        }
        p = stpcpy(p, STR_URL3);
      } else {
        if (lastbeforeline) {
          p = stpcpy(p, " (");
          for (;;) {
            *p = *lastbeforeline; ++p; ++lastbeforeline; if ((!*lastbeforeline) || (*lastbeforeline == '\n')) break;
          }
          *p = ')'; ++p;
        }
      }

    }

    *p = '\n'; ++p;
    /* "[$(date +%s)] PROCESS_SERVICE_CHECK_RESULT;mon2;rsbackup;0;UDPMULTI OK - 666 minutes_ago blablabla" */
    pipe_cursor = p;

  }
  if (debug) write1(pipebuf, pipe_cursor - pipebuf);
  write(fd, pipebuf, pipe_cursor - pipebuf);
  close(fd);

  passive_checks_done:
  /* passive checks done, last check is active check */

  p = stpcpy(linebuf, "\n ");
  p = stpcpy(p, key[0]);
  *p = '\n'; ++p; *p = 0;

  p = strstr(udp_buf, linebuf);
  if (!p) {
/*
    keynotfound:
*/
    stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING "UNKNOWN - key not found\n");
    retvalue = 3;
    goto print_and_exit;
  }
  p += strlen(linebuf);
  lastbeforeline = p;
  for (;;) {
    ++p; if (!*p) break; if (*p == '\n') { ++p; break; }
  }
  lastline = p;
  for (;;) {
    ++p; if (!*p) break; if (*p == '.') *p = '\n'; if (*p == '\n') break;
  }
  *p = 0;
  p = lastline;

  /* construct output line */
  value = 0;
  if (str2uint64(p, &value)) {
    p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING);
    if (value > warn[0]) {
      if (value > crit[0]) {
        p = stpcpy(p, "CRITICAL - ");
        retvalue = 2;
      } else {
        p = stpcpy(p, "WARNING - ");
        retvalue = 1;
      }
    } else {
      p = stpcpy(p, "OK - ");
      retvalue = 0;
    }
  } else {
    p = stpcpy(linebuf, NAGIOSRESULTPREFIXSTRING "UNKNOWN - ");
    retvalue=3;
  }
  p = stpcpy(p, lastline);
  if (unit[0] != NULL) {
    p = stpcpy(p, unit[0]);
  }
  /* append more data if retvalue not "OK" */
  if (retvalue > 0) {
    /* get line before last line */
    for (tmpint = lastline - udp_buf;;) {
      if (!tmpint) break;
      --tmpint;
      if ((unsigned char) udp_buf[tmpint] >= 32) break;
    }
    if (lastline != (udp_buf+tmpint)) {
      udp_buf[tmpint + 1] = 0;
      for (;;) {
        --tmpint;
        if (tmpint < 0) break;
        if ((unsigned char) udp_buf[tmpint] < 32) break;
      }
      /* skip blanks and ctrl chars */
      for (;;) {
        ++tmpint;
        if ((unsigned char) udp_buf[tmpint] > 32) break;
      }
    }
    if (show_http_link) {
      /* construct link */
      p = stpcpy(p, STR_URL1);
      p = stpcpy(p, ip_address);
      *p = ':'; ++p;
      p = stpcpy(p, portstr);
      *p = '/'; ++p;
      p = stpcpy(p, keymulti);
      p = stpcpy(p, STR_URL2);
      p = stpcpy(p, udp_buf + tmpint);
      p = stpcpy(p, STR_URL3);
    } else {
      p = stpcpy(p, " (");
      p = stpcpy(p, udp_buf + tmpint);
      *p = ')'; ++p;
    }
  }
  /* performance data */
  p = stpcpy(p, " | ");
  p = stpcpy(p, key[0]);
  *p = '='; ++p;
  p = stpcpy(p, lastline);
  if (unit[0] != NULL) {
    while (*unit[0] == ' ') ++unit[0];
    for (tmpint = 0;; ++tmpint)
      if (unit[0][tmpint] <= 32) break;
    unit[0][tmpint] = 0;
    p = stpcpy(p, unit[0]);
  }
  *p = ';'; ++p;
  p = stpcpy(p, main_argv[warni]);
  *p = ';'; ++p;
  p = stpcpy(p, main_argv[criti]);
  *p = '\n'; ++p;
  *p = 0;
  print_and_exit:
  print(linebuf);
#ifdef TESTMULTIREQUESTS
  }
#endif
  _exit(retvalue);
  write2("error\n", 6);
  _exit(tmpint);
}
