/**
 * $Id: check_tcp_connect.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * check_tcp_connect is a nagios plugin
 *
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#define MAXRETRIES 6

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/times.h>
#include <arpa/inet.h>
#include <setjmp.h>
#include <stdint.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>

#include "lib/libfwb.h"

#define BUFSIZE 4096

#define NAGIOSRESULTPREFIXSTRING "TCP "
#define NAGIOSRESULTOKSTRING "OK "
#define NAGIOSRESULTWARNSTRING "WARNING "
#define NAGIOSRESULTCRITSTRING "CRITICAL "
#define NAGIOSRESULTUNKNOWNSTRING "UNKNOWN "

static const char version_string[] = "version_string: $Id: check_tcp_connect.c 691 2017-06-04 23:12:04Z fwbergm $";

static const char usage[] = 
  "usage: check_tcp_connect [-T(O|W|C)] -I ip-address -w warnlevel -c critlevel [-p port] [-P resultprefix] [-t timeout] [-D]\n"
  "example: check_tcp_connect -I 127.0.0.1 -w 2000 -c 4000 -p 22\n";
static void printusage() { write1(usage, sizeof usage - 1); exit(1); }

enum { OK, WARNING, CRITICAL, UNKNOWN };

static jmp_buf check_remote_procstate;

static void check_remote_timeout_alarm_sighandler(int num) {
  longjmp(check_remote_procstate,1);
}

static uint32_t len(const char *str) {
  register uint32_t i;

  if (!str) return 0;
  if (!*str) return 0;
  for (i = 1;;) { if (!str[i]) break; ++i; }
  return i;
}

static void print(const char *str) {
  if (!str) return;
  if (!*str) return;
  write1(str, len(str));
}

static char *sputint(register char *buf, uint32_t num, int leadzeroes) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;

  i=num; pbuf=buf;
  o = 0 - leadzeroes;
  if(o) {
    for(d=1;leadzeroes>0;leadzeroes--)d*=10;
  } else {
    d=1000000000;
  }
  for(;d>1;d/=10) {
    c=i/d;
    if((c>0)||(o)) {
      *pbuf++=c+'0'; i=i-(c*d);
      if(c>0)o=1;
      else ++o;
    }
  }
  *pbuf++=i+'0'; *pbuf='\0';
  return pbuf;
}


int main(int argc,char *const argv[]) {
  register char *p;
  struct sockaddr_in server_sockaddr;
  int sockfd;
  char linebuf[BUFSIZE+1];
  char timebuf[BUFSIZE+1];
  char *r;
  int retvalue;
  int c;
  uint64_t warn,crit;
  int port;
  const char *ip_address;
  const char *portstr;
  int tcp_defer_accept;
  int timeout;
  int timeout_state;
  int64_t runtime;
  int runtime_negativeflag;
  struct timespec rt_tspec;
  uint64_t val64;
  char *resultprefix;


  rt_tspec.tv_sec = 0;
  rt_tspec.tv_nsec = 0;
  runtime = -1;
  runtime_negativeflag = 0;
  port = tcp_defer_accept = timeout = 0;
  warn = crit = 0;
  ip_address = (const char *) 0;
  portstr = "80";
  retvalue = 2;
  timeout_state = UNKNOWN;
  resultprefix = NAGIOSRESULTPREFIXSTRING;

  if (argc < 2) { print("missing arguments - help with '-h'\n"); _exit(UNKNOWN); }
  for(c=1; c<argc; c++) {
    if (*argv[c] == '-') {
      switch (argv[c][1]) {
        case 'h': printusage(); break;
        case 'D': tcp_defer_accept = 1; break;
        case 'w': c++; str2uint64(argv[c],&val64);warn=val64; break;
        case 'c': c++; str2uint64(argv[c],&val64);crit=val64; break;
        case 'I': c++; ip_address = argv[c]; break;
        case 'p': c++; portstr = argv[c]; str2uint64((char*)portstr,&val64);port=(int)val64; break;
        case 'P': c++; resultprefix = argv[c]; break;
        case 't': c++; str2uint64(argv[c],&val64);timeout=val64; break;
        case 'T': switch (argv[c][2]) {
                    case 'O': timeout_state = OK; break;
                    case 'W': timeout_state = WARNING; break;
                    case 'C': timeout_state = CRITICAL; break;
                    default: break;
                  } break;
        default: break;
      }
    } else {
      printusage();
    }
  }
  if (ip_address == 0) print(usage);
  if (warn == 0) warn = 5;
  if (crit == 0) crit = 10;
  if (timeout == 0) timeout = 10;
  if (port == 0) port = 80;
  if (len(resultprefix) > 255) resultprefix[255] = 0;
  warn *= 1000;
  crit *= 1000;

  p = stpcpy(linebuf, resultprefix);
  *p++ = ' ';

  /* set timeout */
  signal(SIGALRM,check_remote_timeout_alarm_sighandler);
  alarm(timeout);
  if (setjmp(check_remote_procstate)) {
    switch (timeout_state) {
      case OK: p = stpcpy(p, NAGIOSRESULTOKSTRING "- plugin internal timeout\n"); break;
      case WARNING: p = stpcpy(p, NAGIOSRESULTWARNSTRING "- plugin internal timeout\n"); break;
      case CRITICAL: p = stpcpy(p, NAGIOSRESULTCRITSTRING "- plugin internal timeout\n"); break;
      default: p = stpcpy(p, NAGIOSRESULTUNKNOWNSTRING "- plugin internal timeout\n"); break;
    }
    write1(linebuf, p - linebuf);
    _exit(timeout_state);
  }

  /* send request */
  if (inet_aton(ip_address, &server_sockaddr.sin_addr) == 0) {
    p = stpcpy(p, NAGIOSRESULTUNKNOWNSTRING "- error on inet_aton\n");
    write1(linebuf, p - linebuf);
    _exit(UNKNOWN);
  }
  server_sockaddr.sin_family = AF_INET;
  server_sockaddr.sin_port = htons(port);

  /* obsoleted in POSIX.1-2008:  if (!gettimeofday(&rt_tval, &rt_tz)) ... */
  if (!clock_gettime(CLOCK_REALTIME, &rt_tspec)) runtime = (1000000 * rt_tspec.tv_sec) + (rt_tspec.tv_nsec / 1000);

  if ((sockfd=socket(PF_INET,SOCK_STREAM,0)) == -1) {
    p = stpcpy(p, NAGIOSRESULTUNKNOWNSTRING "- error on opening socket\n");
    write1(linebuf, p - linebuf);
    _exit(UNKNOWN);
  }
  if (tcp_defer_accept) {
    if (setsockopt(sockfd,IPPROTO_TCP,TCP_DEFER_ACCEPT,&tcp_defer_accept,sizeof tcp_defer_accept) != 0) {
      close(sockfd);
      p = stpcpy(p, NAGIOSRESULTCRITSTRING "- error on setsockopt");
      write1(linebuf, p - linebuf);
      _exit(UNKNOWN);
    }
  }
  if (connect(sockfd,(struct sockaddr *)&server_sockaddr,sizeof(server_sockaddr)) == -1) {
    close(sockfd);
    p = stpcpy(p, NAGIOSRESULTCRITSTRING "- error on connect\n");
    write1(linebuf, p - linebuf);
    _exit(CRITICAL);
  }

  /*close(sockfd);*/
  /* obsoleted in POSIX.1-2008:  if (!gettimeofday(&rt_tval, &rt_tz)) ... */
  if (!clock_gettime(CLOCK_REALTIME, &rt_tspec)) runtime = ((1000000 * rt_tspec.tv_sec) + (rt_tspec.tv_nsec / 1000)) - runtime;

  retvalue = 0;

  /* slow response? */
  if (runtime < -1) {
    /* negative runtime: ntp time drift? */
    runtime_negativeflag = 1; /* most simple :-) */
    runtime = 0 - runtime;
  } else {
    /* only take care of runtime if there's no time drift */
    if ((retvalue < 1) && (runtime >= warn)) retvalue = 1;
    if ((retvalue < 2) && (runtime >= crit)) retvalue = 2;
  }

  /* create nagios reply string */
  switch(retvalue) {
    case 0: p = stpcpy(p, NAGIOSRESULTOKSTRING); break;
    case 1: p = stpcpy(p, NAGIOSRESULTWARNSTRING); break;
    case 2: p = stpcpy(p, NAGIOSRESULTCRITSTRING); break;
    default: retvalue = 3;
  }

  if (runtime != -1) {
    p = stpcpy(p, "- ");
    if (runtime_negativeflag) { *p = '-'; ++p; } /* ntp time drift? */
    r = sputint(timebuf, (uint32_t)(runtime/1000000), 0);
    *r++ = '.';
    sputint(r, (uint32_t)(runtime - (1000000 * (runtime/1000000))), 5);
    r = timebuf;
    p = stpcpy(p, r) - 3;
    p = stpcpy(p, " second response time |time=");
    p = stpcpy(p, r);
    p = stpcpy(p, "s\n");
  }
  write1(linebuf, p - linebuf);
  _exit(retvalue);
}
