/*
 * chownu.c
 * small chown with security restrictions to be run with CAP_CHOWN
 * $Id: chownu.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <dirent.h>
#include <stdint.h>
#include <time.h>
#include <errno.h>
#include <sys/mman.h>
#include <sysexits.h>

#include "config.h"
#include "../lib/libfwb.h"

#define USERNAME_SIZE 255
#define PATHNAME_SIZE 16384
#define DEFAULT_MAXDEPTH 50
#ifdef CONFIG_CHOWNU_MINUID
#define MIN_UID CONFIG_CHOWNU_MINUID
#else
#define MIN_UID 7101
#endif
#ifdef CONFIG_CHOWNU_MAXUID
#define MAX_UID CONFIG_CHOWNU_MAXUID
#else
#define MAX_UID 7199
#endif

#define STR1(s) #s
#define STR(s) STR1(s)

/*
config CHOWNU_MINUID
    int "chownu minimal UID to be allowed"
    default 500
    help
      This defines the minimal UID to be allowed in chownu.
      If the user specified on command line has an UID below MINUID then chownu won't run.
      If a file has an owner UID below MINUID then the file will be ignored.
      Default is 500.
config CHOWNU_MAXUID
    int "chownu maximal UID to be allowed"
    default 9999
    help
      This defines the maximal UID to be allowed in chownu.
      If the user specified on command line has an UID above MAXUID then chownu won't run.
      If a file has an owner UID above MAXUID then the file will be ignored.
      Default is 9999.
*/

static void printusage() {
  const char usage[] =
    "usage:\n"
    "  chownu [option(s)] user file [file(s)]\n"
    "options:\n"
    "  -R             directory recursive mode\n"
    "  -d max_depth   specify the maximum recursion depth (default: "STR(DEFAULT_MAXDEPTH)")\n"
    "change the owner of one or more files\n"
    "user id and file owners below "STR(MIN_UID)" and above "STR(MAX_UID)" are ignored\n";

  write1(usage, sizeof usage - 1);
  _exit(EX_USAGE);
}

enum { MODE_NORMAL, MODE_RECURSIVE };

static inline int lchown_void(const char *path, uid_t owner) {
#if defined __NR_lchown32
  int res = syscall(__NR_lchown32, path, owner, -1);
  if (res == -1 && errno == ENOSYS) { /* 38 */
    res = syscall(__NR_lchown, path, owner, -1);
  }
  return res;
#else
  return syscall(__NR_lchown, path, owner, -1);
#endif
}

static void chown_recursive(uid, path, depth, curdepth)
uint32_t uid;
const char *path;
int32_t depth;
int32_t curdepth;
{

  register char *p;
  DIR *dptr;
  struct dirent *eptr;
  struct stat st;
  char pathbuf[PATHNAME_SIZE*2+2];

  ++curdepth;
  if ((depth != 0) && (curdepth > depth)) return;
  dptr = opendir(path);
  if (dptr == NULL) return;
  for(;;) {
    eptr = readdir(dptr); if (eptr == NULL) break;
    p = eptr->d_name;
    if (*p == '.') {
      ++p;
      if (*p == 0) continue;
      if (*p == '.') {
        ++p;
        if (*p == 0) continue;
      }
    }
    p = stpcpy(pathbuf, path);
    *p = '/'; ++p;
    p = stpcpy(p, eptr->d_name);
    if ((p - pathbuf) > PATHNAME_SIZE) _exit(EX_DATAERR);
    p = pathbuf;
    if (lstat(p, &st) == 0){
      if ((st.st_uid != 0) && (st.st_uid >= MIN_UID) && (st.st_uid <= MAX_UID)) {
        lchown_void(p, uid);
      }
      if ((st.st_mode & S_IFMT) == S_IFDIR) {
        chown_recursive(uid, p, depth, curdepth);
      }
    }
  }
  closedir(dptr);
  return;
}

int main(int argc,char *const argv[]) {
  register int c;
  register char *p;
  uint32_t uid;
  const char *pwdbuf;
  off_t pwdlen;
  int s_len;
  int mode;
  int first_arg;
  struct stat st;
  uint32_t depth;

  mode = MODE_NORMAL;
  depth = DEFAULT_MAXDEPTH;
  first_arg = 0;
  for (c = 1; c < argc; ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2]) printusage();
      switch (argv[c][1]) {
        case 'h': printusage();
        case 'R': mode = MODE_RECURSIVE; break;
        case 'd': ++c;
                  if (c >= argc) printusage();
                  p = str2uint32(argv[c], &depth);
                  if (p == 0) printusage();
                  if (*p != 0) printusage();
                  break;
        default:  printusage();
      }
    } else {
      first_arg = c;
      break;
    }
  }
  if (first_arg == 0) printusage();
  if ((first_arg + 1) >= argc) printusage();

  /* open /etc/passwd for uid-mapping */
  p = (char *) map_passwd_file(&pwdbuf, &pwdlen);
  if (p == (char *) -1) _exit(EX_OSFILE);

  /* get UID of user */
  s_len = strlen(argv[first_arg]);
  p = (char *) get_uid_from_passwd(pwdbuf, pwdlen, argv[first_arg], s_len, &s_len);
  if (!p) { munmap((void *) pwdbuf, pwdlen); _exit(EX_NOUSER); }
  if (s_len > USERNAME_SIZE) { /* slightly paranoid */
    munmap((void *) pwdbuf, pwdlen);
    _exit(EX_DATAERR);
  }
  p = str2uint32(p, &uid);
  munmap((void *) pwdbuf, pwdlen);
  if (!p) _exit(EX_OSFILE);
  if ((uid == 0) || (uid < MIN_UID) || (uid > MAX_UID)) {
    _exit(EX_NOPERM);
  }

  /* raise stack for recursive operation */
  if (mode == MODE_RECURSIVE) {
    struct rlimit rl = { 134217728, 134217728 };
    setrlimit(RLIMIT_STACK, &rl);
  }

  /* loop through file args */
  for (c = first_arg+1; c < argc; ++c) {
    if (lstat(argv[c], &st) == 0) {
      if ((st.st_uid != 0) && (st.st_uid >= MIN_UID) && (st.st_uid <= MAX_UID)) {
        lchown_void(argv[c], uid);
      }
      if ((mode == MODE_RECURSIVE) && ((st.st_mode & S_IFMT) == S_IFDIR)) {
        chown_recursive(uid, argv[c], depth, 0);
      }
    }
  }

  _exit(EX_OK);
}
