/*
 * chrootuser.c
 * usage: chrootuser user dir command [options]
 * Frank Bergmann, www.tuxad.com, 2010-11-19
 */

#include "../lib/libfwb.h"
#include <linux/capability.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/prctl.h>
#include <sysexits.h>

/* to build on kernels < 2.6.25: */
#ifndef PR_CAPBSET_DROP
#define PR_CAPBSET_DROP 24
#endif
#ifndef CAP_TO_MASK
#define CAP_TO_MASK(x)      (1 << ((x) & 31)) /* mask for indexed __u32 */
#endif

#define STRINGSIZE 255
#define USAGE "usage: chrootuser user dir command [options]\n"

/* don't rely on libcap, be diet compatible */
struct my_cap_header_struct {
  uint32_t version;
  int pid;
};
struct my_cap_data_struct {
  uint32_t effective;
  uint32_t permitted;
  uint32_t inheritable;
};

int main(int argc,char *const argv[]) {
  register char *p;
  uint32_t uid, gid;
  const char *pwdbuf;
  off_t pwdlen;
  int s_len;
  struct my_cap_header_struct cap_header;
  struct my_cap_data_struct cap_data;

  if (argc < 4) {
    write2(USAGE, sizeof USAGE - 1);
    _exit(EX_USAGE);
  }
  s_len = strlen(argv[1]);

  /* open /etc/passwd for uid-mapping */
  p = (char *) map_passwd_file(&pwdbuf, &pwdlen);
  if (p == (char *) -1) _exit(EX_OSFILE);

  /* get UID of user */
  p = (char *) get_uid_from_passwd(pwdbuf, pwdlen, argv[1], s_len, &s_len);
  if (!p) { munmap((void *) pwdbuf, pwdlen); _exit(EX_NOUSER); }
  if (s_len > STRINGSIZE) { /* slightly paranoid */
    munmap((void *) pwdbuf, pwdlen);
    _exit(EX_DATAERR);
  }
  p = str2uint32(p, &uid);
  if (!p) _exit(EX_OSFILE);
  ++p;
  p = str2uint32(p, &gid);
  munmap((void *) pwdbuf, pwdlen);

  if (chroot(argv[2])) _exit(EX_DATAERR);
  if (chdir("/")) _exit(EX_DATAERR);;

  if (uid) {
    fchown32m(0, uid, gid);
    fchown32m(1, uid, gid);
    fchown32m(2, uid, gid);
    setresgid32m(gid, gid, gid);
    setresuid32m(uid, uid, uid);
  } else {
    /* if superuser is wanted then remove chroot and module capability */
    prctl(PR_CAPBSET_DROP, CAP_SYS_MODULE);
    prctl(PR_CAPBSET_DROP, CAP_SYS_CHROOT);
    cap_header.version = _LINUX_CAPABILITY_VERSION;
    cap_header.pid = 0;
    if (!syscall(__NR_capget, &cap_header, &cap_data)) {
      cap_data.permitted &= ~CAP_TO_MASK(CAP_SYS_CHROOT);
      cap_data.permitted &= ~CAP_TO_MASK(CAP_SYS_MODULE);
    } else {
     /* if capget fails then try no caps as fallback */
      cap_data.permitted = 0;
    }
    cap_data.effective = cap_data.permitted;
    cap_data.effective = 0;
    cap_data.inheritable = 0;
    if (syscall(__NR_capset, &cap_header, &cap_data)) _exit(EX_DATAERR);
  }

  execve(argv[3],argv+3,0);

  _exit(2);
}
