/*
 * direntries.c
 * manage entries (not inodes!) within a directory
 * $Id: direntries-bk.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <dirent.h>
#include <stdint.h>
#include <time.h>
#include <errno.h>
#include "../lib/libfwb.h"


#define PATHNAME_SIZE 16384

const char usage[]="usage:\n"
  "direntries [-l] [directory]\n"
  "direntries -D [-e file-extension] [-P filename-prefix] [-t min-age-in-days] [directory] # delete mode\n";

enum { MODE_NORMAL, MODE_DELETE };

static void print(const char *str) {
  write1(str,strlen(str));
}

static void my_exit(const int retval) {
  _exit(retval);
}

static int64_t str2uint63(register const char *s) {
  register int64_t i;
  register char c;

  c = *s++;
  if ((c<'0') || (c>'9')) return -1;
  i = c-'0';
  for(;;) {
    c=*s;
    if((c<'0')||(c>'9')) {
      return i;
    } else {
      i=(i*10)+c-'0';
      s++;
    }
  }
}

static int str_diff_extension(const char *filename, const char *file_extension, uint32_t n) {
  register const char *s;
  register const char *t;
  int j;
  uint32_t m;

  m = strlen(filename);
  if (m <= n) return -1;
  s = filename + m - n;
  t = file_extension;
  for(;;) {
    if ((j = (*s-*t))) break; if (!*t) break; if (!*s) break; ++s; ++t;
  }
  return j;
}

static void direntries_normal_mode(const char *dir, int list_mode) {
  register uint32_t count;
  register DIR *dptr;
  register struct dirent *eptr;

  dptr = opendir(dir);
  if (dptr == 0) my_exit(1);
  for (count = 0; (eptr = readdir(dptr)); count++) if (list_mode) {
    write1(eptr->d_name, strlen(eptr->d_name)); write1("\n", 1);
  }
  closedir(dptr);
  if (!list_mode) putlong(count);
  my_exit(0);
}

static void direntries_delete_mode(register const char *dir, const char *file_extension, const char *file_prefix, register uint32_t minage) {
  register char *nptr;
  uint32_t count;
  DIR *dptr;
  struct dirent *eptr;
  time_t tt;
  struct stat statbuf;
  char pathname[3*PATHNAME_SIZE];
  uint32_t l_file_prefix;
  uint32_t l_file_extension;

  tt = time(0) - (minage * 86400);
  dptr = opendir(dir);
  if (dptr == 0) my_exit(1);
  l_file_prefix = strlen(file_prefix);
  l_file_extension = strlen(file_extension);

  for (count = 0; (eptr = readdir(dptr));) {
    nptr = eptr->d_name;

    if (file_extension) {
      if (!str_diff_extension(nptr, file_extension, l_file_extension)) {
        *pathname = '\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        if (minage) {
          if (lstat(pathname, &statbuf)) continue;
          if (statbuf.st_mtime >= tt) continue;
        }
        unlink(pathname);
        count++;
      }
      continue;
    }

    if (file_prefix) {
      if (!str_diffn(nptr, file_prefix, l_file_prefix)) {
        *pathname = '\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        if (minage) {
          if (lstat(pathname, &statbuf)) continue;
          if (statbuf.st_mtime >= tt) continue;
        }
        unlink(pathname);
        count++;
      }
      continue;
    }

    *pathname = '\0';
    str_catn(pathname,dir,PATHNAME_SIZE);
    str_catn(pathname,"/",1);
    str_catn(pathname,nptr,PATHNAME_SIZE);
    if (minage) {
      if (lstat(pathname, &statbuf)) continue;
      if (statbuf.st_mtime >= tt) continue;
    }
    unlink(pathname);
    count++;

  }

  closedir(dptr);
  putlong(count);
  my_exit(0);
}

int main(int argc,char *const argv[]) {
  register int c;
  register char *dir;
  register char *file_extension;
  register char *file_prefix;
  register int mode;
  register int list_mode;
  uint32_t minage;
  const char wd[]=".";

  file_prefix = file_extension = 0;
  dir = (char *) wd;
  list_mode = 0;
  minage = 0;
  mode = MODE_NORMAL;
  for (c = 1; c < argc; c++) {
    if (*argv[c] == '-') {
      if (argv[c][2]) { print(usage); _exit(1); }
      switch (argv[c][1]) {
        case 'D': mode = MODE_DELETE; break;
        case 'l': list_mode = 1; break;
        case 't': c++; minage = (uint32_t) str2uint63(argv[c]); break;
        case 'e': c++; file_extension = argv[c]; break;
        case 'P': c++; file_prefix = argv[c]; break;
        default: print(usage); _exit(0);
      }
    } else {
      dir = argv[c];
    }
  }
  switch(mode) {
    case MODE_DELETE:
      direntries_delete_mode(dir, file_extension, file_prefix, minage);
      break;
    default:
      direntries_normal_mode(dir, list_mode);
      break;
  }
  _exit(0);
}
