/*
 * direntries.c
 * manage entries (not inodes!) within a directory
 * $Id: direntries.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <dirent.h>
#include <stdint.h>
#include <time.h>
#include <errno.h>
#include "../lib/libfwb.h"


#define PATHNAME_SIZE 16384

const char usage[]="usage:\n"
  "direntries [directory]\n"
  "direntries -r [directory] # recursive mode\n"
  "direntries -m [-d depth] [directory] # maximum mode\n"
  "direntries -s [-p] [-e file-extension] [-P filename-prefix] [directory] # show mode\n"
  "direntries -D [-e file-extension] [-P filename-prefix] [-t min-age-in-days] [directory] # delete mode\n";

enum { MODE_NORMAL, MODE_RECURSIVE, MODE_MAXIMUM, MODE_SHOW, MODE_DELETE };

static void print(const char *str) { write1(str,strlen(str)); }

static void my_exit(const int retval) {
  _exit(retval);
}

static int64_t str2uint63(const char *s) {
  register char c;
  int64_t i;

  c = *s; ++s;
  if ((c<'0') || (c>'9')) return -1;
  i = c-'0';
  for(;;) {
    c=*s;
    if((c<'0')||(c>'9')) {
      return i;
    } else {
      i=(i*10)+c-'0';
      ++s;
    }
  }
}

static int str_diff_extension(const char *filename, const char *file_extension, uint32_t n) {
  register const char *s;
  register const char *t;
  int j;
  uint32_t m;

  m = strlen(filename);
  if (m <= n) return -1;
  s = filename + m - n;
  t = file_extension;
  for(;;) {
    if ((j = (*s-*t))) break; if (!*t) break; if (!*s) break; ++s; ++t;
  }
  return j;
}

static int walkdir(name,entries,entries_maxdir,pathname_maxdir,pathname_size,depth,curdepth,subdir)
const char *name;
uint32_t *entries;
uint32_t *entries_maxdir;
char *pathname_maxdir;
uint16_t pathname_size;
int32_t depth;
int32_t curdepth;
const char *subdir;
{

  register DIR *dptr;
  struct dirent *eptr;
  struct stat statbuf;
  uint64_t entries_sum;
  char *nptr;
  int dotdirs;
  int32_t hlinks;
  int ret;
  char pathname[PATHNAME_SIZE*3];

  dptr = opendir(name);
  if (dptr == NULL) return 1;
  ret = 0;
  dotdirs = 0;
  hlinks = 2;
  if (lstat(name,&statbuf) == 0) {
    hlinks = statbuf.st_nlink;
    if (hlinks >= 2) hlinks -= 2; /* leaf optimization */
  }
  for(entries_sum = 1;;) {
    eptr = readdir(dptr); if (eptr == NULL) break;
    nptr = eptr->d_name;
    if ((dotdirs < 2) && (*nptr == '.')) {
      if (nptr[1] == '\0') { dotdirs += 1; continue; }
      if (nptr[1] == '.') {
        if (nptr[2] == '\0') { dotdirs += 1; continue; }
      }
    }
    if (hlinks > 0) { /* leaf optimization */
      *pathname = '\0';
      str_catn(pathname, name, PATHNAME_SIZE);
      str_catn(pathname, "/", 1);
      str_catn(pathname, nptr, PATHNAME_SIZE);
      if (subdir != NULL) {
        str_catn(pathname, "/", 1);
        str_catn(pathname, subdir, PATHNAME_SIZE);
      }
      if (lstat(pathname, &statbuf) == 0){
        if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
          hlinks--;
          if (depth > 0) {
            if (curdepth < depth) {
              ret += walkdir(pathname, entries, entries_maxdir, pathname_maxdir, pathname_size, depth, curdepth+1, NULL);
            }
          } else {
            ret += walkdir(pathname, entries, entries_maxdir, pathname_maxdir, pathname_size, depth, 0, NULL);
          } /* depth */
        } else {
          ++entries_sum;
        } /* statbuf */
      } else {
        ++ret;
      } /* lstat */
    } else {
      ++entries_sum;
    } /* hlinks */
  }
  closedir(dptr);
  if (entries_maxdir != 0) {
    if ((pathname_maxdir != NULL) && (entries_sum > *entries_maxdir) && (subdir == NULL)) {
      *entries_maxdir = entries_sum;
      *pathname_maxdir = '\0';
      str_catn(pathname_maxdir, name, pathname_size);
    }
  }
  *entries += entries_sum;
  return ret;
}

static void direntries_normal_mode(const char *dir) {
  register uint32_t count;
  DIR *dptr;
  struct dirent *eptr;

  dptr = opendir(dir);
  if (dptr == 0) my_exit(1);
  for (count = 0; (eptr = readdir(dptr)); ++count);
  closedir(dptr);
  putlong(count);
  my_exit(0);
}

static void direntries_max_mode(char *basedir, const int32_t depth) {
  const char *subdir;
  int ret;
  uint32_t entries;
  uint32_t entries_maxdir;
  char pathname_maxdir[PATHNAME_SIZE+1];
  struct rlimit rl = { 134217728, 134217728 };

  setrlimit(RLIMIT_STACK, &rl);
  entries = entries_maxdir = 0;
  ret = str_index(basedir, "/./"); /* "wildcard" like "*" in shell */
  if (ret > 0) {
    basedir[ret] = '\0';
    subdir = basedir + ret + 3;
    ret = walkdir(basedir, &entries, &entries_maxdir, pathname_maxdir, PATHNAME_SIZE, depth, 0, subdir);
  } else {
    ret = walkdir(basedir, &entries, &entries_maxdir, pathname_maxdir, PATHNAME_SIZE, depth, 0, NULL);
  }
  putlong(entries_maxdir);
  str_catn(pathname_maxdir, "\n", 1);
  print(pathname_maxdir);
  my_exit(ret);
}

static void direntries_recursive_mode(const char *dir) {
  struct rlimit rl = { 134217728, 134217728 };
  uint32_t entries=0;
  setrlimit(RLIMIT_STACK, &rl);
  int ret = walkdir(dir, &entries, NULL, NULL, 0, 0, 0, NULL);
  putlong(entries);
  my_exit(ret);
}

static void direntries_show_mode(const char *dir, const char *file_extension, const char *file_prefix, int print_mode) {
  register char *nptr;
  uint32_t count;
  DIR *dptr;
  struct dirent *eptr;
  char pathname[PATHNAME_SIZE+1];
  uint32_t gifs,jpgs,pngs,jss,csss,phps,htmls,pdfs,exturls;
  uint32_t l_file_prefix;
  uint32_t l_file_extension;

  dptr = opendir(dir);
  if (dptr == NULL) my_exit(1);
  if (file_prefix) l_file_prefix = strlen(file_prefix); else l_file_prefix = 0;
  if (file_extension) l_file_extension = strlen(file_extension); else l_file_extension = 0;
  jpgs = gifs = pngs = jss = csss = phps = htmls = pdfs = exturls = 0;

  for (count = 0; (eptr = readdir(dptr));) {
    nptr = eptr->d_name;

    if (file_extension) {
      if (!str_diff_extension(nptr, file_extension, l_file_extension)) {
	++count;
        if(print_mode) {
          *pathname='\0';
          str_catn(pathname,nptr,PATHNAME_SIZE);
          str_catn(pathname,"\n",1);
          print(pathname);
        }
      }
      continue;
    }

    if (file_prefix) {
      if (!str_diffn(nptr, file_prefix, l_file_prefix)) {
	++count;
        if(print_mode) {
          *pathname='\0';
          str_catn(pathname,nptr,PATHNAME_SIZE);
          str_catn(pathname,"\n",1);
          print(pathname);
        }
      }
      continue;
    }

    ++count;
    if(print_mode) {
      *pathname='\0';
      str_catn(pathname,nptr,PATHNAME_SIZE);
      str_catn(pathname,"\n",1);
      print(pathname);
    }
    if(!str_diff_extension(nptr,".gif",4)) { ++gifs; continue; }
    if(!str_diff_extension(nptr,".jpg",4)) { ++jpgs; continue; }
    if(!str_diff_extension(nptr,".png",4)) { ++pngs; continue; }
    if(!str_diff_extension(nptr,".js",3)) { ++jss; continue; }
    if(!str_diff_extension(nptr,".css",4)) { ++csss; continue; }
    if(!str_diff_extension(nptr,".php",4)) { ++phps; continue; }
    if(!str_diff_extension(nptr,".html",5)) { ++htmls; continue; }
    if(!str_diffn(nptr,"pdf_",4)) { ++pdfs; continue; }
    if(!str_diffn(nptr,"EXTERNAL_URL",12)) { ++exturls; continue; }

  }

  closedir(dptr);

  print("*.gif: "); putlong(gifs);
  print("*.jpg: "); putlong(jpgs);
  print("*.png: "); putlong(pngs);
  print("*.js: "); putlong(jss);
  print("*.css: "); putlong(csss);
  print("*.php: "); putlong(phps);
  print("*.html: "); putlong(htmls);
  print("pdf_*: "); putlong(pdfs);
  print("EXTERNAL_URL*: "); putlong(exturls);
  putlong(count);
  my_exit(0);
}

static void direntries_delete_mode(const char *dir, const char *file_extension, const char *file_prefix, uint32_t minage) {
  register char *nptr;
  uint32_t count;
  DIR *dptr;
  struct dirent *eptr;
  time_t tt;
  struct stat statbuf;
  char pathname[3*PATHNAME_SIZE];
  uint32_t l_file_prefix;
  uint32_t l_file_extension;

  tt = time(0) - (minage * 86400);
  dptr = opendir(dir);
  if (dptr == 0) my_exit(1);
  if (file_prefix) l_file_prefix = strlen(file_prefix); else l_file_prefix = 0;
  if (file_extension) l_file_extension = strlen(file_extension); else l_file_extension = 0;

  for (count = 0; (eptr = readdir(dptr));) {
    nptr = eptr->d_name;

    if (file_extension) {
      if (!str_diff_extension(nptr, file_extension, l_file_extension)) {
        *pathname = '\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        if (minage) {
          if (lstat(pathname, &statbuf)) continue;
          if (statbuf.st_mtime >= tt) continue;
        }
        unlink(pathname);
        ++count;
      }
      continue;
    }

    if (file_prefix) {
      if (!str_diffn(nptr, file_prefix, l_file_prefix)) {
        *pathname = '\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        if (minage) {
          if (lstat(pathname, &statbuf)) continue;
          if (statbuf.st_mtime >= tt) continue;
        }
        unlink(pathname);
        ++count;
      }
      continue;
    }

    *pathname = '\0';
    str_catn(pathname,dir,PATHNAME_SIZE);
    str_catn(pathname,"/",1);
    str_catn(pathname,nptr,PATHNAME_SIZE);
    if (minage) {
      if (lstat(pathname, &statbuf)) continue;
      if (statbuf.st_mtime >= tt) continue;
    }
    unlink(pathname);
    ++count;

  }

  closedir(dptr);
  putlong(count);
  my_exit(0);
}

int main(int argc,char *const argv[]) {
  register int c;
  char *dir;
  char *file_extension;
  char *file_prefix;
  int mode;
  int print_mode;
  int32_t depth;
  uint32_t minage;
  const char wd[]=".";

  file_prefix = file_extension = 0;
  dir = (char *) wd;
  print_mode = 0;
  depth = 2;
  minage = 0;
  mode = MODE_NORMAL;
  for (c = 1; c < argc; ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2]) { print(usage); _exit(1); }
      switch (argv[c][1]) {
        case 'h': print(usage); _exit(0);
        case 'D': mode = MODE_DELETE; break;
        case 's': mode = MODE_SHOW; break;
        case 'm': mode = MODE_MAXIMUM; break;
        case 'r': mode = MODE_RECURSIVE; break;
        case 'p': print_mode = 1; break;
        case 't': ++c; minage = (uint32_t) str2uint63(argv[c]); break;
        case 'd': ++c; depth = (uint32_t) str2uint63(argv[c]); break;
        case 'e': ++c; file_extension = argv[c]; break;
        case 'P': ++c; file_prefix = argv[c]; break;
        default: print(usage); _exit(0);
      }
    } else {
      dir = argv[c];
    }
  }
  if (print_mode && (mode != MODE_SHOW)) { print(usage); _exit(1); }
  switch(mode) {
    case MODE_RECURSIVE:
      direntries_recursive_mode(dir);
      break;
    case MODE_MAXIMUM:
      direntries_max_mode(dir, depth);
      break;
    case MODE_SHOW:
      direntries_show_mode(dir, file_extension, file_prefix, print_mode);
      break;
    case MODE_DELETE:
      direntries_delete_mode(dir, file_extension, file_prefix, minage);
      break;
    default:
      direntries_normal_mode(dir);
      break;
  }
  _exit(0);
}
