/**
 * dirscan.c
 *
 * Compile with:
 * diet gcc -O2 -Wall -o dirscan dirscan.c && sstrip dirscan
 * Frank Bergmann, www.tuxad.com, 11/2008
 * $Id: dirscan.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wmisleading-indentation"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <stdint.h>
#include <limits.h>


#undef USE_STRING_LIB
#define USE_STRING_LIB
#ifdef USE_STRING_LIB
#include <string.h>
#endif

#ifndef write2
#include <sys/syscall.h>
#define write2(b,c) (syscall(__NR_write, 2, b, c))
#endif


#define MAX_CHILDS 100
#define SLEEP 2

extern char **environ;
static const char msg[]="usage: dirscan [-m] [-d changedir] [-I includestring] [-X excludestring] path/to/command [arg(s)]\n";


#ifndef USE_STRING_LIB

int strlen(register const char* in) {
  register const char *t = in;

  if (!t) return 0;
  for (;;) { if (!*t) break; ++t; }
  return (int)(t-in);
}

int strncmp(register const char* s, register const char* t, int limit) {
  register const char* u=t+limit;
  register int j;
  j=0;
  for (;;) { if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t; }
  return j;
}

char *strcpy(register char* out, register const char* in) {
  for (;;) { if (!*in) break; *out++=*in++; }
  *out='\0';
  return out;
}

#endif

void my_exit(int r,const char *s) { write2(s,strlen(s)); _exit(r); }

int32_t str_index(const char* a, const char* b) {
  register const char* s;
  register char z;
  register const char* u;
  register int32_t t;

  s = a;
  t = strlen(b);
  u = s + strlen(s) - t;
  z = *b;
  for (;;) {
    if (s > u) break; if ((*s == z) && (!strncmp(s, b, t))) return s - a; ++s;
    if (s > u) break; if ((*s == z) && (!strncmp(s, b, t))) return s - a; ++s;
    if (s > u) break; if ((*s == z) && (!strncmp(s, b, t))) return s - a; ++s;
    if (s > u) break; if ((*s == z) && (!strncmp(s, b, t))) return s - a; ++s;
  }
  return -1;
}

int main(int argc,char *argv[]) {
  register char *p;
  register int c;
  int firstfree;
  char *exclude_string;
  char *include_string;
  int childstatus;
  int sstring;
  char *changedir1;
  char *changedir2;
  char fnames[MAX_CHILDS][PATH_MAX+2];
  char fpaths[MAX_CHILDS][PATH_MAX+2];
  pid_t cpids[MAX_CHILDS];
  time_t mtimes[MAX_CHILDS];
  pid_t child;
  struct dirent *dent;
  DIR *curdir;
  int sort_by_mtime = 0;
  struct stat stbuf;
  char **args;
  char wd[PATH_MAX+2];
  int wdlen;

  if (argc < 2) my_exit(127, msg);
  changedir1 = changedir2 = 0;
  sstring = -1;
  exclude_string = include_string = 0;
  for (c = 1; (c < argc) && (sstring == -1); ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') my_exit(127, msg);
      switch (argv[c][1]) {
        case 'd': changedir1 = argv[++c]; break;
        case 'D': changedir2 = argv[++c]; break;
        case 'I': include_string = argv[++c]; break;
        case 'X': exclude_string = argv[++c]; break;
        case 'm': sort_by_mtime = 1; break;
        default: my_exit(127, msg);
      }
    } else {
      sstring = c;
      break;
    }
  }
  if (sstring == -1) my_exit(127, msg);

  if (changedir1 && (chdir(changedir1) != 0)) my_exit(111, "can't change dir\n");
  p = getcwd(wd, PATH_MAX);
  if (!p) my_exit(111, "can't get working dir\n");
  wdlen = strlen(wd);

  sstring--;
  args = &argv[sstring];
  /* move array elements by 1 back to add filename for execve() */
  for (c = sstring; c < argc - 1; ++c) {
    argv[c] = argv[c+1];
  }

  for (c = 0; c < MAX_CHILDS; ++c) fnames[c][0] = 0;


  /* main loop */
  for(;;) {

    /* free element in array? */
    firstfree = -1;
    for (c = 0; c < MAX_CHILDS; ++c) {
      if (!*fnames[c]) {
        firstfree = c;
	break;
      }
    }

    /* array full? => sleep and run again */
    if (firstfree == -1) goto sleep_and_waitpid;
    mtimes[firstfree] = 0;

    curdir = opendir(".");
    for (;;) {
  
      next_dirent:
      dent = readdir(curdir);
      if (!dent) break;

      /* does file match any rules? */
      p = dent->d_name;
      if (*p == '.') continue; /* skip names beginning with dot */
      if (exclude_string && (str_index(p, exclude_string) != -1)) continue;
      if (include_string && (str_index(p, include_string) == -1)) continue;

      /* file already processed? */
      for (c = 0; c < MAX_CHILDS; ++c) {
        if (*fnames[c]) {
          if (!strncmp(fnames[c], p, strlen(p))) {
            /* file is processed by a child, next file */
            goto next_dirent;
          }
        }
      }

      /* file's not already processed */
      if (sort_by_mtime) {
        if (lstat(p, &stbuf)) {
          stbuf.st_mtime = 0;
        }
      } else {
        stbuf.st_mtime = 0;
      }
      if (*fnames[firstfree]) {
        if (sort_by_mtime && stbuf.st_mtime) {
          if (stbuf.st_mtime < mtimes[firstfree]) goto file_presorted;
            continue;
        }
        if (strncmp(fnames[firstfree], p, strlen(p)) < 0) continue;
      }

      file_presorted:
      /* file is currently at top of sort order */
      strncpy(fnames[firstfree], p, PATH_MAX);
      mtimes[firstfree] = stbuf.st_mtime;
  
    }
  
    closedir(curdir);
    if (!*fnames[firstfree]) goto sleep_and_waitpid;

    p = strncpy(fpaths[firstfree], wd, PATH_MAX);
    p += wdlen;
    *p = '/'; ++p;
    p = strncpy(p, fnames[firstfree], PATH_MAX);
    argv[argc-1] = fpaths[firstfree];

    /* we just use execve() after forking, use vfork() */
    child = vfork();
    if (child == -1) {
      /* error on vfork() */
      *fnames[firstfree] = 0;
      write2("can't fork\n", 11);
      goto sleep_and_waitpid;
    }

    if (child) {
      /* father */
      cpids[firstfree] = child; /* assigned after execve() */
    } else {
      /* child */
      if (changedir2 && (chdir(changedir2) != 0)) goto no_exec;
      execve(args[0], args, environ);
      no_exec:
      *fnames[firstfree] = 0;
      write2("exec err\n",9);
      _exit(111);
    }

    sleep_and_waitpid:
    sleep(SLEEP);

    /* delete array entries of terminated childs */
    for (c = 0; c < MAX_CHILDS; ++c) {
      if ((*fnames[c]) || (cpids[c] != 0)) {
        if (waitpid(cpids[c], &childstatus, WNOHANG) == cpids[c]) {
          if (WIFEXITED(childstatus) || WIFSIGNALED(childstatus)) {
            *fnames[c] = 0;
            cpids[c] = 0;
          }
        }
      }
    }
  }

  /* never reached */
  _exit(0);
}

