/*
 * ftsm.c
 * file timestamp manager
 * $Id: ftsm.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <utime.h>
#include <dirent.h>
#include <stdint.h>
#include <time.h>
#include <errno.h>
#include "../lib/buffer.h"
#include "../lib/libfwb.h"


#define PATHNAME_SIZE 16384
#define READBUFSIZE 4096
#define WRITEBUFSIZE 8192

char buf0space[READBUFSIZE+1];
char buf1space[WRITEBUFSIZE+1];
buffer *buf1;
int mercurial = 0;

const char usage[]="usage:\n"
  "ftsm [-d depth] directory [directory] >.timestamps\n"
  "ftsm <.timestamps\n";

static void print(const char *str) { write1(str,strlen(str)); }

static void my_exit(const int retval) {
  buffer_flush(buf1);
  _exit(retval);
}

static int64_t str2uint63(const char *s) {
  register char c;
  int64_t i;

  c = *s; ++s;
  if ((c<'0') || (c>'9')) return -1;
  i = c-'0';
  for(;;) {
    c=*s;
    if((c<'0')||(c>'9')) {
      return i;
    } else {
      i=(i*10)+c-'0';
      ++s;
    }
  }
}

int buffer_getstring(buffer *b, char *string, int size, int nlterm) {
  register char *p;
  register ssize_t l;
  register int readbytes = 0;

  if (!string) return EOF;
  if (size < 1) return EOF;
  if (!b) return EOF;
  p = b->c;
  if (!p) return EOF;

  check_end:
  if (p >= b->t) {
    p = b->s;
    l = read(0, p, b->l);
    if (l <= 0) {
      b->c = 0; /* EOF */
      *string = 0;
      if (l < 0) return EOF - 1;
      return EOF;
    }
    b->t = p + l;
  }

  if ( (!nlterm && ((*p == ' ') || (*p == '\t') || (*p == '\n') || (*p == '\r')) )
     || (nlterm && ((*p == '\n') || (*p == '\r'))) ) {
    b->c = p + 1;
    goto out;
  }
  *string = *p;
  ++string;
  ++readbytes;
  --size;
  if (size < 1) goto out;
  ++p;
  b->c = p;
  goto check_end;

  out:
  *string = 0;
  return readbytes;
}

static int walkdir(name,depth,curdepth,subdir)
const char *name;
int32_t depth;
int32_t curdepth;
const char *subdir;
{

  register DIR *dptr;
  struct dirent *eptr;
  struct stat statbuf;
  char *nptr;
  int dotdirs;
  int32_t hlinks;
  int ret;
  char pathname[PATHNAME_SIZE*3];

  dptr = opendir(name);
  if (dptr == NULL) return 1;
  ret = 0;
  dotdirs = 0;
  for(;;) {
    eptr = readdir(dptr); if (eptr == NULL) break;
    nptr = eptr->d_name;
    if ((dotdirs < 2) && (*nptr == '.')) {
      if (nptr[1] == '\0') { dotdirs += 1; continue; }
      if (nptr[1] == '.') {
        if (nptr[2] == '\0') { dotdirs += 1; continue; }
      }
    }
    *pathname = '\0';
    str_catn(pathname, name, PATHNAME_SIZE);
    str_catn(pathname, "/", 1);
    str_catn(pathname, nptr, PATHNAME_SIZE);
    if (subdir != NULL) {
      str_catn(pathname, "/", 1);
      str_catn(pathname, subdir, PATHNAME_SIZE);
    }
    if (lstat(pathname, &statbuf) == 0){
      if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
        hlinks--;
        if (depth > 0) {
          if (curdepth < depth) {
            ret += walkdir(pathname, depth, curdepth+1, NULL);
          }
        } else {
          ret += walkdir(pathname, depth, 0, NULL);
        } /* depth */
      } else {
        if (mercurial) {
          buffer_putstr(buf1, pathname);
          buffer_putc(buf1, ' ');
          buffer_putuint64(buf1, statbuf.st_mtime);
          buffer_putc(buf1, '\n');
        } else {
          buffer_putuint64(buf1, statbuf.st_mtime);
          buffer_putc(buf1, ' ');
          buffer_putstr(buf1, pathname);
          buffer_putc(buf1, '\n');
        }
      } /* statbuf */
    } else {
      ++ret;
    } /* lstat */
  }
  closedir(dptr);
  return ret;
}

int main(int argc,char *const argv[]) {
  register char *p;
  register int c;
/*
  char *file_extension;
  char *file_prefix;
*/
  int32_t depth;
  int save_mode = 0;
  int ret = 0;

  p = buf1space; buffer buf = { p, p, p + WRITEBUFSIZE, 1 }; buf1 = &buf;

/*
  file_prefix = file_extension = 0;
*/
  depth = 0;
  for (c = 1; c < argc; ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2]) { print(usage); _exit(1); }
      switch (argv[c][1]) {
        case 'h': print(usage); _exit(0);
        case 'd': ++c; depth = (uint32_t) str2uint63(argv[c]); break;
/*
        case 'e': ++c; file_extension = argv[c]; break;
        case 'P': ++c; file_prefix = argv[c]; break;
*/
        case 'M': mercurial = 1; break;
        default: print(usage); _exit(0);
      }
    } else {
      save_mode = 1;
      ret += walkdir(argv[c], depth, 1, NULL);
    }
  }
 
  if (!save_mode) {
    char ts_string[PATHNAME_SIZE+1];
    char path[PATHNAME_SIZE+1];
    struct stat statbuf;
    struct utimbuf timebuf;
    int64_t ts;

    p = buf0space; buffer buf0 = { p, p + READBUFSIZE, p + READBUFSIZE, 0, READBUFSIZE };

    if (mercurial) {
      ret = buffer_getstring(&buf0, path, PATHNAME_SIZE, 0);
      while (ret > 0) {
        ret = buffer_getstring(&buf0, ts_string, PATHNAME_SIZE, 1);
        ts = str2uint63(ts_string);
        if (ts < 0) my_exit(ret);
        if (lstat(path, &statbuf) == 0) {
          timebuf.actime = statbuf.st_atime;
          timebuf.modtime = (time_t) ts;
          utime(path, &timebuf);
        }
        ret = buffer_getstring(&buf0, path, PATHNAME_SIZE, 0);
      }
    } else {
      ret = buffer_getstring(&buf0, ts_string, PATHNAME_SIZE, 0);
      while (ret > 0) {
        ts = str2uint63(ts_string);
        if (ts < 0) my_exit(ret);
        ret = buffer_getstring(&buf0, path, PATHNAME_SIZE, 1);
        if (lstat(path, &statbuf) == 0) {
          timebuf.actime = statbuf.st_atime;
          timebuf.modtime = (time_t) ts;
          utime(path, &timebuf);
        }
        ret = buffer_getstring(&buf0, ts_string, PATHNAME_SIZE, 0);
      }
    }
  }

  my_exit(ret);
}
