#define _DARWIN_C_SOURCE
/*
 * lps.c
 * show process info like 'ps' with these columns:
 *   - process age in minutes
 *   - pid
 *   - effective user
 *   - RSS
 *   - VPID (=0 if not OpenVZ)
 *   - CTID/FNID (=0 if not OpenVZ)
 *   - process name/ executable
 *   - string "socket" if process has at least 1 open socket
 *   - (ch)root in fs
 *   - working dir
 *   - file executable
 *   - cmdline
 *  [- HTTP_USER_AGENT]
 * Example usage:
 * lps 1 | sort -n
 * (show all processes except root and sort by age)
 *
 * lps
 * Zeige Prozessinformationen aehnlich wie 'ps' an, mit folgenden Spalten:
 *   - Prozessalter in Minuten (nicht "Laufzeit", sondern "Alter")
 *   - PID
 *   - effektiver Nutzer
 *   - Resident Set Size
 *   - VPID (bei OpenVZ/Virtuozo, ansonsten = 0)
 *   - CTID/FNID (= 0 wenn kein OpenVZ)
 *   - Name des Prozesses wie in /proc/pid/stat angezeigt
 *   - Zeichenkette "socket", wenn der Prozess mind. ein Socket geoeffnet hat
 *   - (ch)root des Prozesses im Dateisystem
 *   - Arbeitsverzeichnis
 *   - ausfuehrbare Prozessdatei
 *   - Kommandozeile
 *  [- HTTP_USER_AGENT]
 * Beispielanwendung:
 * lps 1 | sort -n
 * (zeige alle Prozesse ausser root-Prozesse und sortiere sie nach Alter)
 *
 * $Id: lps.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#define MAX_USERS_IN_PASSWD 4000

#include "../lib/libfwb.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <pwd.h>
#include <stdint.h>
#include <sysexits.h>

#define BUFSIZE 16384
#define WRITEBUFSIZE 2048

static const char usage[]="usage: lps [-s] [-c] [-u] [-U] [-L proclimit] [min-uid [max-uid]]\n";
static const char uid_marker[]="Uid:\011";
/* minimum offset of uid_marker in /proc/PID/status */
#define UID_MIN_OFFSET 45
/* only used for a small speedup */

typedef struct st_lps_global {
  int show_http_user_agent;
  int mode_seconds;
  int mode_user;
  int mode_limit;
  int mode_thread;
  uint32_t mode_count;
  uint32_t proclimit;
  uint32_t uidp[MAX_USERS_IN_PASSWD+1];
  uint32_t uidcnt[MAX_USERS_IN_PASSWD];
  uint32_t max1uid, max2uid, max3uid, max1cnt, max2cnt, max3cnt;
  uint64_t start_time;
  uint32_t uid_min;
  uint32_t uid_max;
  int jiffies_per_second;
  int32_t pagesize;
  const char *pwdbuf;
  off_t pwdlen;
  buffer *wbuf;
} lps_global;

static inline void printusage() {
  write1(usage, sizeof(usage)-1);
  _exit(EX_USAGE);
}

static inline int str_index3(const char* a, register const char* b, int al, int t) {
  register const char* s=a;
  register const char* u=s+al-t;
  for (;;) {
    if (s>u) break; if ((*s == *b) && (!strncmp(s,b,t))) return (s-a)+t; ++s;
    if (s>u) break; if ((*s == *b) && (!strncmp(s,b,t))) return (s-a)+t; ++s;
    if (s>u) break; if ((*s == *b) && (!strncmp(s,b,t))) return (s-a)+t; ++s;
    if (s>u) break; if ((*s == *b) && (!strncmp(s,b,t))) return (s-a)+t; ++s;
  }
  return -1;
}

static char *field(const char *s, uint16_t index) {
  register const char* t=s;

  if (!index) return (char *) s;
  for(;;) {
    if (!*t) return 0; if (*t == ' ') { --index; if (!index) return (char *) ++t; } ++t;
    if (!*t) return 0; if (*t == ' ') { --index; if (!index) return (char *) ++t; } ++t;
    if (!*t) return 0; if (*t == ' ') { --index; if (!index) return (char *) ++t; } ++t;
    if (!*t) return 0; if (*t == ' ') { --index; if (!index) return (char *) ++t; } ++t;
  }
}

static inline int is_socket_link(register const char *s) {
  if (*s != 's') return 0;
  if (*++s != 'o') return 0;
  if (*++s != 'c') return 0;
  if (*++s != 'k') return 0;
  if (*++s != 'e') return 0;
  if (*++s != 't') return 0;
  if (*++s != ':') return 0;
  return 1;
}


void print_pid_line(lps_global *lps, const char *procdir, const char *pid_dir) {
  register char *p;
  register char *s;
  int rf;
  ssize_t c;
  uint32_t val32;
  char name[2*BUFSIZE+1];
  char filebuf[BUFSIZE+1];
  buffer *wbuf;
  uint32_t uid;
  int pid_s_len;
  uint64_t val64;
  uint32_t process_age;
  int pw_name_len;
  DIR *sdirp;
  struct dirent *sdirep;
/*
  DIR *procdirp;
  off_t pwdlen;
*/

  /**
   * verwende kein stat()/lstat(), lese statt dessen .../status;
   * funktioniert bei ganz alten und ganz neuen Linux-Systemen,
   * da Strings tatsaechlich die einzig portable Darstellung 
   * von Zahlen sind: unabhaengig von Typen einer Sprache, unabhaengig
   * Bibliotheken und auch vom Kernel
   */
  s = stpcpy(name, procdir); *s++ = '/';
  s = stpcpy(s, pid_dir);
  stpcpy(s, "/status");
  /* lese UID */
  rf = open(name, O_RDONLY);
  if (rf == -1) return;
  p = filebuf;
  c = read(rf, p, BUFSIZE);
  close(rf);
  if (c < 45) return;
  p[c] = '\0';

  /* suche String "Uid:" with a light offset */
  if ((c = str_index3(p+UID_MIN_OFFSET, uid_marker, c, sizeof(uid_marker)-1)) < 0) return;
  /* p+c zeigt auf Real UID */
  p = strchr(p+c+UID_MIN_OFFSET, '\t') + 1;
  str2uint32(p, &val32); /* Effective UID */
  if (val32 < lps->uid_min) return;
  if (lps->uid_max && (val32 > lps->uid_max)) return;

  if (lps->mode_count) ++lps->mode_count;
  if (lps->mode_user) {
    /* suche nach UID in unserer Liste und zaehle deren Zaehler hoch */
    c = 0;
    for(;;) {
      if (c >= MAX_USERS_IN_PASSWD) break;
      if (lps->uidp[c] == 0) {
        lps->uidp[c] = val32; lps->uidcnt[c] = 1;
        lps->uidp[++c] = 0; break;
      }
      if (lps->uidp[c] == val32) {
        lps->uidcnt[c] += 1;
        break;
      }
      ++c;
    }
  }
  if (lps->mode_user || lps->mode_count) return;
  uid = val32;

  /* lese Prozess stat-Datei */
  s = stpcpy(name, "/proc/");
  p = stpcpy(s, pid_dir);
  pid_s_len = p - s;
  stpcpy(p, "/stat");

  wbuf = lps->wbuf;
  rf = open(name, O_RDONLY);
  if (rf >= 0) {
    p = filebuf;
    c = read(rf, p, BUFSIZE);
    close(rf);
    if (c == -1) return;
    p[c] = '\0';
    /* der 21ste Wert ist "starttime" in Einheit Jiffies */
    s = field(p, 21);
    if (s) {
      /* bei Fehler so verhalten, als waere Prozess gerade gestartet */
      if(!str2uint64(s, &val64)) val64 = lps->start_time;
      s = p;
    } else {
      /* bei Fehler so verhalten, als waere Prozess gerade gestartet */
      val64 = lps->start_time;
    }
    /* verhindere "negatives" Alter */
    if (val64 > lps->start_time) val64 = lps->start_time;
    if (lps->mode_seconds) {
      process_age = (lps->start_time - val64) / lps->jiffies_per_second;
    } else {
      process_age = (lps->start_time - val64) / lps->jiffies_per_second / 60;
    }

    /* Beginn der Ausgabezeile */
    buffer_putuint64(wbuf, process_age); /* Prozessalter */
    buffer_putc(wbuf, ' ');
    buffer_put(wbuf, pid_dir, pid_s_len); /* PID */
    buffer_putc(wbuf, ' ');

    s = (char *) get_name_from_passwd(lps->pwdbuf, lps->pwdlen, uid, &pw_name_len);
    if (s) {
      buffer_put(wbuf, s, pw_name_len); /* Benutzername */
    } else {
      buffer_put(wbuf, "uid", 3);
      buffer_putuint64(wbuf, uid); /* alternativ UID */
    }

    s = field(p, 23);
    if (s) {
      if (!str2uint64(s, &val64)) val64 = 0;
    } else {
      val64 = 0;
    }
    val64 = val64 / 1024 * lps->pagesize; /* in KB */
    buffer_putc(wbuf, ' ');
    buffer_putuint64(wbuf, val64); /* RSS */

    s = field(p, 49);
    if (s) {
      if (!str2uint32(s, &val32)) val32 = 0;
    } else {
      val32 = 0;
    }
    buffer_putc(wbuf, ' ');
    buffer_putuint64(wbuf, val32); /* VPID (OpenVZ) */

    s = field(p, 50);
    if (s) {
      if (!str2uint32(s, &val32)) val32 = 0;
    } else {
      val32 = 0;
    }
    buffer_putc(wbuf, ' ');
    buffer_putuint64(wbuf, val32); /* CTID/FnID (OpenVZ) */

    s = strchr(p, ')'); if (!s) s = p + 10;
    *s = 0;
    val32 = s - p;
    s = strchr(p, '(');
    if (!s) s = p + 5;
    *s = ' ';
    val32 -= s - p;
    buffer_put(wbuf, s, val32); /* Prozessname */

    /* Offene Sockets? */
    s = stpcpy(name, "/proc/");
    s = stpcpy(s, pid_dir);
    s = stpcpy(s, "/fd");

    /* gehe durch das FileDescriptor-Verzeichnis und suche nach Socket-Symlinks */
    sdirp = opendir(name);
    if (!sdirp) goto nosocket;
    *s = '/'; ++s;
    for(;(sdirep = readdir(sdirp)) != NULL;) {
      stpcpy(s, sdirep->d_name);
      c = readlink(name, p, BUFSIZE);
      if (c < 0) continue;
      /* p[c] = 0; */
      if (is_socket_link(p)) goto socket;
    }
    closedir(sdirp);
  
    nosocket:
    buffer_put(wbuf, " ?", 2);
    goto sockfollower;

    socket:
    buffer_put(wbuf, " socket", 7);
    closedir(sdirp);

    sockfollower:

    /* Laeuft der Prozess in einer chroot-Umgebung? */
    s = stpcpy(name, "/proc/");
    s = stpcpy(s, pid_dir);
    stpcpy(s, "/root");
    c = readlink(name, p, BUFSIZE);
    if (c < 0) buffer_put(wbuf, " ?", 2);
    else {
      buffer_putc(wbuf, ' ');
      buffer_put(wbuf, p, c); /* (ch)root */
    }

    /* Arbeitsverzeichnis des Prozesses */
    stpcpy(s, "/cwd");
    c = readlink(name, p, BUFSIZE);
    if (c < 0) buffer_put(wbuf, " ?", 2);
    else {
      buffer_putc(wbuf, ' ');
      buffer_put(wbuf, p, c);
    }

    /* die ausfuehrbare Programmdatei des Prozesses */
    stpcpy(s, "/exe");
    c = readlink(name, p, BUFSIZE);
    if (c < 0) buffer_put(wbuf, " ?", 2);
    else {
      buffer_putc(wbuf, ' ');
      buffer_put(wbuf, p, c);
    }

    /* vollstaendige Befehlszeile des Prozesses */
    stpcpy(s, "/cmdline");
    rf = open(name, O_RDONLY);
    if (rf >= 0) {
      c = read(rf, p, BUFSIZE);
      if (c > 0) {
        int lastnull = 0;
        p[c] = '\0';
        --c;
        val32 = c;
        /* aendere jedes \0 zu Leerzeichen, jedes \n zu einem Fragezeichen */
        for(--c;c > 0;--c) {
          if(!p[c]) {
            p[c] = ' ';
          } else {
            if (!lastnull) lastnull = c + 1;
            if (p[c] == '\n') p[c] = '?';
          }
        }
        p[lastnull] = 0;
        val32 = lastnull;
        buffer_putc(wbuf, ' ');
        buffer_put(wbuf, p, val32);
      } else {
        buffer_put(wbuf, " ?", 2);
      }
      close(rf);
    } else {
      buffer_put(wbuf, " ?", 2);
    }

    if (lps->show_http_user_agent) {
      /* Umgebungsvariablen: HTTP_USER_AGENT */
      stpcpy(s, "/environ");
      rf = open(name, O_RDONLY);
      if (rf >= 0) {
        c = read(rf, p, BUFSIZE);
        if (c > 0) {
          val32 = c;
          /* suche HTTP_USER_AGENT */
          for(c=0;c<val32;) {
            if (!strncmp(p+c, "HTTP_USER_AGENT=", 16)) {
              p += c;
              buffer_putc(wbuf, ' ');
              buffer_put(wbuf, p, strlen(p));
              break;
            } else {
              c += strlen(p+c) + 1;
            }
          }
        }
        close(rf);
      }
    }

    buffer_putc(wbuf, '\n');
  }

  return;
}



void print_pid_task_lines(lps_global *lps, const char *task_dir) {
  register char *s;
  char procdir[2*BUFSIZE+1];
  DIR *procdirp;
  struct dirent *procdirep;

  s = stpcpy(procdir, "/proc/");
  s = stpcpy(s, task_dir);
  stpcpy(s, "/task");
  /* gehe durch das /proc/<PID>/task/ Verzeichnis */
  procdirp = opendir(procdir);
  if (procdirp == NULL) _exit(EX_OSFILE);
  for(;(procdirep = readdir(procdirp)) != NULL;) {
    if ((*procdirep->d_name < '0') || (*procdirep->d_name > '9')) continue;
    print_pid_line(lps, procdir, procdirep->d_name);
  }
  closedir(procdirp);
  return;
}


int main(int argc,char *const argv[]) {
  register char *p;
  register char *s;
  uint32_t val32;
  uint64_t val64;
  ssize_t c;
  int rf;
  char filebuf[BUFSIZE+1];
  char wbufspace[WRITEBUFSIZE+1];
  DIR *procdirp;
  int pw_name_len;
  struct dirent *procdirep;
  lps_global lps;
/*
  int pid_s_len;
  uint32_t uid;
  DIR *sdirp;
  struct dirent *sdirep;
*/


  lps.proclimit =
    lps.uid_min =
    lps.uid_max =
    lps.mode_seconds =
    lps.mode_count =
    lps.mode_user =
    lps.mode_limit =
    lps.mode_thread =
    lps.show_http_user_agent =
    lps.start_time =
    lps.pagesize =
    pw_name_len = 0;
  lps.uidp[0] =
    lps.uidcnt[0] = 0;
  lps.jiffies_per_second=100;

  /* Befehlszeile untersuchen */
  for (c = 1; argc > c; ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') printusage();
      switch (argv[c][1]) {
        case 's': lps.mode_seconds = 1; break;
        case 'c': lps.mode_count = 1; break;
        case 't': lps.mode_thread = 1; break;
        case 'u': lps.mode_user = 1; break;
        case 'U': lps.show_http_user_agent = 1; break;
        case 'L': lps.mode_limit = lps.mode_user = 1;
                  ++c;
                  if (!str2uint32(argv[c], &val32)) printusage();
                  lps.proclimit = val32;
                  break;
        default: printusage();
      }
    } else {
      if (!str2uint32(argv[c], &val32)) _exit(EX_USAGE);
      lps.uid_min = val32;
      ++c;
      if (argc > c) {
        if (!str2uint32(argv[c], &val32)) _exit(EX_USAGE);
        lps.uid_max = val32;
      }
      break;
    }
  }

  /* Ausgabepuffer initialisieren */
  p = wbufspace;
  buffer wbuf = { p, p, p+WRITEBUFSIZE, 1 };
  lps.wbuf = &wbuf;

  /* ermittle Uptime und berechne HZ-Wert */
  p = filebuf;
  if ((!lps.mode_count) && (!lps.mode_user)) { 

    rf = open("/proc/self/stat", O_RDONLY);
    if (rf == -1) _exit(EX_OSFILE);
    c = read(rf, p, BUFSIZE);
    close(rf);
    if (c == -1) _exit(EX_OSFILE);
    p[c] = '\0';
    /* starttime */
    str2uint64(field(p, 21), &val64); /* no error checks */
    lps.start_time = val64;

    rf = open("/proc/uptime", O_RDONLY);
    if (rf == -1) _exit(EX_OSFILE);
    c = read(rf, p, BUFSIZE);
    close(rf);
    if (c == -1) _exit(EX_OSFILE);
    p[c] = '\0';
    /* uptime */
    str2uint64(p, &val64); /* no error checks */
    lps.jiffies_per_second = ((int)(lps.start_time / val64)) + 1;
    /**
     * 10 S/390 (sometimes)
     * 20 User-mode Linux
     * 32 ia64 emulator
     * 64 StrongARM /Shark
     * 100 normal Linux
     * 128 MIPS, ARM
     * 1000 ARM
     * 1024 Alpha, ia64
     * 1200 Alpha
     */
    /* Rundung auf 2 Hertz */
    lps.jiffies_per_second = ((int)(lps.jiffies_per_second / 2)) * 2;
  
#ifdef OBSOLETE_STUFF
    lps.pagesize = sysconf(_SC_PAGESIZE);
    if (lps.pagesize < 1) lps.pagesize = 4096; /* Ausweichmoeglichkeit bei Fehler */
#else
#if ( defined(__alpha__) || defined(__sparc__) )
    lps.pagesize = 8192;
#else
    lps.pagesize = 4096;
#endif
#endif

  }

  /* oeffne /etc/passwd zum Konvertieren UID => Username */
  p = (char *) map_passwd_file(&lps.pwdbuf, &lps.pwdlen);
  if (p == (char *) -1) _exit(EX_OSFILE);

  /* gehe durch das /proc Verzeichnis und alle /proc/<PID>/stat Dateien */
  procdirp = opendir("/proc");
  if (procdirp == NULL) _exit(EX_OSFILE);
  for(;(procdirep = readdir(procdirp)) != NULL;) {

    /* PID? */
    if ((*procdirep->d_name < '0') || (*procdirep->d_name > '9')) continue;

    if (lps.mode_thread) {
      print_pid_task_lines(&lps, procdirep->d_name);
    } else {
      print_pid_line(&lps, "/proc", procdirep->d_name);
    }
  }


  if (lps.mode_count) {
    --lps.mode_count;
    if (lps.mode_user) buffer_put(lps.wbuf, "proccount=", 10);
    buffer_putuint64(lps.wbuf, lps.mode_count);
    buffer_putc(lps.wbuf, '\n');
  }

  if (lps.mode_user) {

    /* gehe durch unsere Liste und ermittle die drei Benutzer mit den meisten Prozessen */
    lps.max1uid = lps.max1cnt = lps.max2uid = lps.max2cnt = lps.max3uid = lps.max3cnt = 0;
    c = 0;
    for(;;) {
      if (c >= MAX_USERS_IN_PASSWD) break;
      if (!lps.uidp[c]) break;
      if (lps.uidcnt[c] > lps.max1cnt) {
        lps.max3uid = lps.max2uid; lps.max3cnt = lps.max2cnt;
        lps.max2uid = lps.max1uid; lps.max2cnt = lps.max1cnt;
        lps.max1uid = lps.uidp[c]; lps.max1cnt = lps.uidcnt[c];
      } else {
        if (lps.uidcnt[c] > lps.max2cnt) {
          lps.max3uid = lps.max2uid; lps.max3cnt = lps.max2cnt;
          lps.max2uid = lps.uidp[c]; lps.max2cnt = lps.uidcnt[c];
        } else {
          if (lps.uidcnt[c] > lps.max3cnt) { lps.max3uid = lps.uidp[c]; lps.max3cnt = lps.uidcnt[c]; }
        }
      }
      ++c;
    }
  
    if (lps.mode_limit) {
      if (lps.max1cnt >= lps.proclimit) {
        s = (char *) get_name_from_passwd(lps.pwdbuf, lps.pwdlen, lps.max1uid, &pw_name_len);
        if (s) {
          buffer_put(lps.wbuf, s, pw_name_len);
        } else {
          buffer_putuint64(lps.wbuf, lps.max1uid);
        }
        buffer_putc(lps.wbuf, '\n');
      }
      goto output_flush;
    }

    if (lps.max1cnt) {
      s = (char *) get_name_from_passwd(lps.pwdbuf, lps.pwdlen, lps.max1uid, &pw_name_len);
      if (s) {
        buffer_put(lps.wbuf, s, pw_name_len);
      } else {
        buffer_putuint64(lps.wbuf, lps.max1uid);
      }
      buffer_putc(lps.wbuf, '=');
      buffer_putuint64(lps.wbuf, lps.max1cnt);
      buffer_putc(lps.wbuf, '\n');
      if (lps.max2cnt) {
        s = (char *) get_name_from_passwd(lps.pwdbuf, lps.pwdlen, lps.max2uid, &pw_name_len);
        if (s) {
          buffer_put(lps.wbuf, s, pw_name_len);
        } else {
          buffer_putuint64(lps.wbuf, lps.max2uid);
        }
        buffer_putc(lps.wbuf, '=');
        buffer_putuint64(lps.wbuf, lps.max2cnt);
        buffer_putc(lps.wbuf, '\n');
        if (lps.max3cnt) {
          s = (char *) get_name_from_passwd(lps.pwdbuf, lps.pwdlen, lps.max3uid, &pw_name_len);
          if (s) {
            buffer_put(lps.wbuf, s, pw_name_len);
          } else {
            buffer_putuint64(lps.wbuf, lps.max3uid);
          }
          buffer_putc(lps.wbuf, '=');
          buffer_putuint64(lps.wbuf, lps.max3cnt);
          buffer_putc(lps.wbuf, '\n');
        }
      }
    }
  }

  output_flush:
  buffer_flush(lps.wbuf);
  closedir(procdirp);
  munmap((void *) lps.pwdbuf, lps.pwdlen);
  _exit(0);

}

