/**
 * mboxscrub.c
 * delete old mails from mbox file
 * Frank Bergmann 10/2016
 * $Id: mboxscrub.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif
#include <features.h>

#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>
#include <limits.h>
#include <sys/wait.h>
#include <time.h>
#include <sysexits.h>

#include "../lib/libfwb.h"

#define BUFSIZE 4096
#define MIN_SEARCH_INTERVAL_BYTES 500000

/* avoid malloc/mmap on 32 bit and diet and tiny systems, better use static buf */
#if __WORDSIZE == 32
#define MAX_MAIL_BUF_SIZE 2147483647
#else
#define MAX_MAIL_BUF_SIZE 4294967295
#endif
static char mail_buf[MAX_MAIL_BUF_SIZE];

#define MBOX_START_OF_MAIL_STRING "\n\nFrom "

static const char usage[] =
  "usage: mboxscrub [-D DAYS_TO_KEEP] /var/spool/mail/USER-MBOX\n";

static void printusage() { write1(usage, sizeof usage - 1); exit(EX_USAGE); }

static const char *days[] = {
  "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
static const char *months[] = {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static const char *is_day(register const char *s, uint8_t *pn) {
  register uint8_t c = 0;
  if (!s) return 0;
  if (*s == 0) return 0;
  if (*s == ' ') ++s;
  if (*s == 0) return 0;
  for (;;) {
    if ((*s == days[c][0]) && (s[1] == days[c][1]) && (s[2] == days[c][2]) && (s[3] <= ' ')) {
      if (pn) *pn = c;
      return s+3;
    }
    ++c;
    if (c == 7) break;
  }
  return 0;
}

static const char *is_month(register const char *s, uint8_t *pn) {
  register uint8_t c = 0;
  if (!s) return 0;
  if (*s == 0) return 0;
  if (*s == ' ') ++s;
  if (*s == 0) return 0;
  for (;;) {
    if ((*s == months[c][0]) && (s[1] == months[c][1]) && (s[2] == months[c][2]) && (s[3] <= ' ')) {
      if (pn) *pn = c;
      return s+3;
    }
    ++c;
    if (c == 12) break;
  }
  return 0;
}

static const char *str2uint8(register const char *s, uint8_t *val) {
  register char c;
  uint8_t i;

  if (!s) return 0;
  c = *s;
  if ((c < '0') || (c > '9')) return 0;
  ++s;
  i = c - '0'; /* 1st digit */
  c = *s;
  if ((c < '0') || (c > '9')) {
    if (val) *val = (uint8_t) i; return s;
  } else {
    i = (i * 10) + c - '0'; ++s;
  } /* 2nd digit */
  c = *s;
  if ((c < '0') || (c > '9')) {
    if (val) *val = (uint8_t) i; return s;
  } else {
    i = (i * 10) + c - '0'; ++s;
  } /* 3rd digit */
  if (val) *val = (uint8_t) i;
  return s;
}

static const char *str2uint16(register const char *s, uint16_t *val) {
  register char c;
  uint16_t i;

  if (!s) return 0;
  c = *s;
  if ((c < '0') || (c > '9')) return 0;
  ++s;
  i = c - '0'; /* 1st digit */
  c = *s;
  if ((c < '0') || (c > '9')) {
    if (val) *val = (uint16_t) i; return s;
  } else {
    i = (i * 10) + c - '0'; ++s;
  } /* 2nd digit */
  c = *s;
  if ((c < '0') || (c > '9')) {
    if (val) *val = (uint16_t) i; return s;
  } else {
    i = (i * 10) + c - '0'; ++s;
  } /* 3rd digit */
  c = *s;
  if ((c < '0') || (c > '9')) {
    if (val) *val = (uint16_t) i; return s;
  } else {
    i = (i * 10) + c - '0'; ++s;
  } /* 4th digit */
  if (val) *val = (uint16_t) i;
  return s;
}

static time_t str2time(register const char *s) {
  register const char *p;
  register const char *r;
  uint8_t mon;
  uint8_t mday;
  uint8_t hour;
  uint8_t minute;
  uint8_t sec;
  uint16_t year;
  struct tm tm;
  time_t t;

  if (!s) return -1;
  p = s;
  for (;;) { if (!p) return -1; if (*p != ' ') break; ++p; }
  r = p;
  p = is_day(p, 0);
  if (!p) p = r;
  p = is_month(p, &mon);
  for (;;) { if (!p) return -1; if (*p != ' ') break; ++p; }
  r = str2uint8(p, &mday);
  if (!r) return -1;
  if (mday > 31) return -1;
  p = r;
  for (;;) { if (!p) return -1; if (*p != ' ') break; ++p; }
  r = str2uint8(p, &hour);
  if (!r) return -1;
  if (hour > 24) return -1;
  p = r;
  if (*p != ':') return -1;
  ++p; if (!p) return -1;
  r = str2uint8(p, &minute);
  if (!r) return -1;
  if (minute > 59) return -1;
  p = r;
  if (*p != ':') return -1;
  ++p; if (!p) return -1;
  r = str2uint8(p, &sec);
  if (!r) return -1;
  if (sec > 59) return -1;
  p = r;
  for (;;) { if (!p) return -1; if (*p != ' ') break; ++p; }
  r = str2uint16(p, &year);
  if (!r) return -1;
  if (year > 2099) return -1;
  if (year < 1970) return -1;
  tm.tm_sec = sec;
  tm.tm_min = minute;
  tm.tm_hour = hour;
  tm.tm_mday = mday;
  tm.tm_mon = mon;
  tm.tm_year = year - 1900;
  tm.tm_isdst = -1;
  t = mktime(&tm);
  /* printf("mon=%d mday=%d hour=%d minute=%d sec=%d year=%d t=%ld\n", mon+1, mday, hour, minute, sec, year, t); */
  return t;
}

static time_t headerfrom2time(register const char *s) {
  register const unsigned char *p;

  if (!s) return -1;
  p = (const unsigned char *) s;
  for (;;) { if (!p) return -1; if (*p > ' ') break; ++p; }
  if (strncmp((const char *) p, "From ", 5) != 0) return -1;
  p += 5;
  for (;;) { if (!p) return -1; if (*p == ' ') break; ++p; }
  for (;;) { if (!p) return -1; if (*p > ' ') break; ++p; }
  return str2time((const char *) p);
}

/**
 * inspired by dietlibc strstr.c
 */
static const char *strstrl(const char *haystack, const char *needle, const char *haystack_limit) {
  size_t nl=strlen(needle);
  size_t hl=haystack_limit-haystack-1;
  size_t i;
  if (nl>hl) return 0;
  if ((haystack_limit-nl)<haystack) return 0;
  for (i=hl-nl+1; __likely(i); --i) {
    if (*haystack==*needle) {
      if (!memcmp(haystack,needle,nl)) {
        return (char*)haystack;
      }
    }
    ++haystack;
  }
  return 0;
}

static const char *next_mail(register const char *p, const char *p_limit, time_t *mail_ts) {
  time_t ts_tmp;

  for (;;) {
    p = strstrl(p, MBOX_START_OF_MAIL_STRING, p_limit);
    if (!p) return 0;
    p += 2; /* 2 LFs of MBOX_START_OF_MAIL_STRING */
    ts_tmp = headerfrom2time(p);
    if (ts_tmp != ((time_t) -1)) {
      *mail_ts = ts_tmp;
      break;
    }
  }
  return p;
}


int main(int argc,char *const argv[]) {
  register const char *p;
  register const char *r;
  const char *search_start;
  const char *search_end;
  const char *mbox_name;
  uint32_t days_to_keep;
  off_t mail_buf_len;
  const char *mail_buf_end;
  int tmpint;
  char line[BUFSIZE+1];
  int has_internal_folder_marker = 0;
  const char *first_mail = 0;
  time_t ts_now;
  time_t ts_keep;
  time_t ts_tmp;
  const char *last_mail_too_old;
  const char *last_mail_too_new;
  int internal_mail_len;
  uint64_t keep_mails_len;
  int mbox_fd;
  ssize_t bytes_written;

  mbox_name = 0;
  days_to_keep = 40;
  if (argc < 2) printusage();
  tmpint = 0;
  for(tmpint=1; tmpint<argc; ++tmpint) {
    if (*argv[tmpint] == '-') {
      if (argv[tmpint][2] != '\0') printusage();
      switch (argv[tmpint][1]) {
        case 'h': printusage(); break;
        case 'D': ++tmpint; if (!str2uint32(argv[tmpint], &days_to_keep)) printusage(); break;
        default: printusage();
      }
    } else {
      mbox_name = argv[tmpint];
    }
  }
  if (mbox_name == 0) printusage();

  //p = (char *) map_file(mbox_name, &mail_buf, &mail_buf_len);
  //if (p == (char *) -1) goto map_error_exit;

  mbox_fd = open(mbox_name, O_RDONLY);
  if (mbox_fd == -1) goto openread_error_exit;
  mail_buf_len = read(mbox_fd, mail_buf, MAX_MAIL_BUF_SIZE);
  if (mail_buf_len == -1) goto read_error_exit;
  close(mbox_fd);
  p = mail_buf;
  first_mail = p;
  mail_buf_end = p + mail_buf_len;
  last_mail_too_old = 0;
  last_mail_too_new = 0;

  /* check for mail folder internal data mail */
  strncpy(line, p, 600); line[599] = 0;
  if (strstr(line, " Mail System Internal Data")) {
    has_internal_folder_marker = 1;
  } else {
    if (strstr(line, "  DON'T DELETE THIS MESSAGE")) {
      has_internal_folder_marker = 1;
    } else {
      if (strstr(line, " FOLDER INTERNAL DATA")) {
        has_internal_folder_marker = 1;
      }
    }
  }

  if (has_internal_folder_marker) {
    p = next_mail(p, mail_buf_end, &ts_tmp);
    if (!p) goto normal_exit; /* empty mailbox */
    first_mail = p;
  } else {
    ts_tmp = headerfrom2time(p);
    if (ts_tmp == ((time_t) -1)) {
      /* TSNH: this should not happen, first email starts at 0 w/o TS */
      p = next_mail(p, mail_buf_end, &ts_tmp);
      if (!p) goto normal_exit; /* empty mailbox */
      first_mail = p;
    }
  }

  ts_now = time(0);
  if (ts_now == ((time_t) -1)) goto time_error_exit;
  ts_keep = ts_now - (days_to_keep * 86400);
  /* check first email */
  if (ts_tmp >= ts_keep) goto normal_exit; /* all mails too new => no mbox change */
  last_mail_too_old = p;

  /* now search mbox */
  search_start = p;
  search_end = mail_buf_end;
  for(;;) {
    p = search_start + ((search_end - search_start) / 2);
    if ((p - search_start) < MIN_SEARCH_INTERVAL_BYTES) p = search_start;
    r = p;
    if (search_start == search_end) break;
    p = next_mail(p, mail_buf_end, &ts_tmp);
    if ((!p) || (ts_tmp >= ts_keep)) {
      /* iteration: interval before */
      if (p) {
        last_mail_too_new = p;
      }
      if (search_end == r) break;
      search_end = r;
    } else {
      /* iteration: next interval */
      last_mail_too_old = p;
      search_start = p;
    }
  }

  /* should have already been catched before when checking first mail:
   * all mails too new => no mbox change */
  if (!last_mail_too_old) goto normal_exit;

  internal_mail_len = first_mail - mail_buf;
  if (!last_mail_too_new) {
    /* keep only folder internal mail - if it does exist */
    keep_mails_len = 0;
  } else {
    keep_mails_len = mail_buf_end - last_mail_too_new;
  }

  /* now write it out */
  mbox_fd = open(mbox_name, O_WRONLY|O_TRUNC);
  if (mbox_fd == -1) goto openwrite_error_exit;
  if (internal_mail_len) {
    bytes_written = write(mbox_fd, mail_buf, internal_mail_len);
    if (bytes_written != internal_mail_len) goto write_error_exit;
  }
  bytes_written = write(mbox_fd, last_mail_too_new, keep_mails_len);
  if (bytes_written != keep_mails_len) goto write_error_exit;
  close(mbox_fd);

  normal_exit:
  _exit(EX_OK);

  read_error_exit:
  close(mbox_fd);
  openread_error_exit:
  _exit(EX_NOINPUT);

  time_error_exit:
  _exit(EX_OSERR);

  openwrite_error_exit:
  _exit(EX_CANTCREAT);

  write_error_exit:
  close(mbox_fd);
  _exit(EX_IOERR);
}

