/**
 * sfgrep-bk.c
 * simple-fixed-grep or string-fast-grep
 * works similar to grep but with strings not patterns
 * Compile with:
 * diet gcc -Os -Wall -o sfgrep sfgrep.c && sstrip sfgrep
 * $Id: sfgrep-bk.c 691 2017-06-04 23:12:04Z fwbergm $
 */

#if __GNUC__ > 5
#pragma GCC diagnostic ignored "-Wmisleading-indentation"
#endif

#define _FILE_OFFSET_BITS 64
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <features.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>

#include <string.h>

/* gcc >= 3 is needed */
#ifndef __expect
#define __expect(foo,bar) __builtin_expect((long)(foo),bar)
#endif
#ifndef __likely
/* macros taken from Linux kernel */
#define __likely(foo) __expect((foo),1)
#define __unlikely(foo) __expect((foo),0)
#endif

#ifndef write1
#include <sys/syscall.h>
#define write1(b,c) (syscall(__NR_write, 1, b, c))
#endif
#ifndef write2
#include <sys/syscall.h>
#define write2(b,c) (syscall(__NR_write, 2, b, c))
#endif

#define READBUFSIZE 1048576
#define WRITEBUFSIZE 8192

static const char msg[]="missing arg(s)\nusage: sfgrep-bk [-q] [-1] [-h] [-H] [-c] [-l] string [file(s)]\n";
static const char warning_open[]=" * warning, could not be opened\n";
static const char warning_read[]=" * warning, could not be read\n";

typedef struct buffer {
  char *s;
  char *c;
  char *t;
  int fd;
} buffer;

typedef struct grep_parms {
  char *search_string;
  buffer *write_buffer;
  char *filename;
  int fnlen;
  int mode_withfilename;
  int mode_withmatch;
  int mode_firstmatch;
  int mode_count;
  int mode_silent;
  uint32_t matchcount;
} grep_parms;

static int32_t buffer_flush(buffer *b) {
  register char *p;
  ssize_t bytes;

  p = b->s;
  if (p == b->c) return 0;
  for (;;) {
    bytes = write(b->fd, p, b->c - p);
    if (bytes == -1) return 1;
    p += bytes;
    if (p == b->c) break;
  }
  b->c = b->s;
  return 0;
}

static int32_t buffer_flush50(buffer *b) {
  register char *p;
  register char *s;
  int32_t bytes;

  s = b->s;
  if (b->c == s) return 0;
  p = s + ((b->t - s) / 2);
  if (b->c < p) return 0;
  p = b->c;
  for (;;) {
    --p;
    if (p <= s) return 0;
    if (*p == '\n') break;
  }
  ++p;
  bytes = write(b->fd, s, p - s);
  if (bytes <= 0) return bytes;
  memcpy(s, s + bytes, b->c - s - bytes);
  b->c -= bytes;
  return bytes;
}

static int32_t buffer_putc(buffer *b, register const char a) {
  register char *p;
  register char *s;

  p = b->c;
  if (p < b->t) goto put;

  s = b->s;
  if (write(b->fd, s, p - s) < 0) return -1;
  p = s;

  put:
  *p = a; ++p;
  b->c = p;
  return buffer_flush50(b);
}

static int32_t buffer_put(buffer *b, register const char *a, register uint32_t l) {
  register char *p;
  register char *s;
  register const char *t;

  if (!l) return 0;
  p = b->c;
  t = b->t;
  if ((p+l) <= t) goto put;

  s = b->s;
  if (write(b->fd, s, p-s) < 0) return -1;
  p = s;
  if ((p+l) > t) {
    b->c = p;
    return write(b->fd, a, l);
  }

  put:
  t = a + l;
  for(;;) {
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
    *p = *a; ++a; ++p; if (a >= t) break;
  }
  b->c = p;
  return buffer_flush50(b);
}

static int32_t buffer_putlongnl(buffer *b, uint32_t num) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;
  char buf[14];

  i=num; o=0;
  pbuf=buf;
  for(d=1000000000;d>1;d/=10) {
    c=i/d;
    if((c>0)||(o)) {
      *pbuf=c+'0';++pbuf;
      i=i-(c*d);
      o=1;
    }
  }
  *pbuf=i+'0'; *++pbuf='\n'; ++pbuf;
  return buffer_put(b, buf, pbuf-buf);
}

static void my_exit(int retval, const char *message) {
  if (message && (*message != '\0')) write2(message, strlen(message));
  _exit(retval);
}




static uint32_t grep_filebuffer(register char *p, register char *t, grep_parms *grep) {
  register char *line;
  register char *s;
  register char *search_string;
  register int mode_withfilename;
  register int mode_silent;
  buffer *write_buffer;

  write_buffer = grep->write_buffer;
  mode_withfilename = grep->mode_withfilename;
  mode_silent = grep->mode_silent;
  search_string = grep->search_string;
  s = search_string;

  for(;;) {

    next_line:
    line = p;

    for(;;) {

      if (__unlikely(*p == '\n')) break;
      if (__unlikely(p >= t)) return 0;

      string_cmp:
      if (__unlikely(*p == *s)) {
        next_char:
        ++p; ++s;
        if (__unlikely(*s == '\0')) {
          /* match */
          s = search_string;
          for(;;) {
            /* don't use: "if (*p == '\n') break; p++;" */
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
          }
          grep->matchcount += 1;
          if (!mode_silent && mode_withfilename && !grep->mode_count) {
            buffer_put(write_buffer, grep->filename, grep->fnlen);
            if (grep->mode_firstmatch && !grep->mode_withmatch) {
              buffer_putc(write_buffer, '\n');
              return 1;
            }
            buffer_putc(write_buffer, ':');
          }
          if (grep->mode_firstmatch && !grep->mode_withmatch) return 1;
          if (!grep->mode_count && !mode_silent) { buffer_put(write_buffer, line, p-line); }
          if (grep->mode_firstmatch) return 1;
          if (p >= t) return 0;
          goto next_line;
        } /* *s == '\0' */
        if (*p == '\n') { s = search_string; break; }
        if (p >= t) return 0;
        if (*p == *s) goto next_char;
        p -= s - search_string - 1;
        s = search_string;
        goto string_cmp;
      } /* *p == *s */

      ++p;
    }

    for(;;) {
      ++p;
      if (p >= t) return 0;
      if (*p != '\n') break;
    }

  }
}


static void grep_file(int fd, grep_parms *grep) {
  register char *p;
  register char *t;
  register char *n;
  char rbufspace[2*READBUFSIZE+1];
  register ssize_t rb;

  p = rbufspace;
  grep->matchcount = 0;
  for(;;) {
    rb = read(fd, p, READBUFSIZE);
    if (rb < 0) {
        if (grep->filename) write2(grep->filename,grep->fnlen);
        write2(warning_read,strlen(warning_read));
        return;
    }
    if ((rb == 0) && (p == rbufspace)) {
      goto allread;
    }
    if (rb == 0) {
      t = p - 1;
      if (*t != '\n') *++t = '\n';
      ++t;
    } else {
      t = p + rb;
    }
    p = rbufspace;
    /* find last \n */
    n = t;
    for(;;) {
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
    }
    if (n <= p) {
      n = t;
    } else {
      ++n;
    }
    if (grep_filebuffer(p, n, grep) && grep->mode_firstmatch) return;
    if (n != t) {
      for(;;) {
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
        *p = *n; ++p; ++n; if (n >= t) break;
      }
    } else {

      if (rb == 0) {
        allread:
        if (grep->mode_count && !(grep->mode_silent)) {
          if (grep->mode_withfilename) {
            buffer_put(grep->write_buffer, grep->filename, grep->fnlen);
            buffer_putc(grep->write_buffer, ':');
          }
          buffer_putlongnl(grep->write_buffer, grep->matchcount);
        }
        return;
      }

    }
  }
}


int main(int argc,char *const argv[]) {
  register char *p;
  register int c;
  char wbufspace[WRITEBUFSIZE+1];
  int fd;
  int file1;
  int sstring;
  int optsend;
  int retval;
  grep_parms grep;

  if (argc < 2) my_exit(127, msg);
  optsend = 0;
  retval = 1;
  grep.mode_firstmatch = 
    grep.mode_withmatch = 
    grep.mode_silent = 
    grep.mode_count = 0;
  sstring = file1 = grep.mode_withfilename = -1;
  for(c=1;c<argc;++c) {
    if (!optsend && (*argv[c] == '-')) {
      if (argv[c][2] != '\0') my_exit(127, msg);
      switch (argv[c][1]) {
        case 'h': grep.mode_withfilename = 0; break;
        case 'H': grep.mode_withfilename = 1; break;
        case 'c': grep.mode_count = 1; break;
        case 'l': grep.mode_firstmatch = 1; grep.mode_withmatch = 0; break;
        case '1': grep.mode_firstmatch = 1; grep.mode_withmatch = 1; break;
        case 'q': grep.mode_silent = 1; break;
        case '-': optsend = 1; break;
        default: break;
      }
    } else {
      if (sstring == -1) {
        sstring = c;
      } else {
        if (file1 == -1) {
          file1 = c;
        } else {
          if (grep.mode_withfilename == -1) grep.mode_withfilename = 1;
        }
      }
    }
  }
  if (sstring == -1) my_exit(127, msg);
  if (grep.mode_count && (grep.mode_withfilename == -1)) grep.mode_withfilename = 1;
  if (grep.mode_withfilename == -1) grep.mode_withfilename = 0;
  if (grep.mode_firstmatch) grep.mode_withfilename = 1;
  /* F*cking illogical GNU-grep compatibility fix: */
  if (grep.mode_firstmatch) grep.mode_count = 0;
  if (grep.mode_silent) {
    grep.mode_withfilename = 0;
    grep.mode_firstmatch = 1;
    grep.mode_count = 0;
  }

  p = wbufspace;
  buffer wbuf = { p, p, p+WRITEBUFSIZE, 1 };
  grep.write_buffer = &wbuf;
  grep.search_string = argv[sstring];

  if (file1 == -1) {
    grep.filename = NULL;
    grep.fnlen = 0;
    grep.mode_withfilename = 0;
    grep_file(0, &grep);
  } else {
    for(c=file1;c<argc;++c) {
      fd = open(argv[c], O_RDONLY);
      if (fd < 0) {
        write2(argv[c],strlen(argv[c]));
        write2(warning_open,strlen(warning_open));
      } else {
        grep.filename = argv[c];
        grep.fnlen = strlen(grep.filename);
        grep_file(fd, &grep);
        if (grep.matchcount) retval = 0;
        close(fd);
      }
    }
  }
  buffer_flush(&wbuf);
  if (grep.matchcount) retval = 0;
  _exit(retval);
}

