/**
 * $Id: show-cipher-preference.c 691 2017-06-04 23:12:04Z fwbergm $
 *
 * show-cipher-preference
 * listen on tcp port, wait for TLS Client Hello and display cipher preference
 * usage: show-cipher-preference 443
 *
 * Compile it with: diet gcc -O2 -Wall ... ; sstrip show-cipher-preference
 * License is GNU GPL v2.
 *
 * Frank Bergmann, http://www.tuxad.com
 *
 */

#include <sys/socket.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>
#include <time.h>

typedef struct st_ciphers {
  uint8_t value[2];
  const char *description;
} ciphers_t;

static ciphers_t ciphers[] = {
  {{0x00,0x00}, "TLS_NULL_WITH_NULL_NULL"},
  {{0x00,0x01}, "TLS_RSA_WITH_NULL_MD5 (NULL-MD5)"},
  {{0x00,0x02}, "TLS_RSA_WITH_NULL_SHA (NULL-SHA)"},
  {{0x00,0x03}, "TLS_RSA_EXPORT_WITH_RC4_40_MD5 (EXP-RC4-MD5)"},
  {{0x00,0x04}, "TLS_RSA_WITH_RC4_128_MD5 (RC4-MD5)"},
  {{0x00,0x05}, "TLS_RSA_WITH_RC4_128_SHA (RC4-SHA)"},
  {{0x00,0x06}, "TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5 (EXP-RC2-CBC-MD5)"},
  {{0x00,0x07}, "TLS_RSA_WITH_IDEA_CBC_SHA (IDEA-CBC-SHA)"},
  {{0x00,0x08}, "TLS_RSA_EXPORT_WITH_DES40_CBC_SHA (EXP-DES-CBC-SHA)"},
  {{0x00,0x09}, "TLS_RSA_WITH_DES_CBC_SHA (DES-CBC-SHA)"},
  {{0x00,0x0A}, "TLS_RSA_WITH_3DES_EDE_CBC_SHA (DES-CBC3-SHA)"},
  {{0x00,0x0B}, "TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA"},
  {{0x00,0x0C}, "TLS_DH_DSS_WITH_DES_CBC_SHA"},
  {{0x00,0x0D}, "TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x0E}, "TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA"},
  {{0x00,0x0F}, "TLS_DH_RSA_WITH_DES_CBC_SHA"},
  {{0x00,0x10}, "TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x11}, "TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA (EXP-EDH-DSS-DES-CBC-SHA)"},
  {{0x00,0x12}, "TLS_DHE_DSS_WITH_DES_CBC_SHA (EDH-DSS-CBC-SHA)"},
  {{0x00,0x13}, "TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA (DHE-DSS-DES-CBC3-SHA EDH-DSS-DES-CBC3-SHA)"},
  {{0x00,0x14}, "TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA (EXP-EDH-RSA-DES-CBC-SHA)"},
  {{0x00,0x15}, "TLS_DHE_RSA_WITH_DES_CBC_SHA (EDH-RSA-DES-CBC-SHA)"},
  {{0x00,0x16}, "TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA (DHE-RSA-DES-CBC3-SHA EDH-RSA-DES-CBC3-SHA)"},
  {{0x00,0x17}, "TLS_DH_anon_EXPORT_WITH_RC4_40_MD5 (EXP-ADH-RC4-MD5)"},
  {{0x00,0x18}, "TLS_DH_anon_WITH_RC4_128_MD5 (ADH-RC4-MD5)"},
  {{0x00,0x19}, "TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA (EXP-ADH-DES-CBC-SHA)"},
  {{0x00,0x1A}, "TLS_DH_anon_WITH_DES_CBC_SHA (ADH-DES-CBC-SHA)"},
  {{0x00,0x1B}, "TLS_DH_anon_WITH_3DES_EDE_CBC_SHA (ADH-DES-CBC3-SHA)"},
  {{0x00,0x1E}, "TLS_KRB5_WITH_DES_CBC_SHA"},
  {{0x00,0x1F}, "TLS_KRB5_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x20}, "TLS_KRB5_WITH_RC4_128_SHA"},
  {{0x00,0x21}, "TLS_KRB5_WITH_IDEA_CBC_SHA"},
  {{0x00,0x22}, "TLS_KRB5_WITH_DES_CBC_MD5"},
  {{0x00,0x23}, "TLS_KRB5_WITH_3DES_EDE_CBC_MD5"},
  {{0x00,0x24}, "TLS_KRB5_WITH_RC4_128_MD5"},
  {{0x00,0x25}, "TLS_KRB5_WITH_IDEA_CBC_MD5"},
  {{0x00,0x26}, "TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA"},
  {{0x00,0x27}, "TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA"},
  {{0x00,0x28}, "TLS_KRB5_EXPORT_WITH_RC4_40_SHA"},
  {{0x00,0x29}, "TLS_KRB5_EXPORT_WITH_DES_CBC_40_MD5"},
  {{0x00,0x2A}, "TLS_KRB5_EXPORT_WITH_RC2_CBC_40_MD5"},
  {{0x00,0x2B}, "TLS_KRB5_EXPORT_WITH_RC4_40_MD5"},
  {{0x00,0x2C}, "TLS_PSK_WITH_NULL_SHA"},
  {{0x00,0x2D}, "TLS_DHE_PSK_WITH_NULL_SHA"},
  {{0x00,0x2E}, "TLS_RSA_PSK_WITH_NULL_SHA"},
  {{0x00,0x2F}, "TLS_RSA_WITH_AES_128_CBC_SHA (AES128-SHA)"},
  {{0x00,0x30}, "TLS_DH_DSS_WITH_AES_128_CBC_SHA (DH-DSS-AES128-SHA)"},
  {{0x00,0x31}, "TLS_DH_RSA_WITH_AES_128_CBC_SHA (DH-RSA-AES128-SHA)"},
  {{0x00,0x32}, "TLS_DHE_DSS_WITH_AES_128_CBC_SHA (DHE-DSS-AES128-SHA)"},
  {{0x00,0x33}, "TLS_DHE_RSA_WITH_AES_128_CBC_SHA (DHE-RSA-AES128-SHA)"},
  {{0x00,0x34}, "TLS_DH_anon_WITH_AES_128_CBC_SHA (ADH-AES128-SHA)"},
  {{0x00,0x35}, "TLS_RSA_WITH_AES_256_CBC_SHA (AES256-SHA)"},
  {{0x00,0x36}, "TLS_DH_DSS_WITH_AES_256_CBC_SHA (DH-DSS-AES256-SHA)"},
  {{0x00,0x37}, "TLS_DH_RSA_WITH_AES_256_CBC_SHA (DH-RSA-AES256-SHA)"},
  {{0x00,0x38}, "TLS_DHE_DSS_WITH_AES_256_CBC_SHA (DHE-DSS-AES256-SHA)"},
  {{0x00,0x39}, "TLS_DHE_RSA_WITH_AES_256_CBC_SHA (DHE-RSA-AES256-SHA)"},
  {{0x00,0x3A}, "TLS_DH_anon_WITH_AES_256_CBC_SHA (ADH-AES256-SHA)"},
  {{0x00,0x3B}, "TLS_RSA_WITH_NULL_SHA256 (NULL-SHA256)"},
  {{0x00,0x3C}, "TLS_RSA_WITH_AES_128_CBC_SHA256 (AES128-SHA256)"},
  {{0x00,0x3D}, "TLS_RSA_WITH_AES_256_CBC_SHA256 (AES256-SHA256)"},
  {{0x00,0x3E}, "TLS_DH_DSS_WITH_AES_128_CBC_SHA256 (DH-DSS-AES128-SHA256)"},
  {{0x00,0x3F}, "TLS_DH_RSA_WITH_AES_128_CBC_SHA256 (DH-RSA-AES128-SHA256)"},
  {{0x00,0x40}, "TLS_DHE_DSS_WITH_AES_128_CBC_SHA256 (DHE-DSS-AES128-SHA256)"},
  {{0x00,0x41}, "TLS_RSA_WITH_CAMELLIA_128_CBC_SHA (CAMELLIA128-SHA)"},
  {{0x00,0x42}, "TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA (DH-DSS-CAMELLIA128-SHA)"},
  {{0x00,0x43}, "TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA (DH-RSA-CAMELLIA128-SHA)"},
  {{0x00,0x44}, "TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA (DHE-DSS-CAMELLIA128-SHA)"},
  {{0x00,0x45}, "TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA (DHE-RSA-CAMELLIA128-SHA)"},
  {{0x00,0x46}, "TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA (ADH-CAMELLIA128-SHA)"},
  {{0x00,0x67}, "TLS_DHE_RSA_WITH_AES_128_CBC_SHA256 (DHE-RSA-AES128-SHA256)"},
  {{0x00,0x68}, "TLS_DH_DSS_WITH_AES_256_CBC_SHA256 (DH-DSS-AES256-SHA256)"},
  {{0x00,0x69}, "TLS_DH_RSA_WITH_AES_256_CBC_SHA256 (DH-RSA-AES256-SHA256)"},
  {{0x00,0x6A}, "TLS_DHE_DSS_WITH_AES_256_CBC_SHA256 (DHE-DSS-AES256-SHA256)"},
  {{0x00,0x6B}, "TLS_DHE_RSA_WITH_AES_256_CBC_SHA256 (DHE-RSA-AES256-SHA256)"},
  {{0x00,0x6C}, "TLS_DH_anon_WITH_AES_128_CBC_SHA256 (ADH-AES128-SHA256)"},
  {{0x00,0x6D}, "TLS_DH_anon_WITH_AES_256_CBC_SHA256 (ADH-AES256-SHA256)"},
  {{0x00,0x84}, "TLS_RSA_WITH_CAMELLIA_256_CBC_SHA (CAMELLIA256-SHA)"},
  {{0x00,0x85}, "TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA (DH-DSS-CAMELLIA256-SHA)"},
  {{0x00,0x86}, "TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA (DH-RSA-CAMELLIA256-SHA)"},
  {{0x00,0x87}, "TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA (DHE-DSS-CAMELLIA256-SHA)"},
  {{0x00,0x88}, "TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA (DHE-RSA-CAMELLIA256-SHA)"},
  {{0x00,0x89}, "TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA (ADH-CAMELLIA256-SHA)"},
  {{0x00,0x8A}, "TLS_PSK_WITH_RC4_128_SHA"},
  {{0x00,0x8B}, "TLS_PSK_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x8C}, "TLS_PSK_WITH_AES_128_CBC_SHA"},
  {{0x00,0x8D}, "TLS_PSK_WITH_AES_256_CBC_SHA"},
  {{0x00,0x8E}, "TLS_DHE_PSK_WITH_RC4_128_SHA"},
  {{0x00,0x8F}, "TLS_DHE_PSK_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x90}, "TLS_DHE_PSK_WITH_AES_128_CBC_SHA"},
  {{0x00,0x91}, "TLS_DHE_PSK_WITH_AES_256_CBC_SHA"},
  {{0x00,0x92}, "TLS_RSA_PSK_WITH_RC4_128_SHA"},
  {{0x00,0x93}, "TLS_RSA_PSK_WITH_3DES_EDE_CBC_SHA"},
  {{0x00,0x94}, "TLS_RSA_PSK_WITH_AES_128_CBC_SHA"},
  {{0x00,0x95}, "TLS_RSA_PSK_WITH_AES_256_CBC_SHA"},
  {{0x00,0x96}, "TLS_RSA_WITH_SEED_CBC_SHA (SEED-SHA)"},
  {{0x00,0x97}, "TLS_DH_DSS_WITH_SEED_CBC_SHA (DH-DSS-SEED-SHA)"},
  {{0x00,0x98}, "TLS_DH_RSA_WITH_SEED_CBC_SHA (DH-RSA-SEED-SHA)"},
  {{0x00,0x99}, "TLS_DHE_DSS_WITH_SEED_CBC_SHA (DHE-DSS-SEED-SHA)"},
  {{0x00,0x9A}, "TLS_DHE_RSA_WITH_SEED_CBC_SHA (DHE-RSA-SEED-SHA)"},
  {{0x00,0x9B}, "TLS_DH_anon_WITH_SEED_CBC_SHA (ADH-SEED-SHA)"},
  {{0x00,0x9C}, "TLS_RSA_WITH_AES_128_GCM_SHA256 (AES128-GCM-SHA256)"},
  {{0x00,0x9D}, "TLS_RSA_WITH_AES_256_GCM_SHA384 (AES256-GCM-SHA384)"},
  {{0x00,0x9E}, "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256 (DHE-RSA-AES128-GCM-SHA256)"},
  {{0x00,0x9F}, "TLS_DHE_RSA_WITH_AES_256_GCM_SHA384 (DHE-RSA-AES256-GCM-SHA384)"},
  {{0x00,0xA0}, "TLS_DH_RSA_WITH_AES_128_GCM_SHA256 (DH-RSA-AES128-GCM-SHA256)"},
  {{0x00,0xA1}, "TLS_DH_RSA_WITH_AES_256_GCM_SHA384 (DH-RSA-AES256-GCM-SHA384)"},
  {{0x00,0xA2}, "TLS_DHE_DSS_WITH_AES_128_GCM_SHA256 (DHE-DSS-AES128-GCM-SHA256)"},
  {{0x00,0xA3}, "TLS_DHE_DSS_WITH_AES_256_GCM_SHA384 (DHE-DSS-AES256-GCM-SHA384)"},
  {{0x00,0xA4}, "TLS_DH_DSS_WITH_AES_128_GCM_SHA256 (DH-DSS-AES128-GCM-SHA256)"},
  {{0x00,0xA5}, "TLS_DH_DSS_WITH_AES_256_GCM_SHA384 (DH-DSS-AES256-GCM-SHA384)"},
  {{0x00,0xA6}, "TLS_DH_anon_WITH_AES_128_GCM_SHA256 (ADH-AES128-GCM-SHA256)"},
  {{0x00,0xA7}, "TLS_DH_anon_WITH_AES_256_GCM_SHA384 (ADH-AES256-GCM-SHA384)"},
  {{0x00,0xA8}, "TLS_PSK_WITH_AES_128_GCM_SHA256"},
  {{0x00,0xA9}, "TLS_PSK_WITH_AES_256_GCM_SHA384"},
  {{0x00,0xAA}, "TLS_DHE_PSK_WITH_AES_128_GCM_SHA256"},
  {{0x00,0xAB}, "TLS_DHE_PSK_WITH_AES_256_GCM_SHA384"},
  {{0x00,0xAC}, "TLS_RSA_PSK_WITH_AES_128_GCM_SHA256"},
  {{0x00,0xAD}, "TLS_RSA_PSK_WITH_AES_256_GCM_SHA384"},
  {{0x00,0xAE}, "TLS_PSK_WITH_AES_128_CBC_SHA256"},
  {{0x00,0xAF}, "TLS_PSK_WITH_AES_256_CBC_SHA384"},
  {{0x00,0xB0}, "TLS_PSK_WITH_NULL_SHA256"},
  {{0x00,0xB1}, "TLS_PSK_WITH_NULL_SHA384"},
  {{0x00,0xB2}, "TLS_DHE_PSK_WITH_AES_128_CBC_SHA256"},
  {{0x00,0xB3}, "TLS_DHE_PSK_WITH_AES_256_CBC_SHA384"},
  {{0x00,0xB4}, "TLS_DHE_PSK_WITH_NULL_SHA256"},
  {{0x00,0xB5}, "TLS_DHE_PSK_WITH_NULL_SHA384"},
  {{0x00,0xB6}, "TLS_RSA_PSK_WITH_AES_128_CBC_SHA256"},
  {{0x00,0xB7}, "TLS_RSA_PSK_WITH_AES_256_CBC_SHA384"},
  {{0x00,0xB8}, "TLS_RSA_PSK_WITH_NULL_SHA256"},
  {{0x00,0xB9}, "TLS_RSA_PSK_WITH_NULL_SHA384"},
  {{0x00,0xBA}, "TLS_RSA_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xBB}, "TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xBC}, "TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xBD}, "TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xBE}, "TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xBF}, "TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0x00,0xC0}, "TLS_RSA_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xC1}, "TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xC2}, "TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xC3}, "TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xC4}, "TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xC5}, "TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA256"},
  {{0x00,0xFF}, "TLS_EMPTY_RENEGOTIATION_INFO_SCSV"},
  {{0x56,0x00}, "TLS_FALLBACK_SCSV"},
  {{0xC0,0x01}, "TLS_ECDH_ECDSA_WITH_NULL_SHA"},
  {{0xC0,0x02}, "TLS_ECDH_ECDSA_WITH_RC4_128_SHA"},
  {{0xC0,0x03}, "TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x04}, "TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x05}, "TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x06}, "TLS_ECDHE_ECDSA_WITH_NULL_SHA (ECDHE-ECDSA-NULL-SHA)"},
  {{0xC0,0x07}, "TLS_ECDHE_ECDSA_WITH_RC4_128_SHA (ECDHE-ECDSA-RC4-SHA)"},
  {{0xC0,0x08}, "TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA (ECDHE-ECDSA-DES-CBC3-SHA)"},
  {{0xC0,0x09}, "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA (ECDHE-ECDSA-AES128-SHA)"},
  {{0xC0,0x0A}, "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA (ECDHE-ECDSA-AES256-SHA)"},
  {{0xC0,0x0B}, "TLS_ECDH_RSA_WITH_NULL_SHA"},
  {{0xC0,0x0C}, "TLS_ECDH_RSA_WITH_RC4_128_SHA"},
  {{0xC0,0x0D}, "TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x0E}, "TLS_ECDH_RSA_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x0F}, "TLS_ECDH_RSA_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x10}, "TLS_ECDHE_RSA_WITH_NULL_SHA (ECDHE-RSA-NULL-SHA)"},
  {{0xC0,0x11}, "TLS_ECDHE_RSA_WITH_RC4_128_SHA (ECDHE-RSA-RC4-SHA)"},
  {{0xC0,0x12}, "TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA (ECDHE-RSA-DES-CBC3-SHA)"},
  {{0xC0,0x13}, "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA (ECDHE-RSA-AES128-SHA)"},
  {{0xC0,0x14}, "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA (ECDHE-RSA-AES256-SHA)"},
  {{0xC0,0x15}, "TLS_ECDH_anon_WITH_NULL_SHA (AECDH-NULL-SHA)"},
  {{0xC0,0x16}, "TLS_ECDH_anon_WITH_RC4_128_SHA (AECDH-RC4-SHA)"},
  {{0xC0,0x17}, "TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA (AECDH-DES-CBC3-SHA)"},
  {{0xC0,0x18}, "TLS_ECDH_anon_WITH_AES_128_CBC_SHA (AECDH-AES128-SHA)"},
  {{0xC0,0x19}, "TLS_ECDH_anon_WITH_AES_256_CBC_SHA (AECDH-AES256-SHA)"},
  {{0xC0,0x1A}, "TLS_SRP_SHA_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x1B}, "TLS_SRP_SHA_RSA_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x1C}, "TLS_SRP_SHA_DSS_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x1D}, "TLS_SRP_SHA_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x1E}, "TLS_SRP_SHA_RSA_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x1F}, "TLS_SRP_SHA_DSS_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x20}, "TLS_SRP_SHA_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x21}, "TLS_SRP_SHA_RSA_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x22}, "TLS_SRP_SHA_DSS_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x23}, "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256 (ECDHE-ECDSA-AES128-SHA256)"},
  {{0xC0,0x24}, "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384 (ECDHE-ECDSA-AES256-SHA384)"},
  {{0xC0,0x25}, "TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256"},
  {{0xC0,0x26}, "TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384"},
  {{0xC0,0x27}, "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256 (ECDHE-RSA-AES128-SHA256)"},
  {{0xC0,0x28}, "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384 (ECDHE-RSA-AES256-SHA384)"},
  {{0xC0,0x29}, "TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256"},
  {{0xC0,0x2A}, "TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384"},
  {{0xC0,0x2B}, "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256 (ECDHE-ECDSA-AES128-GCM-SHA256)"},
  {{0xC0,0x2C}, "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384 (ECDHE-ECDSA-AES256-GCM-SHA384)"},
  {{0xC0,0x2D}, "TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256"},
  {{0xC0,0x2E}, "TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384"},
  {{0xC0,0x2F}, "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 (ECDHE-RSA-AES128-GCM-SHA256)"},
  {{0xC0,0x30}, "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384 (ECDHE-RSA-AES256-GCM-SHA384)"},
  {{0xC0,0x31}, "TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256"},
  {{0xC0,0x32}, "TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384"},
  {{0xC0,0x33}, "TLS_ECDHE_PSK_WITH_RC4_128_SHA"},
  {{0xC0,0x34}, "TLS_ECDHE_PSK_WITH_3DES_EDE_CBC_SHA"},
  {{0xC0,0x35}, "TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA"},
  {{0xC0,0x36}, "TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA"},
  {{0xC0,0x37}, "TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA256"},
  {{0xC0,0x38}, "TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA384"},
  {{0xC0,0x39}, "TLS_ECDHE_PSK_WITH_NULL_SHA"},
  {{0xC0,0x3A}, "TLS_ECDHE_PSK_WITH_NULL_SHA256"},
  {{0xC0,0x3B}, "TLS_ECDHE_PSK_WITH_NULL_SHA384"},
  {{0xC0,0x3C}, "TLS_RSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x3D}, "TLS_RSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x3E}, "TLS_DH_DSS_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x3F}, "TLS_DH_DSS_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x40}, "TLS_DH_RSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x41}, "TLS_DH_RSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x42}, "TLS_DHE_DSS_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x43}, "TLS_DHE_DSS_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x44}, "TLS_DHE_RSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x45}, "TLS_DHE_RSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x46}, "TLS_DH_anon_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x47}, "TLS_DH_anon_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x48}, "TLS_ECDHE_ECDSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x49}, "TLS_ECDHE_ECDSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x4A}, "TLS_ECDH_ECDSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x4B}, "TLS_ECDH_ECDSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x4C}, "TLS_ECDHE_RSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x4D}, "TLS_ECDHE_RSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x4E}, "TLS_ECDH_RSA_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x4F}, "TLS_ECDH_RSA_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x50}, "TLS_RSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x51}, "TLS_RSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x52}, "TLS_DHE_RSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x53}, "TLS_DHE_RSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x54}, "TLS_DH_RSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x55}, "TLS_DH_RSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x56}, "TLS_DHE_DSS_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x57}, "TLS_DHE_DSS_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x58}, "TLS_DH_DSS_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x59}, "TLS_DH_DSS_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x5A}, "TLS_DH_anon_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x5B}, "TLS_DH_anon_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x5C}, "TLS_ECDHE_ECDSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x5D}, "TLS_ECDHE_ECDSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x5E}, "TLS_ECDH_ECDSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x5F}, "TLS_ECDH_ECDSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x60}, "TLS_ECDHE_RSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x61}, "TLS_ECDHE_RSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x62}, "TLS_ECDH_RSA_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x63}, "TLS_ECDH_RSA_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x64}, "TLS_PSK_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x65}, "TLS_PSK_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x66}, "TLS_DHE_PSK_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x67}, "TLS_DHE_PSK_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x68}, "TLS_RSA_PSK_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x69}, "TLS_RSA_PSK_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x6A}, "TLS_PSK_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x6B}, "TLS_PSK_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x6C}, "TLS_DHE_PSK_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x6D}, "TLS_DHE_PSK_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x6E}, "TLS_RSA_PSK_WITH_ARIA_128_GCM_SHA256"},
  {{0xC0,0x6F}, "TLS_RSA_PSK_WITH_ARIA_256_GCM_SHA384"},
  {{0xC0,0x70}, "TLS_ECDHE_PSK_WITH_ARIA_128_CBC_SHA256"},
  {{0xC0,0x71}, "TLS_ECDHE_PSK_WITH_ARIA_256_CBC_SHA384"},
  {{0xC0,0x72}, "TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256 (ECDHE-ECDSA-CAMELLIA128-SHA256)"},
  {{0xC0,0x73}, "TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_CBC_SHA384 (ECDHE-ECDSA-CAMELLIA256-SHA384)"},
  {{0xC0,0x74}, "TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x75}, "TLS_ECDH_ECDSA_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x76}, "TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256 (ECDHE-RSA-CAMELLIA128-SHA256)"},
  {{0xC0,0x77}, "TLS_ECDHE_RSA_WITH_CAMELLIA_256_CBC_SHA384 (ECDHE-RSA-CAMELLIA256-SHA384)"},
  {{0xC0,0x78}, "TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x79}, "TLS_ECDH_RSA_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x7A}, "TLS_RSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x7B}, "TLS_RSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x7C}, "TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x7D}, "TLS_DHE_RSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x7E}, "TLS_DH_RSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x7F}, "TLS_DH_RSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x80}, "TLS_DHE_DSS_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x81}, "TLS_DHE_DSS_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x82}, "TLS_DH_DSS_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x83}, "TLS_DH_DSS_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x84}, "TLS_DH_anon_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x85}, "TLS_DH_anon_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x86}, "TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x87}, "TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x88}, "TLS_ECDH_ECDSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x89}, "TLS_ECDH_ECDSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x8A}, "TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x8B}, "TLS_ECDHE_RSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x8C}, "TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x8D}, "TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x8E}, "TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x8F}, "TLS_PSK_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x90}, "TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x91}, "TLS_DHE_PSK_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x92}, "TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256"},
  {{0xC0,0x93}, "TLS_RSA_PSK_WITH_CAMELLIA_256_GCM_SHA384"},
  {{0xC0,0x94}, "TLS_PSK_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x95}, "TLS_PSK_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x96}, "TLS_DHE_PSK_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x97}, "TLS_DHE_PSK_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x98}, "TLS_RSA_PSK_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x99}, "TLS_RSA_PSK_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x9A}, "TLS_ECDHE_PSK_WITH_CAMELLIA_128_CBC_SHA256"},
  {{0xC0,0x9B}, "TLS_ECDHE_PSK_WITH_CAMELLIA_256_CBC_SHA384"},
  {{0xC0,0x9C}, "TLS_RSA_WITH_AES_128_CCM"},
  {{0xC0,0x9D}, "TLS_RSA_WITH_AES_256_CCM"},
  {{0xC0,0x9E}, "TLS_DHE_RSA_WITH_AES_128_CCM"},
  {{0xC0,0x9F}, "TLS_DHE_RSA_WITH_AES_256_CCM"},
  {{0xC0,0xA0}, "TLS_RSA_WITH_AES_128_CCM_8"},
  {{0xC0,0xA1}, "TLS_RSA_WITH_AES_256_CCM_8"},
  {{0xC0,0xA2}, "TLS_DHE_RSA_WITH_AES_128_CCM_8"},
  {{0xC0,0xA3}, "TLS_DHE_RSA_WITH_AES_256_CCM_8"},
  {{0xC0,0xA4}, "TLS_PSK_WITH_AES_128_CCM"},
  {{0xC0,0xA5}, "TLS_PSK_WITH_AES_256_CCM"},
  {{0xC0,0xA6}, "TLS_DHE_PSK_WITH_AES_128_CCM"},
  {{0xC0,0xA7}, "TLS_DHE_PSK_WITH_AES_256_CCM"},
  {{0xC0,0xA8}, "TLS_PSK_WITH_AES_128_CCM_8"},
  {{0xC0,0xA9}, "TLS_PSK_WITH_AES_256_CCM_8"},
  {{0xC0,0xAA}, "TLS_PSK_DHE_WITH_AES_128_CCM_8"},
  {{0xC0,0xAB}, "TLS_PSK_DHE_WITH_AES_256_CCM_8"},
  {{0xC0,0xAC}, "TLS_ECDHE_ECDSA_WITH_AES_128_CCM"},
  {{0xC0,0xAD}, "TLS_ECDHE_ECDSA_WITH_AES_256_CCM"},
  {{0xC0,0xAE}, "TLS_ECDHE_ECDSA_WITH_AES_128_CCM_8"},
  {{0xC0,0xAF}, "TLS_ECDHE_ECDSA_WITH_AES_256_CCM_8"},
  {{0xCC,0xA8}, "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256 (ECDHE-RSA-CHACHA20-POLY1305)"},
  {{0xCC,0xA9}, "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256 (ECDHE-ECDSA-CHACHA20-POLY1305)"},
  {{0xCC,0xAA}, "TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256 (DHE-RSA-CHACHA20-POLY1305)"},
  {{0xCC,0xAB}, "TLS_PSK_WITH_CHACHA20_POLY1305_SHA256 (PSK-CHACHA20-POLY1305)"},
  {{0xCC,0xAC}, "TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256 (ECDHE-PSK-CHACHA20-POLY1305)"},
  {{0xCC,0xAD}, "TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256 (DHE-PSK-CHACHA20-POLY1305)"},
  {{0xCC,0xAE}, "TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256 (RSA-PSK-CHACHA20-POLY1305)"},
  {{0,0}, 0}
};

static const char *curves[] = {
  "-",
  "sect163k1",
  "sect163r1",
  "sect163r2",
  "sect193r1",
  "sect193r2",
  "sect233k1",
  "sect233r1",
  "sect239k1",
  "sect283k1",
  "sect283r1",
  "sect409k1",
  "sect409r1",
  "sect571k1",
  "sect571r1",
  "secp160k1",
  "secp160r1",
  "secp160r2",
  "secp192k1",
  "secp192r1",
  "secp224k1",
  "secp224r1",
  "secp256k1",
  "secp256r1",
  "secp384r1",
  "secp521r1"
};

typedef struct buffer {
  char *s;     /* first char in bufspace */
  char *c;     /* cursor */
  char *t;     /* first char after bufspace */
  int fd;      /* er... maybe the file descriptor? ;-) */
  uint32_t l;  /* for read buffers */
} buffer;

#ifndef EOF
#define EOF (-1)
#endif

#define READBUFSIZE 4096
#define WRITEBUFSIZE 16384
static char buf1space[WRITEBUFSIZE+1];
static buffer *buf1;

static int64_t str2uint63(register const char *s) {
  register int64_t i;
  register char c;

  c = *s;
  if ((c < '0') || (c > '9')) return -1;
  ++s;
  i = c - '0';
  for(;;) {
    c = *s;
    if ((c < '0') || (c > '9')) {
      return i;
    } else {
      i = (i * 10) + c - '0';
      ++s;
    }
  }
}

static int buffer_put(buffer *b, register const char *a, register uint32_t l) {
  register char *p;
  register char *s;
  register const char *t;

  if (!l) return 0;
  if (!b) return -1;
  p = b->c;
  t = b->t;
  if ((p + l) <= t) goto put;

  s = b->s;
  if (write(b->fd, s, p-s) < 0) return -1;
  p = s;
  if ((p + l) > t) {
    b->c = p;
    return write(b->fd, a, l);
  }

  put:
  t = a + l;
  for(;;) {
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
  }
  b->c = p;
  return 0;
}

static int buffer_putuint32(buffer *b, uint32_t num) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;
  char buf[10];

  if (!b) return -1;
  i = num; o = 0;
  pbuf = buf;
  for (d = 1000000000; d > 1; d /= 10) {
    c = i / d;
    if ((c > 0) || (o)) {
      *pbuf = c + '0';
      ++pbuf;
      i = i - (c * d);
      o = 1;
    }
  }
  *pbuf = i + '0'; ++pbuf;
  return buffer_put(b, buf, pbuf-buf);
}

static int buffer_putc(buffer *b, register const char a) {
  register char *p;
  register char *s;

  if (!b) return -1;
  p = b->c;
  if (p < b->t) goto put;

  s = b->s;
  if (write(b->fd, s, p - s) < 0) return -1;
  p = s;

  put:
  *p = a;
  ++p;
  b->c = p;
  return 0;
}

static int buffer_puthex8(buffer *b, register uint8_t a) {
  register uint8_t digit;
  buffer_putc(b, '0');
  buffer_putc(b, 'x');
  digit = a / 16;
  a = a - (digit * 16);
  buffer_putc(b, (digit > 9) ? digit + '7' : digit + '0');
  digit = a;
  buffer_putc(b, (digit > 9) ? digit + '7' : digit + '0');
  return 0;
}

static int buffer_putstr(buffer *b, register const char *a) {
  register char *p;
  register char *s;
  register const char *t;
  register uint32_t l;

  if ((!a) || (!b)) return -1;
  l = strlen(a);
  if (!l) return 0;
  p = b->c;
  t = b->t;
  if ((p + l) <= t) goto put;

  s = b->s;
  if (write(1, s, p-s) < 0) return -1;
  p = s;
  if ((p + l) > t) {
    b->c = p;
    return write(1, a, l);
  }

  put:
  t = a + l;
  for(;;) {
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
    *p = *a; ++p; ++a; if (a >= t) break;
  }
  b->c = p;
  return 0;
}

static int buffer_flush(buffer *b) {
  register char *p;
  ssize_t bytes;

  if (!b) return 1;
  p = b->s;
  if (p == b->c) return 0;
  for (;;) {
    bytes = write(b->fd, p, b->c - p);
    if (bytes == -1) return 1;
    p += bytes;
    if (p == b->c) break;
  }
  b->c = b->s;
  return 0;
}

static void buffer_putversion(buffer *b, uint16_t version) {
  buffer_putstr(b, "Version: ");
  switch(version) {
    case 0x300: buffer_putstr(b, "SSLv3"); break;
    case 0x301: buffer_putstr(b, "TLSv1"); break;
    case 0x302: buffer_putstr(b, "TLSv1.1"); break;
    case 0x303: buffer_putstr(b, "TLSv1.2"); break;
    case 0x304: buffer_putstr(b, "TLSv1.3"); break;
    default: buffer_putstr(b, "(unknown "); buffer_putuint32(b, version); buffer_putc(b, ')'); break;
  }
  buffer_putc(b, '\n');
}

static void my_exit(const int retval) {
  buffer_flush(buf1);
  _exit(retval);
}

int main(int argc, char *argv[]) {
  register unsigned char *p;
  int i;
  int sockfd;
  int connsock;
  struct sockaddr_in servin;
  struct sockaddr_in peerin;
  socklen_t peerin_len;
  char net_buf_rcv[READBUFSIZE + 1];
  uint16_t net_servport;
  int64_t value;
  int read_len;
  uint16_t record_len;
  uint16_t message_len;
  uint16_t session_id_len;
  uint16_t cipher_suite_len;
  uint16_t extensions_len;
  uint16_t tmplen;
  int last_byte;

  if (argc < 2) {
    net_servport = 443;
  } else {
    value = str2uint63(argv[1]);
    if (value < 0) _exit(1);
    net_servport = (uint16_t) value;
  }

  p = (unsigned char *) buf1space;
  buffer buf = { (char *)p, (char *)p, (char *)p + WRITEBUFSIZE, 1 };
  buf1 = &buf;

  connsock = -1;
  sockfd = socket(PF_INET, SOCK_STREAM, 0);
  if (sockfd == -1) my_exit(2);

  i = 1;
  i = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i);
  if (i) my_exit(3);

  memset(&servin, 0, sizeof servin);
  servin.sin_family = AF_INET;
  servin.sin_port = htons(net_servport);
  if (bind(sockfd, (struct sockaddr *) &servin, sizeof servin)) my_exit(4);
  if (listen(sockfd, 2)) my_exit(5);
  
  i = 65536;
  if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &i, sizeof i)) my_exit(6);

  peerin_len = sizeof peerin;
  memset(&peerin, 0, peerin_len);

  p = (unsigned char *) net_buf_rcv;
  if((connsock = accept(sockfd, (struct sockaddr *) &peerin, &peerin_len)) <= 0) my_exit(7);
  (void) close(sockfd);
  read_len = (int)read(connsock, p, READBUFSIZE);
  (void) close(connsock);
  if (read_len < 10) my_exit(8);
/*
  buffer_putstr(buf1, "Read: ");
  buffer_putuint32(buf1, read_len);
  buffer_putc(buf1, '\n');
*/

  /* type */
  if (p[0] != 22) my_exit(9);

  /* version */
  value = p[1] * 256 + p[2];
  buffer_putversion(buf1, value);
  
  /* Record Length */
  record_len = p[3] * 256 + p[4];
  last_byte = record_len + 5;
  if (last_byte > read_len) my_exit(10);
  --last_byte;

  /* message type: must be 1 for Client Hello */
  if (p[5] != 1) my_exit(11);

  /* print Record Length */
  buffer_putstr(buf1, "Record Length: ");
  buffer_putuint32(buf1, record_len);
  buffer_putc(buf1, '\n');

  /* message length */
  message_len = p[6];
  message_len = 256 * message_len + p[7];
  message_len = 256 * message_len + p[8];
  buffer_putstr(buf1, "Message Length: ");
  buffer_putuint32(buf1, message_len);
  buffer_putc(buf1, '\n');

  /* version */
  value = p[9] * 256 + p[10];
  buffer_putversion(buf1, value);

  /* ServerRandom */
  value = p[11];
  value = 256 * value + p[12];
  value = 256 * value + p[13];
  value = 256 * value + p[14];
  buffer_putstr(buf1, "ServerRandom, Time: ");
  buffer_putuint32(buf1, value);
  buffer_putstr(buf1, ",\n");
  { time_t t; t = value; buffer_putstr(buf1, ctime(&t)); }
  value = p[14];
  value = 256 * value + p[13];
  value = 256 * value + p[12];
  value = 256 * value + p[11];
  buffer_putstr(buf1, "(time reversed:) ");
  buffer_putuint32(buf1, value);
  buffer_putstr(buf1, ",\n");
  { time_t t; t = value; buffer_putstr(buf1, ctime(&t)); }

  /* Session ID Length */
  session_id_len = p[43];
  buffer_putstr(buf1, "Session ID Length: ");
  buffer_putuint32(buf1, session_id_len);
  buffer_putc(buf1, '\n');

  /* Cipher Suite Length */
  cipher_suite_len = p[44+session_id_len];
  cipher_suite_len = 256 * cipher_suite_len + p[45+session_id_len];
  buffer_putstr(buf1, "Cipher Suite Length: ");
  buffer_putuint32(buf1, cipher_suite_len);
  buffer_putc(buf1, '\n');

  i = 46 + session_id_len;
  last_byte = i + cipher_suite_len - 1;
  for (;;) {
    uint8_t b1;
    uint8_t b2;
    int tmpint;
    b1 = p[i];
    ++i;
    b2 = p[i];
    tmpint = 0;
    for (;;) {
      if (ciphers[tmpint].description == 0) break;
      if (ciphers[tmpint].value[0] == b1 &&
          ciphers[tmpint].value[1] == b2) {
        buffer_puthex8(buf1, b1); buffer_putc(buf1, ' ');
        buffer_puthex8(buf1, b2); buffer_putc(buf1, ' ');
        buffer_putstr(buf1, ciphers[tmpint].description);
        buffer_putc(buf1, '\n');
        break;
      }
      ++tmpint;
    }
    ++i;
    if (i > last_byte) break;
  }

  /* Compression Methods Length */
  value = p[i]; ++i;
  buffer_putstr(buf1, "Compression Methods Length: ");
  buffer_putuint32(buf1, value);
  buffer_putc(buf1, '\n');
  i += value;

  /* Extensions Length */
  extensions_len = p[i]; ++i;
  extensions_len = 256 * extensions_len + p[i]; ++i;
  buffer_putstr(buf1, "Extensions Length: ");
  buffer_putuint32(buf1, extensions_len);
  buffer_putc(buf1, '\n');
  tmplen = i + extensions_len - 2;

  /* step through extensions */
  for (;;) {
    uint16_t elen;
    if (i >= tmplen) break;

    buffer_putstr(buf1, "Extension: ");
    value = p[i];
    buffer_puthex8(buf1, p[i]); buffer_putc(buf1, ' ');
    ++i;
    value = value * 256 + p[i];
    buffer_puthex8(buf1, p[i]);
    ++i;
    elen = p[i];
    ++i;
    elen = elen * 256 + p[i];
    ++i;
    buffer_putstr(buf1, ", Extension Length: ");
    buffer_putuint32(buf1, elen);
    if (value == 10) {
      buffer_putstr(buf1, ", EC list:");
      uint16_t c = i;
      for (;;) {
        if (c >= i + elen) break;
        value = p[c]; ++c;
        value = value * 256 + p[c]; ++c;
        buffer_putc(buf1, ' ');
        buffer_putstr(buf1, curves[value]);
      }
    }
    buffer_putc(buf1, '\n');

    i += elen;
  }

  my_exit(0);

  return 0;
}
